/*
 * Copyright (C) 2004 Sun Microsystems, Inc.
 *
 * newton.c
 *
 * Newton-Raphson minimizer.
 *
 * Provided by Russell A. Brown (russ.brown@sun.com)
 */

#include <stdio.h>
#include <math.h>
#include "memutil.h"

#undef timers
#ifdef timers
#include <time.h>
#endif

REAL_T mme( REAL_T *, REAL_T *, int * );

/*
 * Adjust the Hessian matrix to put the overall translations and
 * rotations at a relatively high frequency.
 *
 * See Nguyen and Case, J. Phys. Chem. 89, 4020 (1985).
 *
 * Calling parameters are as follows:
 *
 * h      the Hessian matrix to be modified
 * x[n]   the atomic (x,y,z) coordinates
 * m      the atomic masses
 * d      unit vector matrix
 * n      the number of coordinates
 * natom  the number of atoms
 */

static
void level( REAL_T *h, REAL_T *x, REAL_T *m, 
      REAL_T *d, int n, int natom )
{

  REAL_T xx, yy, zz, mass, xsum, ysum, zsum, msum;
  REAL_T vnor1, vnor2, vnor3, vnor4;
  int i, j, ic, iat, i6, j6, jn;

  /* Normalize the translational and rotational degrees of freedom. */

  xsum = ysum = zsum = msum = 0.0;

  for (i = 1; i <= natom; i++) {

    xx = x[3 * (i - 1) + 1];
    yy = x[3 * (i - 1) + 2];
    zz = x[3 * (i - 1) + 3];
    mass = m[i];

    xsum += xx * xx * mass;
    ysum += yy * yy * mass;
    zsum += zz * zz * mass;
    msum += mass;
  }

  vnor1 = sqrt(xsum + ysum);
  vnor2 = sqrt(ysum + zsum);
  vnor3 = sqrt(xsum + zsum);
  vnor4 = sqrt(msum);

  /*
   * Set up the translational and rotational unit vectors.
   * See Eckart, Phys. Rev. 47, 552 (1935).
   */

  for (i = 1; i <= 6 * n; i ++) {
    d[i] = 0.0;
  }

  iat = 1;
  for (i = 1; i <= n; i += 3) {

    i6 = 6 * i;

    xx = x[3 * (iat - 1) + 1];
    yy = x[3 * (iat - 1) + 2];
    zz = x[3 * (iat - 1) + 3];
    mass = sqrt(m[iat]);

    /*
     * For reference, here is the code for d as a matrix:
     *
     *      d[i][1] = mass / vnor4;
     *      d[i+1][2] = mass / vnor4;
     *      d[i+2][3] = mass / vnor4;
     *
     *      d[i][4] = -mass * yy / vnor1;
     *      d[i+1][4] = mass * xx / vnor1;
     *
     *      d[i+1][5] = -mass * zz / vnor2;
     *      d[i+2][5] = mass * yy / vnor2;
     *
     *      d[i][6] = mass * zz / vnor3;
     *      d[i+2][6] = -mass * xx / vnor3;
     */

    d[i6 - 5] = mass / vnor4;
    d[i6 + 2] = mass / vnor4;
    d[i6 + 9] = mass / vnor4;

    d[i6 - 2] = -mass * yy / vnor1;
    d[i6 + 4] = mass * xx / vnor1;

    d[i6 + 5] = -mass * zz / vnor2;
    d[i6 + 11] = mass * yy / vnor2;

    d[i6 + 0] = mass * zz / vnor3;
    d[i6 + 12] = -mass * xx / vnor3;

    iat++;
  }

  /*
   * Now perform the level shift.
   *
   * Before d was converted from a matrix to a vector, the syntax was:
   *
   *      h[i][j] += 3500.0 * d[i][ic] * d[j][ic];
   */

#ifdef MULTIPLE_THREADS
#pragma omp parallel for private (i, i6, j, j6, jn, ic)
#endif

  for (i = 0; i < n; i++) {
    i6 = 6 * i;
    for (j = 0; j < n; j++) {
      j6 = 6 * j;
      jn = n * j;
      for (ic = 1; ic <= 6; ic++) {
        h[i + jn] += 3500.0 * d[i6 + ic] * d[j6 + ic];
      }
    }
  }
}


/* 
 * Newton-Raphson minimizer
 *
 * Calling parameters are as follows:
 *
 * x[n]    contains the variables to be updated
 * n       is the number of variables
 * f       will contain the value of the final value of the function f
 * func    pointer to function that computes the energy, gradient and hessian
 * rmsgrad quit when the rms of the gradient is less than rmsgrad
 * nradd   floating point value to add to the diagonal of the hessian
 * maxiter quit when iter exceeds maxiter
 *
 *
 *
 *     return codes:
 *             >0    converged, final iteration number
 *             -1    exceeded the maximum number of iterations
 *
 */

int newton( REAL_T x[], int *n, REAL_T *f, 
      REAL_T ( *func)( REAL_T*, REAL_T*, REAL_T*,
             REAL_T*, int*, int* ), 
      REAL_T *rmsgrad, REAL_T *nradd, int *maxiter )

{
  REAL_T *d, *g, *m;
  REAL_T *h;
  REAL_T sumg, temp, dgrad, fnew, scale, fmin;
  int niter, i, j, ncopy, natom, nrot, ret_val, odd, lmin, lsearch, iter_mme;
  int *indx;
  char uplo;
  int one = 1;
  int info = 0;

#ifdef timers
  clock_t begintime, endtime;
  unsigned long derivtime, newtontime;
#endif

  /* All arrays, even the x array, will be indexed from 1 to *n. */

  --x;

  /* Allocate some dynamic vectors and matrices. */

  ncopy = *n;
  indx = ivector( 0, ncopy );     /* permutation array    */
  g = vector( 1, ncopy );         /* gradient vector      */
  m = vector( 1, ncopy );         /* atomic masses        */
  d = vector( 1, 6 * ncopy );     /* scratch for level()  */
  h = vector( 0, ncopy*ncopy );   /* hessian (linearized) */

  /* Initialize rms cutoff, iteration counter and timing variables. */

  dgrad = (*rmsgrad) * (*rmsgrad) * ncopy;
  niter = 0;
  iter_mme = -1;

#ifdef timers
  derivtime = newtontime = 0;
  endtime = clock();
#endif

  /* Here is the minimization loop. */

  do {

    /* Increment the iteration counter. */

    ++niter;

#ifdef timers
    begintime = clock();
    newtontime += begintime - endtime;
#endif

    /*
     * Compute the function value in "f",
     * its gradient (with respect to x) in "g", and
     * its hessian (with respect to x) in "h".
     * The atomic masses are returned in "m".
     * The number of atoms is returnd in "natom".
     *
     * The &x[1], &g[1] and &m[1] calling parameters
     * map from 1..n indexing in this newton function
     * to 0..n-1 indexing in *func (which defaults to mme2).
     * This technique is not used for h because it must be
     * indexed from 1 to n.
     */

    *f = ( *func )( &x[ 1 ], &g[ 1 ], h, &m[ 1 ], &natom, &niter );

    /*
     * The time to process the derivative step is
     * the difference between endtime and begintime.
     */

#ifdef timers
    endtime = clock();
    derivtime += endtime - begintime;
#endif

    /*
     * Calculate the rms value (squared) of the gradient vector,
     * and quit if that value is below the rms cutoff.
     */

    sumg = 0.0;
    for (i = 1; i <= ncopy; i++) {
      sumg += g[i] * g[i];
    }
    if (sumg < dgrad) break;

    /*
     * "Level-shift" the eigenvalues that correspond to overall
     * translation and rotation of the molecule to high frequencies.
     * The Hessian matrix h is altered by the level function.
     */

    level(h, x, m, d, ncopy, natom);

    /*  Add a constant to the diagonal of the Hessian, ostensibly to
     *  make it positive definite.
     */

    for (i = 0; i < ncopy; i++) h[i + ncopy*i] += *nradd;


    /* LAPACK solution to the NR equations:  */

    uplo = 'U';
    dposv_( &uplo, n, &one, h, n, &g[1], n, &info );

    if( info ){
      printf( "dposv returns %d; giving up....\n", info );
      exit(1);
    }

    /* update the coordinates:  */
    for (i = 1; i <= ncopy; i++) x[i] -= g[i];

  } while (niter < *maxiter);

#ifdef timer
  printf("\nderivative time = %d   Newton-Raphson time = %d\n",
   derivtime, newtontime);
#endif

  /* Prepare return value. */

  if (niter >= *maxiter) ret_val = -1;
  else ret_val = niter;

  /* Free the dynamic vectors and matrices. */

  free_ivector( indx, 0, ncopy );
  free_vector( g, 1, ncopy );
  free_vector( m, 1, ncopy );
  free_vector( d, 1, 6 * ncopy );
  free_vector( h, 0, ncopy*ncopy );

  return ret_val;
} 
