/*  sff.c:  simple force field: implement routines to read a "prmtop"
 *      file from AMBER, and calculate the energy.  Implements only
 *      some of AMBER's functionality: bonds, angles, dihedrals, and
 *      nonbonded interactions with a distance-dependent dielectric.
 *      Does not (yet) include support for period systems.
 *
 *      Main interface is through routines "mme" and "mme_init".
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <ctype.h>
#include <time.h>
#ifdef MULTIPLE_THREADS
#include <omp.h>
#include <sys/time.h>
#endif

#include "nab.h"
#include "memutil.h"
#include "debug.h"
						   /*  general force field variables   */

#ifdef flex
static char *mmoinputptr;
static int mmoinputlim;
#endif

static REAL_T t1, t2, tnonb, tpair, tbond, tangl, tphi, tcons, teffr, tdiag;
							/* timer variables  */

static char *gopts;			/*  points to the mm_options string  */
static REAL_T cut = 20.0;		/* non-bonded cutoff */
static REAL_T cut2;          /* non-bonded cutoff squared */
static int ntpr = 10;			/*  print frequency  */
static int nsnb = 10;			/*  non-bonded update frequency  */
static REAL_T scnb = 2.0;		/*  scale factor for 1-4 nonbonds  */
static REAL_T scee = 1.2;		/*  scale factor for 1-4 electrostatics  */

static REAL_T rgbmax = 15.0;   /*  maximum dist. for gb radius calculation */
static REAL_T gbalpha, gbbeta, gbgamma;

static PARMSTRUCT_T *prm;

static int *frozen = NULL;          /* frozen atoms */
static int nfrozen = 0;             /* number of frozen atoms */

static int *constrained = NULL;
static int nconstrained = 0;
static REAL_T *x0;		     /*  will hold coordinates to be constrained  */
static REAL_T wcons = 0.0;	     /*  weight of constraints    */

typedef struct kdnode {          /* a kd-tree node */
  INT_T n;
  struct kdnode *lo, *hi;
} KDNODE;

static int *upairs = NULL;     /* array of the number of pairs in upper triangle */
static int *lpairs = NULL;     /* array of the number of pairs in lower triangle */
static int *uprstr = NULL;     /* array of the pairs offset in upper triangle */
static int *lprstr = NULL;     /* array of the pairs offset in lower triangle */
static int *pairlist = NULL;   /* array of the list of pairs */
static int nb_pairs = -1;      /* the total number of pairs */

static int gb = 0;			/* non-zero if generalized Born is to be run */
static int gb_debug = 0;		/* non-zero if extra printout for gen. Born  */
static int e_debug = 0;			/* non-zero if extra printout for energies   */

/*LCPO stuff */

	/* augmented VdW radii, parameters (see JCC 20,2,217 ff.) */
static REAL_T *P0 = NULL, *P1 = NULL, *P2 = NULL, *P3 = NULL, *P4 = NULL;
static int *ineighbor;			/* neighborlist-array */
static int gbsa = 0;
static REAL_T surften = 0.005;	/* 'surface tension' for GB/SA */
static REAL_T esurf = 0.0;		/* nonpolar solvation */

static REAL_T epsext = 78.5;	/* exterior dielectric constant for GB */
static REAL_T kappa = 0.0;		/* Debye-Huckel screeing parameter for GB */

static int maxnb;
static int dield = 1;			/* dielectric function to be used */
static REAL_T dielc = 1.0;      /* dielectric constant for non-GB simulations */
static REAL_T k4d = 0.0;
static int dim;					/* dimension: 3 or 4 */


							/*  MD variables:  */

static REAL_T t = 0.;			/* initial time */
static REAL_T dt = 0.001;		/* time step, ps. */
static REAL_T tautp = 999999.0;	/* temp. coupling parm., ps */
static REAL_T gamma_ln = 0.0;   /* Langevin collison parameter, in ps^-1  */
static REAL_T temp0 = 300.;		/* target temperature, K */
static REAL_T boltz2 = 9.93595e-4;
static REAL_T vlimit = 10.0;	/* maximum velocity component */
static REAL_T genmass = 10.;	/* general masses, for now all the same */
static int ntpr_md = 10;		/* print frequency for KE,PE,temp */
static int ntwx = 0;			/* trajectory snapshot frequency  */
static FILE *binposfp;			/* file pointer for trajectories  */
static int zerov = 0;			/* if true, use zero initial velocities */
static REAL_T tempi = 0.0;		/* initial temperature */
static int idum = -1;			/* random number seed */

static int irattle = 0;			/* turn on/off bond constraints in md */

#ifdef MULTIPLE_THREADS
static int  *Istr;               /* start of data for atom i in prm->Iblo */
static int  *Jblo;               /* j-analog of prm->Iblo */
static int  *JExclAt;            /* j-analog of prm->ExclAt */
static int  *Jstr;               /* start of data for atom j in Jblo */
REAL_T ts, tf;
REAL_T hrt0, hrt1, hrt2, hrt3, hrt4, hrt5, hrt6, hrt7, hrt8, hrt9, hrt10, hrt11;
#endif

REAL_T second();
REAL_T gauss();

#ifdef MULTIPLE_THREADS
REAL_T  hirestime( void );
#endif

int	mme_init( MOLECULE_T *, char *, char *, REAL_T *, FILE * );
REAL_T	mme( REAL_T *, REAL_T *, int * );
REAL_T	mme_rattle( REAL_T *, REAL_T *f, int * );
int	get_masses( REAL_T * );
int	mm_options( char * );
int	md( int,  int, REAL_T *, REAL_T *, REAL_T *, REAL_T (*)() );
int	mme_timer( void );

/*   The energy functions themselves:   */
#include "eff.c"
#include "sff2.c"
#include "sff4.c"

#ifdef MULTIPLE_THREADS
REAL_T hirestime(){

#  ifdef SUN
  hrtime_t nsec;
  nsec = gethrtime();
  return(((REAL_T)nsec)*1.0e-09);
#  else
  REAL_T nsec;
  nsec = 0.0;
  return( nsec );
#  endif
}
#endif

int	mme_init( MOLECULE_T *m, char *aexp, char *aexp2, REAL_T *x0i, FILE *bfpi )

{

  int i, j;
  REAL_T scale, n2o2, natom_float, maxnb_float;
  int at1, at2;
  char atypi[3];
  int *numb;

#ifdef MULTIPLE_THREADS
  int in, iexcl, jexcl, iexcl_last, jexcl_last, nexcl;

  typedef struct atomstr {  /* Structure for temporary atom storage. */
    INT_T num;
    struct atomstr *nxt;
  } ATOMSTR;

  ATOMSTR *Iatoms, *Iatom;  /* Temporary array and pointer for storing atom names */
  ATOMSTR **Jatoms;         /* Array of pointers into the temporary array */
  char *skipar;             /* matrix built from the excluded atom list */
#endif

  dim = 3;
  prm = m->m_prm;
  x0 = x0i;
  binposfp = bfpi;

  /*
   * The excluded atom list atom i is a contiguous set of atom numbers in the
   * prm->ExclAt array that are excluded from pair interaction with atom i.
   * These atom numbers are all greater that i and they are sorted in increasing
   * order.  Create another excluded atom list for atom j where the atom
   * numbers are all less than j and sort them in increasing order.  Also
   * create "start" arrays to store for each atom i and j the starting offset
   * of the contingous data in the excluded atom arrays.  These start arrays
   * remove the loop-carried dependence of excluded atom list expansion and
   * permit parallel processing via OpenMP.  The start and excluded atom arrays
   * are used in the egb and egb2 functions.
   */

#ifdef MULTIPLE_THREADS

  
  /* Initialize the hi resolution timing variables. */
  
  hrt0 = hrt1 = hrt2 = hrt3 = hrt4 = hrt5 = 0.0;
  hrt6 = hrt7 = hrt8 = hrt9 = hrt10 = hrt11 = 0.0;
  ts = hirestime();

  /* Allocate the start arrays and the j-equivalent of the prm->Iblo array. */

  Istr = ivector(0, prm->Natom);
  Jstr = ivector(0, prm->Natom);
  Jblo = ivector(0, prm->Natom);

  /*
   * Walk the prm->Iblo array to load the i-start array and to obtain
   * the total length of the prm->ExclAt array.
   */

  iexcl = 0;
  for (i = 0; i < prm->Natom; i++) {
    Istr[i] = iexcl;
    iexcl += prm->Iblo[i];
  }

  /* Allocate the JExclAt array to be the same length as prm->ExclAt. */

  JExclAt = ivector(0, iexcl);

  /*
   * Allocate the array that will point to the linked lists to be
   * processed and initialize its elements to NULL pointers.
   */

  if ( (Jatoms = (ATOMSTR **) malloc(prm->Natom * sizeof(ATOMSTR *))) == NULL) {
    printf("Error allocating Jatoms array in mme_init!\n");
    exit(1);
  }
  for (i = 0; i < prm->Natom; i++) {
    Jatoms[i] = NULL;
  }
  
  /*
   * Allocate the Iatoms array to have the same number of entries
   * as prm->ExclAt.  This array will store the linked lists.
   */

  if ( (Iatoms = (ATOMSTR *) malloc(iexcl * sizeof(ATOMSTR))) == NULL) {
    printf("Error allocating Iatoms array in mme_init!\n");
  }

  /*
   * Expand the excluded atom list from largest i to smallest i,
   * and get the atom numbers of all atoms j that are excluded
   * from atom i.  For each such atom j that is greater than zero,
   * allocate a record in the Iatoms array and store atom i into
   * the record.  Catenate the record onto the head of the linked
   * list that is stored the jth element of the Jatom array.
   *
   * Each linked list will be ordered by smallest i to largest i.
   *
   * Atom numbers in the excluded atom list lie in the range 1..n
   * whereas array indices lie in the range 0..n-1.
   */

  Iatom = Iatoms;
  for (i = prm->Natom - 1; i >= 0; i--) {
    jexcl = Istr[i];
    jexcl_last = jexcl + prm->Iblo[i];
    for (j = jexcl; j < jexcl_last; j++) {
      nexcl = prm->ExclAt[j] - 1;
      if ( (nexcl >= 0) && (nexcl < prm->Natom) ) {
	Iatom->num = i;
	Iatom->nxt = Jatoms[nexcl];
	Jatoms[nexcl] = Iatom++;
      }
    }
  }

  /*
   * Walk the linked lists from smallest j to largest j, and build
   * in the JExclAt array a contiguous excluded atom list of atoms i
   * that are excluded from atom j.  Store the size of each list in
   * the corresponding entry of the Jblo array.
   *
   * Atom numbers in the excluded atom list lie in the range 1..n
   * whereas array indices lie in the range 0..n-1.
   */

  jexcl = 0;
  for (j = 0; j < prm->Natom; j++) {
    nexcl = 0;
    while (Jatoms[j] != NULL) {
      JExclAt[jexcl++] = Jatoms[j]->num + 1;
      Jatoms[j] = Jatoms[j]->nxt;
      nexcl++;
    }
    Jblo[j] = nexcl;
  }
    
  /* Walk the Jblo array to set the j-start array. */

  jexcl = 0;
  for (j = 0; j < prm->Natom; j++) {
    Jstr[j] = jexcl;
    jexcl += Jblo[j];
  }

  /* Free the temporary arrays. */

  free(Iatoms);
  free(Jatoms);

#endif

  if (gbsa == 1) {
    numb = ivector(0, prm->Natom);
    for (i = 0; i < prm->Natom; i++)
      numb[i] = 0;

    for (i = 0; i < prm->Nbona; i++) {
      at1 = prm->BondAt1[i] / 3;
      at2 = prm->BondAt2[i] / 3;
      numb[at1] += 1;
      numb[at2] += 1;
    }

    free_vector( P0, 0, prm->Natom );
    free_vector( P1, 0, prm->Natom );
    free_vector( P2, 0, prm->Natom );
    free_vector( P3, 0, prm->Natom );
    free_vector( P4, 0, prm->Natom );
    P0 = vector(0, prm->Natom);
    P1 = vector(0, prm->Natom);
    P2 = vector(0, prm->Natom);
    P3 = vector(0, prm->Natom);
    P4 = vector(0, prm->Natom);
    ineighbor = ivector(0, 30 * prm->Natom);
    /* 30 seems sufficient. this should, however, be changed */
    /* if it proves unsafe */
    /* mabye to size 80000, which is well beyond what's needed */
    /* for a whole tRNA */



    for (i = 0; i < prm->Natom; i++) {
      atypi[0] = toupper( prm->AtomSym[i * 4] );
      atypi[1] = toupper( prm->AtomSym[i * 4 + 1] );
      atypi[2] = '\0';


      if (strncmp(atypi, "CT", 2) == 0) {

	if (numb[i] == 1) {
	  P0[i] = 1.70 + 1.40;
	  P1[i] = 0.7887;
	  P2[i] = -0.28063;
	  P3[i] = -0.0012968;
	  P4[i] = 0.00039328;
	} else if (numb[i] == 2) {
	  P0[i] = 1.70 + 1.4;
	  P1[i] = 0.56482;
	  P2[i] = -0.19608;
	  P3[i] = -0.0010219;
	  P4[i] = 0.0002658;
	} else if (numb[i] == 3) {
	  P0[i] = 1.70 + 1.4;
	  P1[i] = 0.23348;
	  P2[i] = -0.072627;
	  P3[i] = -0.00020079;
	  P4[i] = 0.00007967;
	} else if (numb[i] == 4) {
	  P0[i] = 1.70 + 1.4;
	  P1[i] = P2[i] = P3[i] = P4[i] = 0.0;
	}
      } else if (strncmp(atypi, "C", 1) == 0) {
	if (numb[i] == 2) {
	  P0[i] = 1.70 + 1.4;
	  P1[i] = 0.51245;
	  P2[i] = -0.15966;
	  P3[i] = -0.00019781;
	  P4[i] = 0.00016392;
	} else if (numb[i] == 3) {
	  P0[i] = 1.70 + 1.4;
	  P1[i] = 0.070344;
	  P2[i] = -0.019015;
	  P3[i] = -0.000022009;
	  P4[i] = 0.000016875;
	} else if (numb[i] == 4) {
	  P0[i] = 1.70 + 1.4;
	  P1[i] = P2[i] = P3[i] = P4[i] = 0.0;
	} else {
	  P0[i] = 1.70 + 1.4;
	  P1[i] = P2[i] = P3[i] = P4[i] = 0.0;
	  fprintf( stderr, "using default parameters for atom type %s\n", 
		   atypi );
	}
      } else if (strncmp(atypi, "O ", 2) == 0) {
	P0[i] = 1.6 + 1.4;
	P1[i] = 0.68563;
	P2[i] = -0.1868;
	P3[i] = -0.00135573;
	P4[i] = 0.00023743;
      } else if (strncmp(atypi, "O2", 2) == 0) {
	P0[i] = 1.6 + 1.4;
	P1[i] = 0.88857;
	P2[i] = -0.33421;
	P3[i] = -0.0018683;
	P4[i] = 0.00049372;
      } else if (strncmp(atypi, "O", 1) == 0) {
	if (numb[i] == 1) {
	  P0[i] = 1.6 + 1.4;
	  P1[i] = 0.77914;
	  P2[i] = -0.25262;
	  P3[i] = -0.0016056;
	  P4[i] = 0.00035071;
	} else if (numb[i] == 2) {
	  P0[i] = 1.6 + 1.4;
	  P1[i] = 0.49392;
	  P2[i] = -0.16038;
	  P3[i] = -0.00015512;
	  P4[i] = 0.00016453;
	}
      } else if (strncmp(atypi, "N3", 2) == 0) {
	if (numb[i] == 1) {
	  P0[i] = 1.65 + 1.4;
	  P1[i] = 0.078602;
	  P2[i] = -0.29198;
	  P3[i] = -0.0006537;
	  P4[i] = 0.00036247;
	} else if (numb[i] == 2) {
	  P0[i] = 1.65 + 1.4;
	  P1[i] = 0.22599;
	  P2[i] = -0.036648;
	  P3[i] = -0.0012297;
	  P4[i] = 0.000080038;
	} else if (numb[i] == 3) {
	  P0[i] = 1.65 + 1.4;
	  P1[i] = 0.051481;
	  P2[i] = -0.012603;
	  P3[i] = -0.00032006;
	  P4[i] = 0.000024774;
	}
      } else if (strncmp(atypi, "N", 1) == 0) {
	if (numb[i] == 1) {
	  P0[i] = 1.65 + 1.4;
	  P1[i] = 0.73511;
	  P2[i] = -0.22116;
	  P3[i] = -0.00089148;
	  P4[i] = 0.0002523;
	} else if (numb[i] == 2) {
	  P0[i] = 1.65 + 1.4;
	  P1[i] = 0.41102;
	  P2[i] = -0.12254;
	  P3[i] = -0.000075448;
	  P4[i] = 0.00011804;
	} else if (numb[i] == 3) {
	  P0[i] = 1.65 + 1.4;
	  P1[i] = 0.062577;
	  P2[i] = -0.017874;
	  P3[i] = -0.00008312;
	  P4[i] = 0.000019849;
	}
      } else if (strncmp(atypi, "SH", 2) == 0) {
	P0[i] = 1.9 + 1.4;
	P1[i] = 0.7722;
	P2[i] = -0.26393;
	P3[i] = 0.0010629;
	P4[i] = 0.0002179;
      } else if (strncmp(atypi, "S", 1) == 0) {
	P0[i] = 1.9 + 1.4;
	P1[i] = 0.54581;
	P2[i] = -0.19477;
	P3[i] = -0.0012873;
	P4[i] = 0.00029247;
      } else if (strncmp(atypi, "P", 1) == 0) {
	if (numb[i] == 3) {
	  P0[i] = 1.9 + 1.4;
	  P1[i] = 0.3865;
	  P2[i] = -0.18249;
	  P3[i] = -0.0036598;
	  P4[i] = 0.0004264;
	} else if (numb[i] == 4) {
	  P0[i] = 1.9 + 1.4;
	  P1[i] = 0.03873;
	  P2[i] = -0.0089339;
	  P3[i] = 0.0000083582;
	  P4[i] = 0.0000030381;
	}
      } else if (strncmp(atypi, "H", 1) == 0) {
	P0[i] = 1.40;
	P1[i] = P2[i] = P3[i] = P4[i] = 0.0;
      } else {
	P0[i] = 1.70 + 1.4;
	P1[i] = 0.51245;
	P2[i] = -0.15966;
	P3[i] = -0.00019781;
	P4[i] = 0.00016392;
	fprintf(stderr, "Using carbon SA parms for atom type %s\n",
		atypi);
      }

    }
    free_ivector(numb, 0, prm->Natom);

  }
  /*GBSA initialization ends here */

  /*  alter the charges if dielc is not unity  */
  if( dielc != 1.0 ){
    assert( prm != NULL );
    scale = 1.0/sqrt( dielc );
    printf( "scaling charges by %8.3f\n", scale );
    for( i=0; i<prm->Natom; i++ ){
      prm->Charges[i] *= scale;
    }
  }

  free_ivector( frozen, 0, prm->Natom );
  free_ivector( constrained, 0, prm->Natom );
  free_ivector( upairs, 0, prm->Natom );
  free_ivector( lpairs, 0, prm->Natom );
  free_ivector( uprstr, 0, prm->Natom );
  free_ivector( lprstr, 0, prm->Natom );

  frozen = ivector(0, prm->Natom);
  constrained = ivector(0, prm->Natom);
  upairs = ivector(0, prm->Natom);
  lpairs = ivector(0, prm->Natom);
  uprstr = ivector(0, prm->Natom);
  lprstr = ivector(0, prm->Natom);

  if (binposfp != NULL)
    writebinposhdr(binposfp);

  /*
   * Estimate for the number of non-bonded pairs.
   * The maximum number is doubled relative to
   * earlier implementations because the pair list
   * now contains upper and lower triangle pairs.
   */

  free_ivector(pairlist, 0, maxnb);
  nb_pairs = -1;
		
  natom_float = (REAL_T) prm->Natom;
  maxnb_float = 2.0 * natom_float * cut * cut * cut / 1.25;
  n2o2 = natom_float * (natom_float - 1.0);
  maxnb_float = maxnb_float < n2o2 ?  maxnb_float : n2o2;

  if (maxnb_float > 2.147e9){
    printf("Unreasonably large number of pairs: %10.4e\n", maxnb_float);
    exit(1);
  } else {
    maxnb = maxnb_float;
    printf("allocating space for %u non-bonded pairs\n", maxnb);
    pairlist = ivector(0, maxnb);
  }

  nfrozen = set_belly_mask(m, aexp, frozen);
  if (nfrozen)
    printf("freezing %d atoms using expression %s for moving atoms\n",
	   nfrozen, aexp);

  nconstrained = set_cons_mask(m, aexp2, constrained);
  if (nconstrained)
    if (aexp2 == NULL)
      printf("constrained all %d atoms\n", nconstrained);
    else
      printf("constrained %d atoms using expression %s\n",
	     nconstrained, aexp2);

  /*  set up GB/OBC parameters:  */
  if( gb==2 ){
    gbalpha = 0.8; gbbeta = 0.0; gbgamma = 2.909125;
  } else if ( gb==5 ){
    gbalpha = 1.0; gbbeta = 0.8; gbgamma = 4.85;
  }

  return (0);
}

REAL_T	mme( REAL_T *x,  REAL_T *f, int *iter )

{
  REAL_T ebh, eba, eth, eta, eph, epa, enb, eel, enb14, eel14, ecn, e_gb, frms;
  REAL_T ene[20];
  int i, k;

  t1 = second();
  if (*iter == 0) {
    printf
      ("    iter    Total    bad         vdW     elect.     cons.   genBorn      frms\n");
    tnonb = tpair = tbond = tangl = tphi = tcons = 0.0;
  }
  if (nb_pairs < 0 || (*iter>0 && *iter % nsnb == 0)) {
    nb_pairs = nblist(x, lpairs, upairs, lprstr, uprstr, pairlist);
    t2 = second();
    tpair += t2 - t1;
    t1 = t2;
  }

  /* Initialize the gradient vector. */

  for (i = 0; i < 3 * prm->Natom; i++) {
    f[i] = 0.0;
  }

  ebh = ebond(prm->Nbonh, prm->BondHAt1, prm->BondHAt2,
	      prm->BondHNum, prm->Rk, prm->Req, x, f);
  eba = ebond(prm->Mbona, prm->BondAt1, prm->BondAt2,
	      prm->BondNum, prm->Rk, prm->Req, x, f);
  ene[3] = ebh + eba;
  t2 = second();
  tbond += t2 - t1;
  t1 = t2;

  eth = eangl(prm->Ntheth, prm->AngleHAt1, prm->AngleHAt2,
	      prm->AngleHAt3, prm->AngleHNum, prm->Tk, prm->Teq, x, f);
  eta = eangl(prm->Ntheta, prm->AngleAt1, prm->AngleAt2,
	      prm->AngleAt3, prm->AngleNum, prm->Tk, prm->Teq, x, f);
  ene[4] = eth + eta;
  t2 = second();
  tangl += t2 - t1;
  t1 = t2;

  eph = ephi(prm->Nphih, prm->DihHAt1, prm->DihHAt2,
	     prm->DihHAt3, prm->DihHAt4, prm->DihHNum,
	     prm->Pk, prm->Pn, prm->Phase, x, f);
  epa = ephi(prm->Mphia, prm->DihAt1, prm->DihAt2,
	     prm->DihAt3, prm->DihAt4, prm->DihNum,
	     prm->Pk, prm->Pn, prm->Phase, x, f);
  ene[5] = eph + epa;

  ene[6] = 0.0;       /*  hbond term not in Amber-94 force field */
  nbond(prm->N14pairs, prm->N14pairlist, uprstr, 0, x, f, &enb14, &eel14,
	scnb, scee);
  if (e_debug) {
    EXPR("%9.3f", enb14);
    EXPR("%9.3f", eel14);
  }
  ene[7] = enb14;
  ene[8] = eel14;
  t2 = second();
  tphi += t2 - t1;
  t1 = t2;

  if (nconstrained) {
    ecn = econs(x, f);
    t2 = second();
    tcons += t2 - t1;
    t1 = t2;
  } else
    ecn = 0.0;
  ene[9] = ecn;

  if (gb) {
    e_gb = egb(&prm->Natom, x, f, prm->Fs, prm->Rborn, prm->Charges,
		  &kappa, &epsext, &enb, &eel);
    ene[1] = enb;
    ene[2] = eel;
    ene[10] = e_gb;
    ene[11] = esurf;
    if (e_debug) {
      EXPR("%9.3f", enb);
      EXPR("%9.3f", eel);
      EXPR("%9.3f", e_gb);
      EXPR("%9.3f", esurf);
    }
  } else {
    nbond(upairs, pairlist, uprstr, 1, x, f, &enb, &eel, 1.0, 1.0);
    ene[1] = enb;
    ene[2] = eel;
    ene[10] = 0.0;
    ene[11] = 0.0;
    if (e_debug) {
      EXPR("%9.3f", enb);
      EXPR("%9.3f", eel);
    }
    t2 = second();
    tnonb += t2 - t1;
    t1 = t2;
  }

  ene[0] = 0.0;
  for (k = 1; k <= 11; k++)
    ene[0] += ene[k];

  for (k = 0; k < prm->Natom; k++) {	/* zero out frozen forces */
    if (frozen[k]) {
      f[3 * k + 0] = 0.0;
      f[3 * k + 1] = 0.0;
      f[3 * k + 2] = 0.0;
    }
  }

#ifdef PRINT_DERIV
  k = 0;
  for (i = 0; i < 105; i++) {
    k++;
    printf("%10.5f", f[i]);
    if (k % 8 == 0)
      printf("\n");
  }
  printf("\n");
#endif

  frms = 0.0;
  for (i = 0; i < 3 * prm->Natom; i++)
    frms += f[i] * f[i];
  frms = sqrt(frms / (3 * prm->Natom));

  if (*iter > -1 && (*iter == 0 || *iter % ntpr == 0)) {
    printf("ff:%4d %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2e\n",
	   *iter, ene[0], ene[3] + ene[4] + ene[5],
	   ene[1] + ene[7], ene[2] + ene[8], ene[9]+ene[11], ene[10], frms);
    fflush(stdout);
  }

  if( *iter == -1 ){
    printf( "     bond:  %15.9f\n", ene[3] );
    printf( "    angle:  %15.9f\n", ene[4] );
    printf( " dihedral:  %15.9f\n", ene[5] );
    printf( "    enb14:  %15.9f\n", ene[7] );
    printf( "    eel14:  %15.9f\n", ene[8] );
    printf( "      enb:  %15.9f\n", ene[1] );
    printf( "      eel:  %15.9f\n", ene[2] );
    printf( "      egb:  %15.9f\n", ene[10] );
    printf( "    econs:  %15.9f\n", ene[9] );
    printf( "    esurf:  %15.9f\n", ene[11] );
    printf( "    Total:  %15.9f\n", ene[0] );
  }

  return (ene[0]);
}

REAL_T	mme_rattle( REAL_T *x, REAL_T *f, int *iter )
/*
REAL_T *x, *f;
int *iter;
*/
{
	REAL_T ene;


	/*
	 * Version of mme() with rattle for bond length constraints, for
	 * use in minimization. This is quite loosely based on  Duan, Kumar, 
	 * Rosenberg, and Kollman, J. Computat. Chem. 16, 1351-1356 (1995).
	 */

	int dumx, nratH, nrat, Trat, at1, at2, atyp;
	int nat3, i, done, niter, maxiter, imax;
	static REAL_T *minv, *dumv, *deltag;
	REAL_T dtx;
	REAL_T maxf, tmpf;
	REAL_T inP, krat, dist2, xr, yr, zr, delta;
	REAL_T bn, bnx, bny, bnz, frms;

	nat3 = 3 * prm->Natom;
	dtx = 0.001 * 20.455;
	maxiter = 1000;

	niter = 0;
	done = 0;

	/*
	 * inverse masses and velocities are required by the
	 * RATTLE routine that's called below, but are dummy variables
	 * since here only positions are affected
	 */

	if (!minv) {
		minv = (REAL_T *) calloc(nat3, sizeof(REAL_T));
		dumv = (REAL_T *) calloc(nat3, sizeof(REAL_T));
		deltag = (REAL_T *) calloc(nat3, sizeof(REAL_T));
		for (i = 0; i < nat3; i++) {
			minv[i] = 2.0;
			dumv[i] = 0.0;
		}
	}

	for (i = 0; i < nat3; i++)  deltag[i] = 0.0;
	maxf = -1.0;

	/* call the 'position part' of RATTLE to repair bond lengths */
	dumx = rattle(dtx, x, x, dumv, minv);

	/* get energy and gradient for the RATTLEd structure */
	ene = mme(x, f, iter);

	/* repair the gradient */

	nratH = prm->Nbonh;
	nrat = prm->Mbona;
	Trat = nratH + nrat;

	while ((done == 0) && (niter < 50)) {
		niter++;
		done = 1;
		for (i = 0; i < Trat; i++) {

			if (i < nratH) {
				at1 = prm->BondHAt1[i];
				at2 = prm->BondHAt2[i];
				atyp = prm->BondHNum[i] - 1;
				krat = prm->Req[atyp];
			} else {
				at1 = prm->BondAt1[i - nratH];
				at2 = prm->BondAt2[i - nratH];
				atyp = prm->BondNum[i - nratH] - 1;
				krat = prm->Req[atyp];
			}

			xr = x[at2] - x[at1];
			yr = x[at2 + 1] - x[at1 + 1];
			zr = x[at2 + 2] - x[at1 + 2];

			dist2 = xr * xr + yr * yr + zr * zr;
			/* 'inP' comes from the german expression for dot product */
			/* since thats exactly what's calculated here */
			inP =
				xr * (f[at1] - f[at2]) + yr * (f[at1 + 1] - f[at2 + 1]) +
				zr * (f[at1 + 2] - f[at2 + 2]);

#define EPS 0.0000001
			if (inP > EPS) done = 0;
			f[at1] -= (xr / (2.0 * dist2)) * inP;
			f[at1 + 1] -= (yr / (2.0 * dist2)) * inP;
			f[at1 + 2] -= (zr / (2.0 * dist2)) * inP;
			f[at2] += (xr / (2.0 * dist2)) * inP;
			f[at2 + 1] += (yr / (2.0 * dist2)) * inP;
			f[at2 + 2] += (zr / (2.0 * dist2)) * inP;
		}
	}

#ifdef DEBUG
	frms = 0.0;
	for (i = 0; i < nat3 / 3; i++) {
		tmpf =
			f[3 * i] * f[3 * i] + f[3 * i + 1] * f[3 * i + 1] + f[3 * i +
																  2] *
			f[3 * i + 2];
		frms += tmpf;

		if (sqrt(tmpf) > maxf) {
			maxf = sqrt(tmpf);
			imax = i;

		}
	}
	/* the last for loop is pretty inefficient ;-) */
	/* but the output is informative nonetheless */

	frms = sqrt(frms / (REAL_T) nat3);
	printf("Ene %18.14f frms %18.14f\n", ene, frms);
	printf("Max grad %f for atom %d\n", maxf, imax + 1);
	/* output accuracy is a little exaggerated */
	/* we want to see all those digits */
#endif

	return ene;
}

int	get_masses( REAL_T *minv )
									/* dim = 3 or 4, for 3D or 4D problem */
/*
 REAL_T *minv;
*/
{

	int i, k;
	REAL_T am;

	for (k = 0, i = 0; i < prm->Natom; i++) {
		am = 1. / prm->Masses[i];
		minv[k + 0] = am;
		minv[k + 1] = am;
		minv[k + 2] = am;
		if (dim == 4)
			minv[k + 3] = am;
		k += dim;
	}
	return (0);
}

int	mm_options( char *opts )
/*
char *opts;
*/
{
	gopts = opts;
#ifdef flex
	mmoinputlim = strlen(opts);
	mmoinputptr = gopts;
#endif
	mmolex();
	return (0);
}

int	md( int n,  int maxstep, REAL_T *x, REAL_T *f, REAL_T *v, REAL_T ( *mme )() )
/*
 REAL_T(*mme) ();
int n, maxstep;
REAL_T *x, *f, *v;
*/
{

	REAL_T dtx, dt5, rndf, dttp, ekin0, ekin, epot, etot, tscal, temp, sd;
	REAL_T gammai, c_implic, c_explic, sdfac;
	int nstep, i, iat;
	REAL_T zero, invmass;
	static REAL_T *xold, *sqrmass, *accel, *minv;

	/*  space for holding inverse masses  */
	if( !minv ){
		minv = (REAL_T *) calloc(n, sizeof(REAL_T));
		if( minv == NULL ){
			fprintf( stderr, "unable to allocate space for minv\n" );
			exit(1);
		}
	}

	/*  space for holding extra set of coords  */
	if( irattle && !xold ){
		xold = (REAL_T *) calloc(n, sizeof(REAL_T));
		if( xold == NULL ){
			fprintf( stderr, "unable to allocate space for xold\n" );
			exit(1);
		}
	}

	/*  space for holding square root of mass */
	if( (gamma_ln != 0.0) && !sqrmass ){
		sqrmass = (REAL_T *) calloc(n, sizeof(REAL_T));
		if( sqrmass == NULL ){
			fprintf( stderr, "unable to allocate space for sqrmass\n" );
			exit(1);
		}
	}

	/*  space for holding the accelerations */
	if( !accel ){
		accel = (REAL_T *) calloc(n, sizeof(REAL_T));
		if( accel == NULL ){
			fprintf( stderr, "unable to allocate space for accel\n" );
			exit(1);
		}
	}

	dtx = dt * 20.455;
	dt5 = 0.5 * dtx;
	rndf = n - 3 * nfrozen;
	if( irattle ) rndf -= prm->Nbonh + prm->Mbona;
	ekin0 = boltz2 * rndf * temp0;	/* target kinetic energy  */
	dttp = dt / tautp;
	zero = 0.0;
	gammai = gamma_ln/20.455;
	c_implic = 1.0/(1.0+gammai*dt5);
	c_explic = 1.0 - gammai*dt5;
	sdfac = sqrt( 4.0*gammai*boltz2*temp0/dtx );

	invmass = 1. / genmass;
	if (prm){
		get_masses(minv);
	} else { /* if no prmtop file is present, see all masses to genmass */
		for (i = 0; i < n; i++) minv[i] = invmass;
	}

	if (zerov) {
		for (i = 0; i < n; i++)
			v[i] = 0.0;
		ekin = 0.0;
	} else if (tempi > 0.0) {
		ekin = 0.0;
		for (i = 0; i < n; i++) {
			if ( frozen[i / dim] ) {
				v[i] = 0.0;
			} else {
				sd = sqrt(2. * boltz2 * tempi * minv[i]);
				v[i] = gauss(&zero, &sd, &idum);
				ekin += v[i] * v[i] / minv[i];
			}
		}
		ekin *= 0.5;
		temp = ekin / (boltz2 * rndf);

	} else {
		for (ekin = 0., i = 0; i < n; i++)
			ekin += v[i] * v[i] * minv[i];
		ekin *= 0.5;
	}

/*  main loop for the dynamics:  Use velocity Verlet algorithm  */

	nstep = 0;
	epot = mme(x, f, &nstep);

	if( gammai == 0.0 ){  
		for (i = 0; i < n; i++){
			accel[i] = -f[i] * minv[i] * dt5;
		}
	} else {
		for (i = 0; i < n; i++){
			if( frozen[ i/dim ] ){
				accel[i] = 0.0;
			} else {
				sqrmass[i] = 1.0/sqrt(minv[i]);
				sd = sdfac*sqrmass[i];
				accel[i] = (-f[i] + gauss(&zero, &sd, &idum)) * minv[i] * dt5;
			}
		}
	}


	for (nstep = 1; nstep <= maxstep; nstep++) {

		if (ekin > 0.01)
			tscal = sqrt(1. + dttp * (ekin0 / ekin - 1.));
		else
			tscal = 1.0;
		ekin = 0.0;

		if( irattle )
			for (i = 0; i < n; i++) xold[i] = x[i];

		/*  update v[i] by half-timestep, then x[] by a full timestep:  */
		if( gammai == 0.0 ){
			for (i = 0; i < n; i++) {
				v[i] = (v[i] + accel[i]) * tscal;
				v[i] = v[i] > vlimit ? vlimit : v[i];
				v[i] = v[i] < -vlimit ? -vlimit : v[i];
				x[i] += v[i] * dtx;
			}
		} else {
			for (i = 0; i < n; i++) {
				v[i] = c_explic*v[i] + accel[i];
				v[i] = v[i] > vlimit ? vlimit : v[i];
				v[i] = v[i] < -vlimit ? -vlimit : v[i];
				x[i] += v[i] * dtx;
			}
		}

		if( irattle ) rattle( dtx, x, xold, v, minv );

		/*  get the new energies and forces:  */
		epot = mme(x, f, &nstep);

		/*  update the velocities by the second half-timestep:  */
		if( gammai == 0.0 ){
			for (i = 0; i < n; i++){
				accel[i] = -f[i] * minv[i] * dt5;
				v[i] = (v[i] + accel[i]) * tscal;
			}
		} else {
			for (i = 0; i < n; i++){
				if( !frozen[ i/dim ] ){
					sd = sdfac*sqrmass[i];
					accel[i] = (-f[i] + gauss(&zero, &sd, &idum)) * minv[i] * dt5;
					v[i] = (v[i] + accel[i]) * c_implic;
				}
			}
		}

		if( irattle ) rattle2( dtx, x, v, minv );

		/*  get the kinetic and total energies; update the time:  */
		for (i = 0; i < n; i++) ekin += v[i]*v[i]/minv[i];
		ekin *= 0.5;
		etot = ekin + epot;
		temp = ekin / (boltz2 * rndf);
		t += dt;

		if (nstep % ntpr_md == 0 || nstep == 1) {
			printf("md:       %5d %10.3f %10.2f %10.2f %10.2f %10.2f\n",
				   nstep, t, ekin, epot, ekin + epot, temp);
			fflush(stdout);
		}
		if (ntwx > 0 && nstep % ntwx == 0 && binposfp != NULL)
			writebinposfrm(n / 3, x, binposfp);
	}

#ifdef MULTIPLE_THREADS
	tf = hirestime();
	hrt0 += tf - ts;
	ts = tf;
#endif

#undef PRINT_HR_TIME
#if defined(PRINT_HR_TIME) && defined(MULTIPLE_THREADS)
	printf("\nsingle thread  = %8.3f\n\n",
	       hrt0 + hrt6 + hrt7 + hrt8 + hrt9 + hrt11);
	printf("md, mme, etc.  = %8.3f\n", hrt0);
	printf("build kd tree  = %8.3f\n", hrt9);
	printf("bond stretch 1 = %8.3f\n", hrt6);
	printf("bond bending 1 = %8.3f\n", hrt7);
	printf("bond torsion 1 = %8.3f\n", hrt8);
	printf("non-bond int 1 = %8.3f\n", hrt11);
	printf("\nmulti thread   = %8.3f\n\n",
	       hrt1 + hrt2 + hrt3 + hrt4 + hrt5 + hrt10);
	printf("search kd tree = %8.3f\n", hrt10);
	printf("Born 1 loop 1  = %8.3f\n", hrt1);
	printf("Born 1 loop 2  = %8.3f\n", hrt2);
	printf("Born 1 loop 3  = %8.3f\n", hrt3);
	printf("Born 1 loop 4  = %8.3f\n", hrt4);
	printf("Born 1 loop 5  = %8.3f\n", hrt5);
#endif

	return (0);
}

int	mme_timer( void )
{
	printf("\nTiming summary:\n");
	printf("   bonds       %8.1f\n", tbond);
	printf("   angles      %8.1f\n", tangl);
	printf("   torsions    %8.1f\n", tphi);
	printf("   pairlist    %8.1f\n", tpair);
	if( teffr > 0.0 ) printf("   eff. radii  %8.1f\n", teffr);
	printf("   nonbonds    %8.1f\n", tnonb);
	if( tdiag > 0.0 ) printf("   diag. GB    %8.1f\n", tdiag);
	printf("   constraints %8.1f\n", tcons);
	printf("   Total       %8.1f\n\n",
		   tbond + tangl + tphi + tpair + teffr + tnonb + tdiag + tcons);
	return (0);
}

/*   #include "shiftnbond.c"   */
#include "rattle.c"
#include "lex.mm_options.c"
