/*
 * Copyright (C) 2004 Sun Microsystems, Inc.
 *
 *  sff.c:  simple force field: implement routines to read a "prmtop"
 *      file from AMBER, and calculate the energy.  Implements only
 *      some of AMBER's functionality: bonds, angles, dihedrals, and
 *      nonbonded interactions with a distance-dependent dielectric.
 *      Does not (yet) include support for period systems.
 *
 *      Interface is through routine "mme2"
 *
 *      Second derivatives added to egb2 by Russ Brown (russ.brown@sun.com).
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "nab.h"
#include "memutil.h"
#include "debug.h"

REAL_T econs2( REAL_T *x, REAL_T *f, REAL_T *g )

{
  int i,n;
  REAL_T e_cons, rx, ry, rz;

  n = 3*prm->Natom;
  e_cons = 0.0;
  for (i = 0; i < prm->Natom; i++) {
    if (constrained[i]) {
      rx = x[3 * i + 0] - x0[3 * i + 0];
      ry = x[3 * i + 1] - x0[3 * i + 1];
      rz = x[3 * i + 2] - x0[3 * i + 2];
      e_cons += wcons * (rx * rx + ry * ry + rz * rz);
      f[3 * i + 0] += 2. * wcons * rx;
      f[3 * i + 1] += 2. * wcons * ry;
      f[3 * i + 2] += 2. * wcons * rz;

      g[3*i + 0 + n*(3*i + 0)] += 2. * wcons;
      g[3*i + 1 + n*(3*i + 1)] += 2. * wcons;
      g[3*i + 2 + n*(3*i + 2)] += 2. * wcons;
    }
  }
  return (e_cons);
}

/* Calculate bond stretching energy and derivatives. */

REAL_T ebond2( int nbond, int *a1, int *a2, int *atype,
	       REAL_T *Rk, REAL_T *Req, REAL_T *x, REAL_T *f, REAL_T *g)

{
  int i, k, l, n, at1, at2, atyp;
  REAL_T e_bond, r, r2, rinv, r2inv, rx, ry, rz, dfx, dfy, dfz, db, df, dg, e;
  REAL_T di[3], dj[3], d2ii[3][3], d2jj[3][3], d2ij[3][3], ijterm;

  n = 3*prm->Natom;
  e_bond = 0.0;
  for (i = 0; i < nbond; i++) {
    at1 = a1[i];
    at2 = a2[i];
    atyp = atype[i] - 1;

    /* Calculate the interatomic distance and its (x,y,z) components. */

    rx = x[at1 + 0] - x[at2 + 0];
    ry = x[at1 + 1] - x[at2 + 1];
    rz = x[at1 + 2] - x[at2 + 2];
    r2 = rx * rx + ry * ry + rz * rz;
    r2inv = 1.0 / r2;
    r = sqrt(r2);
    rinv = r * r2inv;

    /*
     * Calculate the first and second derivatives of the interatomic
     * distance Dij with respect to the cartesian coordinates of atoms
     * i and j.  The results are placed into five arrays:
     *
     *   di[] for the first derivatives with respect to atom i
     *   dj[] for the first derivatives with respect to atom j
     *   d2ii[] for the second derivatives with respect to atom i
     *   d2jj[] for the second derivatives with respect to atom j
     *   d2ij[] for the second derivatives with respect to atoms i and j
     *
     * Some useful symmetry obtains.  The d2ii, d2jj and d2ij arrays
     * are symmetric in that their lower and upper triangles are
     * the transposes of one another.  Also, d2jj equals d2ii, and
     * d2ij is the negative of d2ii.
     *
     * As was exploited by David Case and his colleagues, the vector dj
     * is the negative of the vector di.
     */
      
    di[0] = rx;
    di[1] = ry;
    di[2] = rz;

    dj[0] = -rx;
    dj[1] = -ry;
    dj[2] = -rz;

    /* Load the upper triangle of d2ii. */

    d2ii[0][0] = 1.0 - rx * rx * r2inv;
    d2ii[0][1] = -rx * ry * r2inv;
    d2ii[0][2] = -rx * rz * r2inv;

    d2ii[1][1] = 1.0 - ry * ry * r2inv;
    d2ii[1][2] = -ry * rz * r2inv;

    d2ii[2][2] = 1.0 - rz * rz * r2inv;

    /* Finish loading the rest of all of the matrices. */

    for (k = 0; k < 3; k++) {

      /* Load the upper triangles of d2jj and d2ij. */

      for (l = k; l < 3; l++) {
	d2jj[k][l] = d2ii[k][l];
	d2ij[k][l] = -d2ii[k][l];
      }

      /* Load the symmetric elements of d2ii, d2jj and d2ij. */

      for (l = k+1; l < 3; l++) {
	d2ii[l][k] = d2ii[k][l];
	d2jj[l][k] = d2jj[k][l];
	d2ij[l][k] = d2ij[k][l];
      }
    }

    /*
     * Calculate the bond stretching energy as:
     *
     *      e = Rk*(r - Req)^2
     */

    db = r - Req[atyp];
    df = Rk[atyp] * db;
    e = df * db;
    e_bond += e;

    /*
     * For the calculation of the 1st derivatives of the energy, the
     * 1st derivatives of the bond distance are multiplied by df:
     *
     *      de/dx = df*(dr/dx)
     *
     * where ds/dx represents the 1st derivative of the bond distance, and: 
     *
     *      df = de/dr = 2*Rk*(r - Req)
     *
     * because the chain rule gives:
     *
     *      de/dx = (de/dr)*(dr/dx) = df*(dr/dx)
     *
     * Divide df by r because this factor is not present in dr/dx.
     */

    df *= 2.0 * rinv;

    dfx = df * rx;
    dfy = df * ry;
    dfz = df * rz;

    f[at1 + 0] += dfx;
    f[at1 + 1] += dfy;
    f[at1 + 2] += dfz;
    f[at2 + 0] -= dfx;
    f[at2 + 1] -= dfy;
    f[at2 + 2] -= dfz;

    /*
     * For the calculation of the 2nd derivatives of the energy, the 1st and 2nd
     * derivatives of the bond distance are multiplied by dg and df as follows:
     *
     *      d2e/dx2dx1 = df*(d2r/dx2dx1) + dg*(dr/dx2)*(dr/dx1)
     *
     * where d2s/dx2dx1 represents a 2nd derivative of the bond distance,
     * dr/dx1 and dr/dx2 represent 1st derivatives of the bond distance, and:
     *
     *      dg = (d/dr)df = (d/dr)(de/dr) = 2*Rk
     *
     * because the chain rule gives:
     *
     *      d2e/dx2dx1 = (d/dx2)(de/dx1)
     *
     *                 = (d/dx2)[(de/dr)*(dr/dx1)]
     *
     *                 = (d/dx2)[df*(dr/dx1)]
     *
     *                 = [(d/dx2)df]*(dr/dx1) + df*[(d/dx2)(dr/dx1)]
     *
     *                 = [(d/dx2)(de/dr)]*(dr/dx1) + df*(d2r/dx2dx1)
     *
     *                 = {[(d/dr)(de/dr)]*(dr/dx2)}*(dr/dx1) + df*(d2r/dx2dx1)
     *
     *                 = dg*(dr/dx2)*(dr/dx1) + df*(d2r/dx2dx1)
     *
     * Divide dg by r^2 because this factor is not present in (dr/dx1)*(dr/dx2).
     */

    dg = 2.0 * Rk[atyp] * r2inv;

    for (k = 0; k < 3; k++) {
      for (l = 0; l < 3; l++) {
        g[at1 + k + n*(at1 + l)] += df * d2ii[k][l] + dg * di[k] * di[l];
        g[at2 + k + n*(at2 + l)] += df * d2jj[k][l] + dg * dj[k] * dj[l];

	ijterm = df * d2ij[k][l] + dg * di[k] * dj[l];
        g[at1 + k + n*(at2 + l)] += ijterm;
        g[at2 + l + n*(at1 + k)] += ijterm;
      }
    }
  }

  if (e_debug)
    EXPR("%9.3f", e_bond);
  return (e_bond);
}

void difang(REAL_T cst, REAL_T * s, REAL_T * dc, REAL_T ** ddc)
{

	/*
	 *     ----- sets up second derivatives for angle-like potentials -----
	 *           cst ... cosine of theta
	 *           s   ... bond vectors ij and kj
	 *           dc  ... first der. of costheta w/respct to the cartesian
	 *                   differences in s
	 *           ddc ... output of second derivative of costheta w/respect
	 *                   to the cartesian differences
	 */

	REAL_T bij, boi2, boij3, boji3, bkj, boj2, br, bt;
	int i, i1, j;

	/*     ----- first set up distances needed -----   */

	bij = s[1]*s[1] + s[2]*s[2] + s[3]*s[3];
	boi2 = 1.0 / bij;
	bij = sqrt(bij);
	bkj = s[4]*s[4] + s[5]*s[5] + s[6]*s[6];
	boj2 = 1.0 / bkj;
	bkj = sqrt(bkj);
	br = boi2;

	/*
	 *     ----- set up ddc first as second derivative of costheta w/respect
	 *           to cartesian differences -----
	 */

	for (i = 1; i <= 6; i++) {
		if (i == 4) br = boj2;
		ddc[i][i] = -br*(2.0*s[i]*dc[i] + cst*(1.0 - s[i]*s[i]*br));
	}

	for (i = 1; i <= 3; i++) {
		ddc[i][i+3] = -boi2*s[i]*dc[i+3] - boj2*s[i+3]*dc[i]
			- boi2*boj2*(s[i]*s[i+3]*cst - bij*bkj);
	}

	for (i = 1; i <= 2; i++) {
		i1 = i + 1;
		for (j = i1; j <= 3; j++) {
			ddc[i][j] =
				-boi2*(s[i]*dc[j] + s[j]*dc[i] - boi2*cst*s[i]*s[j]);
		}
	}

	for (i = 4; i <= 5; i++) {
		i1 = i + 1;
		for (j = i1; j <= 6; j++) {
			ddc[i][j] =
				-boj2*(s[i]*dc[j] + s[j]*dc[i] - boj2*cst*s[i]*s[j]);
		}
	}

	boij3 = boj2/(bij*bkj);
	boji3 = boi2/(bij*bkj);
	bt = boi2*boj2*cst;
	for (i = 1; i <= 3; i++) {
		for (j = 4; j <= 6; j++) {
			if (j != i + 3)
				ddc[i][j] =
					-boij3*s[i+3]*s[j] - boji3*s[i]*s[j-3] + bt*s[i]*s[j];
		}
	}

	/*    ----- symmetrize the ddc array -----   */

	for (i = 1; i <= 6; i++) {
		for (j = i; j <= 6; j++) {
			ddc[j][i] = ddc[i][j];
		}
	}
	return;
}

/* Calculate the bond bending energy and derivatives.*/

REAL_T eangl2( int nang, int *a1, int *a2, int *a3, int *atype,
	       REAL_T * Tk, REAL_T * Teq, REAL_T * x, REAL_T * f,
	       REAL_T * g )
{
	int i, j, n, ia, atyp, at1, at2, at3, ist, iof, inew, jst, jof, jnew;
	REAL_T x1, y1, z1, x2, y2, z2, x3, y3, z3;
	REAL_T c, st, st2c, dtheta, theta, df, ddf, e, e_theta;
	REAL_T rij, rkj, rij2, rkj2, rrik;

	REAL_T s[7], dc[7], dr[10];
	REAL_T **ddc, **ddr;

	ddc = matrix(1,6,1,6);
	ddr = matrix(1,9,1,9);

	n = 3*prm->Natom;
	e_theta = 0.0;
	for (ia = 0; ia < nang; ia++) {
		at1 = a1[ia];
		at2 = a2[ia];
		at3 = a3[ia];
		atyp = atype[ia] - 1;

		/* Set up the variables x1..x3, y1..y3 and z1..z3 for use later. */

		x1 = x[at1 + 0];
		y1 = x[at1 + 1];
		z1 = x[at1 + 2];
		x2 = x[at2 + 0];
		y2 = x[at2 + 1];
		z2 = x[at2 + 2];
		x3 = x[at3 + 0];
		y3 = x[at3 + 1];
		z3 = x[at3 + 2];

		s[1] = x1 - x2;
		s[2] = y1 - y2;
		s[3] = z1 - z2;
		s[4] = x3 - x2;
		s[5] = y3 - y2;
		s[6] = z3 - z2;
		rij2 = s[1]*s[1] + s[2]*s[2] + s[3]*s[3];
		rkj2 = s[4]*s[4] + s[5]*s[5] + s[6]*s[6];
		rij = sqrt(rij2);
		rkj = sqrt(rkj2);
		rrik = rij * rkj;
		c = (s[1]*s[4] + s[2]*s[5] + s[3]*s[6]) / rrik;

		c = c > 1.0 ? 1.0 : c;
		c = c < -1.0 ? -1.0 : c;
		theta = acos(c);
		dtheta = theta - Teq[atyp];

		/* df and ddf are derivatives of E with respect to c == costheta: */
		df = dtheta * Tk[atyp];
		ddf = 2. * Tk[atyp];
		e = df * dtheta;
		e_theta += e;

		/*
		 * Calculate the sine then limit small values of the sine to the range
		 * -10**-3..+10**-3 so as to avoid division by zero.
		 */

		st = sin(theta);
		if (st > 0 && st < 1.e-3)
			st = 1.e-3;
		else if (st < 0 && st > -1.e-3)
			st = -1.e-3;
		df *= 2.0;   /* check this!!  */

		/* dc = derivative of c with respect to cartesian differences: */

		for( i=1; i<=3; i++ ){
			dc[i] = (s[i+3]/rkj - c*s[i]/rij)/rij;
			dc[i+3] = (s[i]/rij - c*s[i+3]/rkj)/rkj;
		}

		/* get ddc = second derivative of c with respect to 
		   cartesian differences:  */
		difang( c, s, dc, ddc );

		/* change ddc to second. der. of theta with respect to 
		   cartesian differences:  */
		st2c = c/(st*st);
		for( i=1; i<=6; i++ ){
			for( j=i; j<=6; j++ ){
				ddc[i][j] = -(ddc[i][j] + dc[i]*dc[j]*st2c)/st;
				ddc[j][i] = ddc[i][j];
			}
		}

		/* change dc to derivative of theta w/ respect to cartesian
		   differences:  */
		for( i=1; i<=6; i++ ){
			dc[i] = -dc[i]/st;
		}

		/* dr will hold -derivates of theta w/ respect to cartesians: */
		for( i=1; i<=3; i++ ){
			dr[i] = -dc[i];
			dr[i+6] = -dc[i+3];
			dr[i+3] = dc[i] + dc[i+3];
		}

		/* update the forces:  */
		f[at1 + 0] -= df*dr[1];
		f[at1 + 1] -= df*dr[2];
		f[at1 + 2] -= df*dr[3];
		f[at2 + 0] -= df*dr[4];
		f[at2 + 1] -= df*dr[5];
		f[at2 + 2] -= df*dr[6];
		f[at3 + 0] -= df*dr[7];
		f[at3 + 1] -= df*dr[8];
		f[at3 + 2] -= df*dr[9];

		
		/* ddr will hold second derivatives of theta with respect to carts: */
		for( i=1; i<=3; i++ ){
			for( j=i; j<=3; j++ ){
				ddr[i][j] = ddc[i][j];
				ddr[i+6][j+6] = ddc[i+3][j+3];
			}
			for( j=1; j<=3; j++ ){
				ddr[i][j+6] = ddc[i][j+3];
				ddr[i][j+3] = -ddc[i][j] - ddc[i][j+3];
				ddr[i+3][j+6] = -ddc[i+3][j] - ddc[i+3][j+3];
			}
		}
		for( i=1; i<=3; i++ ){
			for( j=i; j<=3; j++ ){
				ddr[i+3][j+3] = -ddr[i][j+3] - ddr[i+3][j+6];
			}
		}
/*
		for( i=1; i<=9; i++ ){
			for( j=i; j<=9; j++ ){
				ddr[j][i] = ddr[i][j];
			}
		}
*/

		for( i=1; i<=9; i++ ){
			ist = at3;
			if( i<=6 ) ist = at2;
			if( i<=3 ) ist = at1;
			iof = i%3 ; if( iof == 0 ) iof =3;
			inew = ist + iof - 1;

			for( j=i; j<=9; j++ ){
				jst = at3;
				if( j<=6 ) jst = at2;
				if( j<=3 ) jst = at1;
				jof = j%3; if( jof == 0 ) jof =3;
				jnew = jst + jof - 1;

				g[inew + n*jnew] += ddf*dr[i]*dr[j] + df*ddr[i][j];
				if( inew != jnew )
					g[jnew + n*inew] += ddf*dr[i]*dr[j] + df*ddr[i][j];
			}
		}
	}

	free_matrix( ddc, 1,6,1,6 );
	free_matrix( ddr, 1,6,1,6 );

	if (e_debug)
		EXPR("%9.3f", e_theta);
	return (e_theta);
}

/* Calculate the dihedral torsion energy and derivatives. */

REAL_T ephi2( int nphi, int *a1, int *a2,
	      int *a3, int *a4, int *atype,
	      REAL_T * Pk, REAL_T * Pn, REAL_T * Phase,
	      REAL_T * x, REAL_T * f, REAL_T * g )
{
	int i, j, k, l, n, at1, at2, at3, at4, atyp, iper,iphi;
	int inew, ist, iof, jnew, jst, jof;
	REAL_T e, df, ddf, e_tors;
	REAL_T xij, yij, zij, xkj, ykj, zkj, xkl, ykl, zkl;
	REAL_T dx, dy, dz, gx, gy, gz, bi, bk, ct, boi2, boj2, z1, z2, bioj, bjoi;
	REAL_T ct2, arg, z11, z12, z22;
	REAL_T t[7], dc[7], dr[13];
	REAL_T dum,s1,s2,s3,s4,s5,s6,s7,s8,s9,s10,s11,s12;
	REAL_T **dtx, **ddr, **ddc;

	ddc = matrix( 1,6, 1,6 );
	dtx = matrix( 1,6, 1,12 );
	ddr = matrix( 1,12, 1,12 );

	n = 3*prm->Natom;
	e_tors = 0.0;
	for (iphi = 0; iphi < nphi; iphi++) {

		/*
		 * Get the atom numbers.  Atom 3 may be negative to specify that
		 * the end group interactions are to be ignored.  Atom 4 may be 
		 * negative to specify that the dihedral angle is an "improper
		 * torsion".  However, neither of these conditions is actually
		 * treated in this function, so just ignore negative values.
		 */

		at1 = a1[iphi];
		at2 = a2[iphi];
		at3 = abs(a3[iphi]);
		at4 = abs(a4[iphi]);
		atyp = atype[iphi] - 1;


		/*     ----- calculation of the dihedral -----     */


		xij = x[at1 + 0] - x[at2 + 0];
		yij = x[at1 + 1] - x[at2 + 1];
		zij = x[at1 + 2] - x[at2 + 2];
		xkj = x[at3 + 0] - x[at2 + 0];
		ykj = x[at3 + 1] - x[at2 + 1];
		zkj = x[at3 + 2] - x[at2 + 2];
		xkl = x[at3 + 0] - x[at4 + 0];
		ykl = x[at3 + 1] - x[at4 + 1];
		zkl = x[at3 + 2] - x[at4 + 2];

		dx = yij * zkj - zij * ykj;
		dy = zij * xkj - xij * zkj;
		dz = xij * ykj - yij * xkj;
		gx = zkj * ykl - ykj * zkl;
		gy = xkj * zkl - zkj * xkl;
		gz = ykj * xkl - xkj * ykl;

		bi = dx * dx + dy * dy + dz * dz;
		bk = gx * gx + gy * gy + gz * gz;
		ct = dx * gx + dy * gy + dz * gz;

		/*    ----- branch if linear dihedral -----   */

		assert(bi > 0.01);
		assert(bk > 0.01);

		boi2 = 1. / bi;
		boj2 = 1. / bk;
		bi = sqrt(bi);
		bk = sqrt(bk);
		z1 = 1. / bi;
		z2 = 1. / bk;
		bioj = bi * z2;
		bjoi = bk * z1;
		ct = ct * z1 * z2;
		if (ct > 1.)
			ct = 1.;
		if (ct < (-1.))
			ct = -1.;

		/*     ----- ct value above is actually -cosphi; here we change
		 *           its sign -----
		 */
		ct = -ct;

		/*     ----- calculate the energy and derivatives -----


		 *     ----- get df = first der. of potential w/respect to cosphi; and
		 *           ddf = second der. of potntial w/respect to cosphi -----

		 *           the torsional potential is assumed to have the form:
		 *            e = pk(ic) * (1.0+phase*cos(pn(ic)*phi)
		 *            where phase = 1.0 or -1.0, and pn = 1,2,3,4, or 6

		 *     ----- energy terms for dihedrals are expressed in terms of
		 *           cosphi in order to eliminate problems for planar angles ----
		 */

	  multi_term:
		iper = fabs(Pn[atyp]) + 0.0001;
		assert(iper != (fabs(Pn[atyp]) - 0.0001));
		assert(iper >= 1 && iper <= 6);
		switch (iper) {

		case 1:
			e = ct;
			df = 1.0;
			ddf = 0.0;
			break;

		case 2:
			e = 2.0 * ct * ct - 1.0;
			df = 4.0 * ct;
			ddf = 4.0;
			break;

		case 3:
			ct2 = ct * ct;
			e = ct * (4.0 * ct2 - 3.0);
			df = 12.0 * ct2 - 3.0;
			ddf = 24.0 * ct;
			break;

		case 4:
			ct2 = ct * ct;
			e = 1.0 + ct2 * 8.0 * (ct2 - 1.0);
			df = 16.0 * ct * (ct2 + ct2 - 1.0);
			ddf = 16.0 * (6.0 * ct2 - 1.0);
			break;

		case 6:
			ct2 = ct * ct;
			e = ct2 * (ct2 * (ct2 * 32.0 - 48.0) + 18.0) - 1.0;
			df = ct * (ct2 * (ct2 * 192.0 - 192.0) + 36.0);
			ddf = ct2 * (ct2 * 960.0 - 576.0) + 36.0;

		default:
			fprintf(stderr, "bad periodicity: %d\n", iper);
			exit(1);
		}

		if (Phase[atyp] == 0.0) {
			arg = Pk[atyp];
		} else {
			arg = -Pk[atyp];
			assert(fabs(3.14159 - Phase[atyp]) < 0.01);
		}

		e = Pk[atyp] + arg * e;
		df = df * arg;
		ddf = ddf * arg;

		e_tors += e;

		t[1] = dx;
		t[2] = dy;
		t[3] = dz;
		t[4] = -gx;
		t[5] = -gy;
		t[6] = -gz;

		/*
		 *     ----- now, set up array dc = first der. of cosphi w/respect
		 *           to the cartesian differences t -----
		 */

		z11 = z1 * z1;
		z12 = z1 * z2;
		z22 = z2 * z2;
		for (i = 1; i <= 3; i++) {
			dc[i] = t[i+3] * z12 - ct * t[i] * z11;
			dc[i+3] = t[i] * z12 - ct * t[i+3] * z22;
		}

		/*
		 *     ----- subroutine difang will now create array ddc which is second
		 *           derivative of cosphi with respect to the t s -----
		 */

		difang(ct, t, dc, ddc);

		/* ---now set up array s, given on page 118 of cff book-- */

		s1 = xij;
		s2 = yij;
		s3 = zij;
		s4 = xkj;
		s5 = ykj;
		s6 = zkj;
		s7 = -xkj;
		s8 = -ykj;
		s9 = -zkj;
		s10 = -xkl;
		s11 = -ykl;
		s12 = -zkl;

		/*
		 *     ----- set up dtx[i][j] = derivative of t[i] w/respect to x[j]
		 *           see p. 120 of cff book -----
		 */

		for (i = 1; i <= 6; i++) {
			for (j = 1; j <= 12; j++) {
				dtx[i][j] = 0.0;
			}
		}

		dtx[1][2] = s6;
		dtx[1][3] = -s5;
		dtx[1][5] = s3 - s6;
		dtx[1][6] = s5 - s2;
		dtx[1][8] = -s3;
		dtx[1][9] = s2;
		dtx[2][1] = -s6;
		dtx[2][3] = s4;
		dtx[2][4] = s6 - s3;
		dtx[2][6] = s1 - s4;
		dtx[2][7] = s3;
		dtx[2][9] = -s1;
		dtx[3][1] = s5;
		dtx[3][2] = -s4;
		dtx[3][4] = s2 - s5;
		dtx[3][5] = s4 - s1;
		dtx[3][7] = -s2;
		dtx[3][8] = s1;
		dtx[4][5] = s12;
		dtx[4][6] = -s11;
		dtx[4][8] = s9 - s12;
		dtx[4][9] = s11 - s8;
		dtx[4][11] = -s9;
		dtx[4][12] = s8;
		dtx[5][4] = -s12;
		dtx[5][6] = s10;
		dtx[5][7] = s12 - s9;
		dtx[5][9] = s7 - s10;
		dtx[5][10] = s9;
		dtx[5][12] = -s7;
		dtx[6][4] = s11;
		dtx[6][5] = -s10;
		dtx[6][7] = s8 - s11;
		dtx[6][8] = s10 - s7;
		dtx[6][10] = -s8;
		dtx[6][11] = s7;

		/*
		 *     ----- set up dr array, containing -first derivative of cosphi with
		 *           respect to cartesians -----
		 */

		for (i = 1; i <= 12; i++) {
			dum = 0.0;
			for (j = 1; j <= 6; j++) {
				dum += dc[j]*dtx[j][i];
			}
			dr[i] = -dum;
		}
       

		/*     ----- update the force array ----- */

		f[at1 + 0] -= df * dr[1];
		f[at1 + 1] -= df * dr[2];
		f[at1 + 2] -= df * dr[3];
		f[at2 + 0] -= df * dr[4];
		f[at2 + 1] -= df * dr[5];
		f[at2 + 2] -= df * dr[6];
		f[at3 + 0] -= df * dr[7];
		f[at3 + 1] -= df * dr[8];
		f[at3 + 2] -= df * dr[9];
		f[at4 + 0] -= df * dr[10];
		f[at4 + 1] -= df * dr[11];
		f[at4 + 2] -= df * dr[12];

		/*
		 *     ----- now set up the ddr array = second der. of cosphi w/respect
		 *           to cartesians; first we take the first term of last formula
		 *           on p. 113 of cff book -----
		 */

		for (i = 1; i <= 12; i++) {
			for (j = i; j <= 12; j++) {
				ddr[i][j] = 0.0;
				for (k = 1; k <= 6; k++) {
					for (l = 1; l <= 6; l++) {
						ddr[i][j] = 
						    ddr[i][j] + ddc[k][l]*dtx[k][i]*dtx[l][j];
					}
				}
			}
		}

		/*     ----- now do the second term of this equation -----  */

		ddr[2][9] += dc[1];
		ddr[3][5] += dc[1];
		ddr[6][8] += dc[1];
		ddr[2][6] -= dc[1];
		ddr[3][8] -= dc[1];
		ddr[5][9] -= dc[1];
		ddr[1][6] += dc[2];
		ddr[3][7] += dc[2];
		ddr[4][9] += dc[2];
		ddr[1][9] -= dc[2];
		ddr[3][4] -= dc[2];
		ddr[6][7] -= dc[2];
		ddr[1][8] += dc[3];
		ddr[2][4] += dc[3];
		ddr[5][7] += dc[3];
		ddr[1][5] -= dc[3];
		ddr[2][7] -= dc[3];
		ddr[4][8] -= dc[3];
		ddr[5][12] += dc[4];
		ddr[6][8] += dc[4];
		ddr[9][11] += dc[4];
		ddr[5][9] -= dc[4];
		ddr[6][11] -= dc[4];
		ddr[8][12] -= dc[4];
		ddr[4][9] += dc[5];
		ddr[6][10] += dc[5];
		ddr[7][12] += dc[5];
		ddr[4][12] -= dc[5];
		ddr[6][7] -= dc[5];
		ddr[9][10] -= dc[5];
		ddr[4][11] += dc[6];
		ddr[5][7] += dc[6];
		ddr[8][10] += dc[6];
		ddr[4][8] -= dc[6];
		ddr[5][10] -= dc[6];
		ddr[7][11] -= dc[6];

		/*     ----- Now form the second derivative matrix -----  */

		for (i = 1; i <= 12; i++) {
			ist = at4;
			if (i <= 9) ist = at3;
			if (i <= 6) ist = at2;
			if (i <= 3) ist = at1;
			iof = i % 3; if (iof == 0) iof = 3;
			inew = ist + iof - 1;

			for (j = i; j <= 12; j++) {
				jst = at4;
				if (j <= 9) jst = at3;
				if (j <= 6) jst = at2;
				if (j <= 3) jst = at1;
				jof = j % 3; if (jof == 0) jof = 3;
				jnew = jst + jof - 1;

				g[inew + n*jnew] += ddf*dr[i]*dr[j] + df*ddr[i][j];
				if (inew != jnew)
					g[jnew + n*inew] += ddf*dr[i]*dr[j] + df*ddr[i][j];
			}
		}


#ifdef PRINT_EPHI
		printf("%4d %4d %4d %4d %4d %9.4f\n", i + 1, at1 / 3, at2 / 3,
			   at3 / 3, at4 / 3, e);
#endif

		/* A negative value of Pn means that more terms are needed in the expansion. */

		if (Pn[atyp] < 0) {
			atyp++;
			goto multi_term;
		}
	}
	if (e_debug)
		EXPR("%9.3f", e_tors);

	free_matrix( dtx, 1,12, 1,12 );
	free_matrix( ddr, 1,12, 1,12 );

	return (e_tors);
}

/* 
 * Calculate the non-bonded energy, 1st and 2nd derivatives.
 * This function is complicated by the fact that it must
 * process two forms of pair lists: the 1-4 pair list and
 * the non-bonded pair list.  The non-bonded pair list
 * must be modified by the excluded atom list whereas the
 * 1-4 pair list is used unmodified.  Also, the non-bonded
 * pair list comprises lower and upper triangles whereas
 * the 1-4 pair list comprises an upper triangle only.
 *
 * Calling parameters are as follows:
 *
 * npairs - the number of pairs on either pair list
 * pairlist - either the 1-4 pair list or the non-bonded pair list
 * offset - the starting address of the upper triangle; ignored
 *          for the 1-4 pair list
 * non14 - set to 1 for the non-bonded pair list, 0 for the 1-4 pair list
 * x - the atomic coordinate array
 * f - the gradient vector
 * g - the Hessian matrix mapped onto a linear vector
 * enb - Van der Waals energy return value, passed by reference
 * eel - Coulombic energy return value, passed by reference
 * enbfac - scale factor for Van der Waals energy
 * eelfac - scale factor for Coulombic energy
 */

int nbond2( int *npairs, int *pairlist, int *offset, int non14,
	    REAL_T *x, REAL_T *f, REAL_T *g, REAL_T *enb, REAL_T *eel,
	    REAL_T enbfac, REAL_T eelfac )
{
  int i, j, k, l, n, jn, ic, npr, lpair, iaci, iexcl, nexcl;
  int *iexw;
  REAL_T r, r2, r2inv, df2, dg2, r6, r10, f1, f2, df, dg;
  REAL_T dis, kij, d0, diff, rinv, rs, rssq, eps1, epsi, cgijr, pow, cgi;
  REAL_T dumx, dumy, dumz, enbfaci, eelfaci;
  int nhbpair, ibig, isml;
#if 0
  REAL_T hbener, nb14;
  nhbpair = 0;
  hbener = 0.0; 
  nb14 = 0.0;
#endif

  REAL_T xi, yi, zi, xij, yij, zij;

  REAL_T di[3], dj[3], d2ii[3][3], d2jj[3][3], d2ij[3][3], ijterm;

#define SIG 0.3
#define DIW 78.0
#define C1 38.5

  /*
   * For the various methods of calculating the nonbonded energy
   * both the energy and its 1st and 2nd derivatives are calculated.
   *
   * For the calculation of the 1st derivatives of the energy, the
   * 1st derivatives the nonbonded distance r are multiplied by df:
   *
   *      de/dx = df*(dr/dx)
   *
   * where dr/dx represents the 1st derivative of the distance, and 
   * 
   *      df = de/dr
   *
   * is required by the chain rule:
   *
   *      de/dx = (de/dr)*(dr/dx) = df*(dr/dx)
   *
   * For the calculation of the 2nd derivatives of the energy, the 1st and
   * 2nd derivatives of the distance are multiplied by dg and df as follows:
   *
   *      d2e/dxjdxi = df*(d2r/dxjdxi) + dg*(dr/dxj)*(dr/dxi)
   *
   * where d2r/dxjdxi represents a 2nd derivative of the distance,
   * dr/dxi and dr/dxj represent 1st derivatives of the distance, and:
   *
   *      dg = (d/dr)df
   *
   * is required by the chain rule:
   *
   *      d2e/dxjdxi = (d/dxj)(de/dxi)
   *
   *                 = (d/dxj)[(de/dr)*(dr/dxi)]
   *
   *                 = (d/dxj)[df*(dr/dxi)]
   *
   *                 = [(d/dxj)df]*(dr/dxi) + df*[(d/dxj)(dr/dxi)]
   *
   *                 = [(d/dxj)(de/dr)]*(dr/dxi) + df*(d2r/dxjdxi)
   *
   *                 = {[(d/dr)(de/dr)]*(dr/dxj)}*(dr/dxi) + df*(d2r/dxjdxi)
   *
   *                 = dg*(dr/dxj)*(dr/dxi) + df*(d2r/dxjdxi)
   *
   */

  n = 3*prm->Natom;
  *enb = 0.;
  *eel = 0.;
  enbfaci = 1. / enbfac;
  eelfaci = 1. / eelfac;

  /* Initialize offsets into the pair list and the excluded atom list. */

  lpair = 0;
  iexcl = 0;

  /*
   * Allocate and initialize the iexw array used for skipping excluded
   * atoms.  Note that because of the manner in which iexw is used, it
   * is necessary to initialize it before only the first iteration of
   * the following loop.
   */

  iexw = ivector(-1, prm->Natom);
  for (i = -1; i < prm->Natom; i++) {
    iexw[i] = -1;
  }

  /* Loop over all atoms i except for the final atom i. */

  for (i = 0; i < prm->Natom - 1; i++) {

    /*
     * If the 'non14' calling parameter is set, ignore the offset
     * into the pair list that accumulates with each iteration of
     * the i-loop, and use instead the beginning address of the upper
     * triangle of the pair list.
     */

    if (non14 != 0) {
      lpair = offset[i];
    }

    /* Check whether there are any atoms j on the pair list of atom i. */

    nexcl = prm->Iblo[i];
    npr = npairs[i];
    if (npr > 0) {
      iaci = prm->Ntypes * (prm->Iac[i] - 1);
      dumx = 0.0;
      dumy = 0.0;
      dumz = 0.0;
      xi = x[3 * i + 0];
      yi = x[3 * i + 1];
      zi = x[3 * i + 2];
      cgi = eelfaci * prm->Charges[i];

      /*
       * Expand the excluded list into the iexw array by storing i
       * at array address j.
       */

      for (j = 0; j < nexcl; j++) {
	iexw[prm->ExclAt[iexcl + j] - 1] = i;
      }

      /* Select atoms j from the pair list. */

      for (jn = 0; jn < npr; jn++) {
	j = pairlist[jn + lpair];

	/*
	 * If the 'non14' calling parameter is set, check whether
	 * this i,j pair is exempted by the excluded atom list.
	 */

	if (non14 == 0 || iexw[j] != i) {
	  xij = xi - x[3 * j + 0];
	  yij = yi - x[3 * j + 1];
	  zij = zi - x[3 * j + 2];
	  r2 = xij * xij + yij * yij + zij * zij;
	  r2inv = 1.0 / r2;
	  r = sqrt(r2);
	  rinv = r * r2inv;

	  /*
	   * Calculate the first and second derivatives of the interatomic
	   * distance Dij with respect to the cartesian coordinates of atoms
	   * i and j.  The results are placed into five arrays:
	   *
	   *   di[] for the first derivatives with respect to atom i
	   *   dj[] for the first derivatives with respect to atom j
	   *   d2ii[] for the second derivatives with respect to atom i
	   *   d2jj[] for the second derivatives with respect to atom j
	   *   d2ij[] for the second derivatives with respect to atoms i and j
	   *
	   * Some useful symmetry obtains.  The d2ii, d2jj and d2ij arrays
	   * are symmetric in that their lower and upper triangles are
	   * the transposes of one another.  Also, d2jj equals d2ii, and
	   * d2ij is the negative of d2ii.
	   *
	   * As was exploited by David Case and his colleagues, the vector dj
	   * is the negative of the vector di.
	   */
      
	  di[0] = xij;
	  di[1] = yij;
	  di[2] = zij;

	  dj[0] = -xij;
	  dj[1] = -yij;
	  dj[2] = -zij;

	  /* Load the upper triangle of d2ii. */

	  d2ii[0][0] = 1.0 - xij * xij * r2inv;
	  d2ii[0][1] = -xij * yij * r2inv;
	  d2ii[0][2] = -xij * zij * r2inv;

	  d2ii[1][1] = 1.0 - yij * yij * r2inv;
	  d2ii[1][2] = -yij * zij * r2inv;

	  d2ii[2][2] = 1.0 - zij * zij * r2inv;

	  /* Finish loading the rest of all of the matrices. */

	  for (k = 0; k < 3; k++) {

	    /* Load the upper triangles of d2jj and d2ij. */

	    for (l = k; l < 3; l++) {
	      d2jj[k][l] = d2ii[k][l];
	      d2ij[k][l] = -d2ii[k][l];
	    }

	    /* Load the symmetric elements of d2ii, d2jj and d2ij. */

	    for (l = k+1; l < 3; l++) {
	      d2ii[l][k] = d2ii[k][l];
	      d2jj[l][k] = d2jj[k][l];
	      d2ij[l][k] = d2ij[k][l];
	    }
	  }

	  /* Calculate the energy and derivatives according to dield. */

	  if (dield == -3) {

	    /* special code Ramstein & Lavery dielectric, 94 force field */

	    rs = SIG * r;
	    rssq = rs * rs;
	    pow = exp(-rs);
	    eps1 = rssq + rs + rs + 2.0;
	    epsi = 1.0 / (DIW - C1 * pow * eps1);
	    cgijr = cgi * prm->Charges[j] * rinv * epsi;
	    *eel += cgijr;
	    df2 = -cgijr * (1.0 + C1 * pow * rs * rssq * epsi);
	    dg2 =
	      cgijr*(epsi*(C1*pow*SIG*(rssq*(2.0*(epsi*C1*pow*SIG*rssq + rinv)
					     + SIG) - 2.0*SIG*rs)) + 2.0*r2inv);
	    ic = prm->Cno[iaci + prm->Iac[j] - 1] - 1;
	    if( ic >= 0 ){
	      r6 = r2inv * r2inv * r2inv;
	      f2 = prm->Cn2[ic] * r6;
	      f1 = prm->Cn1[ic] * r6 * r6;
	      *enb += (f1 - f2) * enbfaci;
	      df = (df2 + (6.0 * f2 - 12.0 * f1) * enbfaci) * rinv;
	      dg = dg2 + (156.0 * f1 - 42.0 * f2 * enbfaci) * r2inv;
	    } else {
	      df = df2 * rinv;
	      dg = dg2;
	    }

	  } else if (dield == -4) {

	    /* distance-dependent dielectric code, 94 ff */
	    /* epsilon = r  */

	    rs = cgi * prm->Charges[j] * r2inv;
	    df2 = -2.0 * rs;
	    dg2 = 6.0 * rs * r2inv;
	    *eel += rs;
	    ic = prm->Cno[iaci + prm->Iac[j] - 1] - 1;
	    if( ic >= 0 ){
	      r6 = r2inv * r2inv * r2inv;
	      f2 = prm->Cn2[ic] * r6;
	      f1 = prm->Cn1[ic] * r6 * r6;
	      *enb += (f1 - f2) * enbfaci;
	      df = (df2 + (6.0 * f2 - 12.0 * f1) * enbfaci) * rinv;
	      dg = dg2 + (156.0 * f1 - 42.0 * f2 * enbfaci) * r2inv;
	    } else {
	      df = df2 * rinv;
	      dg = dg2;
	    }

	  } else if (dield == -5) {

	    /* non-bonded term from yammp  */

	    dis = r;
	    ic = prm->Cno[iaci + prm->Iac[j] - 1] - 1;
	    d0 = prm->Cn2[ic];
	    if (dis < d0) {
	      kij = prm->Cn1[ic];
	      diff = dis - d0;
	      *enb += kij * diff * diff;
	      df = 2.0 * kij * diff;
	      dg = 2.0 * kij;
	    } else {
	      df = 0.0;
	      dg = 0.0;
	    }
	  } else {

	    /*
	     * Code for various dielectric models.
	     * The df2 variable should hold r(dV/dr).
	     */

	    if (dield == 0) {

	      /* epsilon = r  */

	      rs = cgi * prm->Charges[j] * r2inv;
	      df2 = -2.0 * rs;
	      dg2 = 6.0 * rs * r2inv;
	      *eel += rs;

	    } else if (dield == 1) {
	      
	      /* epsilon = 1  */

	      rs = cgi * prm->Charges[j] * rinv;
	      df2 = -rs;
	      dg2 = 2.0 * rs * r2inv;
	      *eel += rs;

	    } else if (dield == -2) {

	      /* Ramstein & Lavery dielectric, PNAS 85, 7231 (1988). */

	      rs = SIG * r;
	      rssq = rs * rs;
	      pow = exp(-rs);
	      eps1 = rssq + rs + rs + 2.0;
	      epsi = 1.0 / (DIW - C1 * pow * eps1);
	      cgijr = cgi * prm->Charges[j] * rinv * epsi;
	      *eel += cgijr;
	      df2 = -cgijr * (1.0 + C1 * pow * rs * rssq * epsi);
	      dg2 =
		cgijr*(epsi*(C1*pow*SIG*(rssq*(2.0*(epsi*C1*pow*SIG*rssq + rinv)
					       + SIG) - 2.0*SIG*rs)) + 2.0*r2inv);
	    }

	    /* Calculate either Van der Waals or hydrogen bonded term. */

	    ic = prm->Cno[iaci + prm->Iac[j] - 1];
	    if (ic > 0 || enbfac != 1.0) {
	      if (ic > 0) {
		ic--;
	      } else {
		ibig = prm->Iac[i] > prm->Iac[j] ?
		  prm->Iac[i] : prm->Iac[j];
		isml = prm->Iac[i] > prm->Iac[j] ?
		  prm->Iac[j] : prm->Iac[i];
		ic = ibig * (ibig - 1) / 2 + isml - 1;
	      }
	      r6 = r2inv * r2inv * r2inv;
	      f2 = prm->Cn2[ic] * r6;
	      f1 = prm->Cn1[ic] * r6 * r6;
	      *enb += (f1 - f2) * enbfaci;
	      df = (df2 + (6.0 * f2 - 12.0 * f1) * enbfaci) * rinv;
	      dg = dg2 + (156.0 * f1 - 42.0 * f2) * enbfaci * r2inv;
#if 0
	      if( enbfac != 1.0 ) nb14 += (f1-f2)*enbfaci;
#endif
	    } else {
	      ic = -ic - 1;
	      r10 = r2inv * r2inv * r2inv * r2inv * r2inv;
	      f2 = prm->HB10[ic] * r10;
	      f1 = prm->HB12[ic] * r10 * r2inv;
	      *enb += (f1 - f2) * enbfaci;
	      df = (df2 + (10.0 * f2 - 12.0 * f1) * enbfaci) * rinv;
	      dg = dg2 + (156.0 * f1 - 110.0 * f2 * enbfaci) * r2inv;
	      nhbpair++;
#if 0
	      hbener += (f1-f2)*enbfaci;
#endif
	    }
	  }

	  /* Divide by r here instead dividing the derivatives by r. */

	  df *= rinv;
	  dg *= r2inv;

	  /* Update the gradient and the Jacobian. */

	  dumx += df * di[0];
	  dumy += df * di[1];
	  dumz += df * di[2];

	  f[3 * j + 0] += df * dj[0];
	  f[3 * j + 1] += df * dj[1];
	  f[3 * j + 2] += df * dj[2];

	  /*
	   * Calculate the second derivatives of Dij for the Van der Waals
	   * and Coulombic interactions.
	   */

	  for (k = 0; k < 3; k++) {
	    for (l = 0; l < 3; l++) {
	      g[3*i + k + n*(3*i + l)] += df * d2ii[k][l] + dg * di[k] * di[l];
	      g[3*j + k + n*(3*j + l)] += df * d2jj[k][l] + dg * dj[k] * dj[l];

	      ijterm = df * d2ij[k][l] + dg * di[k] * dj[l];
	      g[3*i + k + n*(3*j + l)] += ijterm;
	      g[3*j + l + n*(3*i + k)] += ijterm;
	    }
	  }
	}
      }

      /* For atom i, the gradient is updated in the i-loop only. */

      f[3 * i + 0] += dumx;
      f[3 * i + 1] += dumy;
      f[3 * i + 2] += dumz;
    }

    /* Update the offsets into the pair list and the excluded atom list. */

    lpair += npr;
    iexcl += nexcl;
  }

  /* Deallocate the iexw array. */

  free_ivector(iexw, -1, prm->Natom);

  return (0);
}

/*
 * Calculate the generalized Born energy, 1st and 2nd derivatives.
 *
 * Calling parameters are as follows:
 *
 * natom - input: the number of atoms
 * x - input: the atomic (x,y,z) coordinates
 * f - updated: the gradient vector
 * g - updated: the Hessian matrix
 * fs - input: overlap parameters
 * rborn - input: atomic radii
 * q - input: atomic charges
 * kappa - input: inverse of the Debye-Huckel length
 * diel_ext - input: solvent dielectric constant
 * enb - updated: Lennard-Jones energy
 * eelt - updated: gas-phase electrostatic energy
 */

REAL_T egb2( int *natom, REAL_T *x, REAL_T *f, REAL_T *g, REAL_T *fs,
	     REAL_T *rborn, REAL_T *q, REAL_T *kappa, REAL_T *diel_ext,
	     REAL_T *enb, REAL_T *eelt)

#define BOFFSET 0.09
#define KSCALE 0.73

#define PIx4 12.5663706143591724639918
#define PIx2  6.2831853071795862319959
#define PIx1  3.1415926535897931159979

  /*FGB taylor coefficients follow */
  /* from A to L :                 */
  /* 1/3 , 2/5 , 3/7 , 4/9 , 5/11  */
  /* 4/3 , 12/5 , 24/7 , 40/9 , 60/11 */
  /* 20/3, 84/5, 216/7, 440/9, 780/11 */

#define TA 0.33333333333333333333
#define TB 0.4
#define TC 0.42857142857142857143
#define TD 0.44444444444444444444
#define TDD 0.45454545454545454545

#define TE 1.33333333333333333333
#define TF 2.4
#define TG 3.42857142857142857143
#define TH 4.44444444444444444444
#define THH 5.45454545454545454545

#define TI 6.66666666666666666667
#define TJ 16.8
#define TK 30.8571428571428571429
#define TL 48.8888888888888888889
#define TLL 70.9090909090909090909
{

#ifdef MULTIPLE_THREADS
  int iexcl_last;
#endif

  int i, j, k, l, m, n, i3, in3, in, j3, jn, jn3, nn, n3, nn3;
  int iexcl, jexcl, jexcl_last, nexcl, npairs, ic, iaci;
  int *iexw;
  REAL_T elec, evdw, epol, dielfac, qi, qj, qiqj, fgbi, fgbk, rb2, expmkf;
  REAL_T xi, yi, zi, xj, yj, zj, xij, yij, zij;
  REAL_T dumbo,tmpsd;
  REAL_T df, dg, dx, dy, dz;
  REAL_T dedx, dedy, dedz;
  REAL_T dij1i, dij2i, dij3i, temp1, temp2;
  REAL_T qi2h, qid2h, datmp, da2tmp;
  REAL_T theta, ri1i;

  REAL_T dij, sumi;
  REAL_T eel, f6, f12, rinv, r2inv, r4inv, r6inv;
  REAL_T r2, ri, ri2, riinv, rj, rjinv, rirjinv, rj2, sj, sj2, thi;
  REAL_T uij, efac, temp4, temp5, temp6, temp7;
  REAL_T sumda, sumdn, sumd2;
  REAL_T *reff, *dreff, *sumdeijda, *sumepol, *sumelec, *sumevdw, *psi;
  REAL_T *sumdeijd2, *sumdeijdn, *sumdeijdg, *sumdeijdh;
  REAL_T d_dij, e_dij, f_dij, g_dij, h_dij, i_dij;
  REAL_T j_dij, k_dij, l_dij, m_dij, n_dij, o_dij;
  REAL_T a_dij, b_dij, c_dij, p_dij, q_dij, r_dij;
  REAL_T temp8, temp9, temp10, temp11, temp12, temp13;
  REAL_T temp14, temp15, temp16, temp17, temp18, temp19;
  REAL_T temp20, temp21, temp22, temp23;

  int count, count2, icount;
  REAL_T si, sumAij, sumAjk, sumAijAjk, sumdAijddijdxi;
  REAL_T sumdAijddijdyi, sumdAijddijdzi, sumdAijddijdxiAjk;
  REAL_T sumdAijddijdyiAjk, sumdAijddijdziAjk, rij, tmpaij, Aij,
    dAijddij;
  REAL_T dAijddijdxj, dAijddijdyj, dAijddijdzj;
  REAL_T sumdAjkddjkdxj, sumdAjkddjkdyj, sumdAjkddjkdzj, p3p4Aij;
  REAL_T xk, yk, zk, rjk2, djk1i, rjk, vdw2dif, tmpajk, Ajk, sumAjk2,
    dAjkddjk;
  REAL_T dAjkddjkdxj, dAjkddjkdyj, dAjkddjkdzj, lastxj, lastyj, lastzj;
  REAL_T dAidxj, dAidyj, dAidzj, Ai, Aidxi, Aidyi, Aidzi, dAidxi, dAidyi,
    dAidzi;
  REAL_T totsasa;
  REAL_T rgbmax1i, rgbmax2i, rgbmaxpsmax2;

  REAL_T dreffi, sreffi, sdreff, sumdh, dxj, dyj, dzj;

  REAL_T di[3], dj[3], d2ii[3][3], d2jj[3][3], d2ij[3][3];

  /*
   * Smooth "cut-off" in calculating GB effective radii.
   * Implementd by Andreas Svrcek-Seiler and Alexey Onufriev.
   * The integration over solute is performed up to rgbmax and includes
   * parts of spheres; that is an atom is not just "in" or "out", as
   * with standard non-bonded cut.  As a result, calclated effective
   * radii are less than rgbmax. This saves time, and there is no
   * discontinuity in dReff/drij.
   *
   * Only the case rgbmax > 5*max(sij) = 5*fsmax ~ 9A is handled; this is
   * enforced in mdread().  Smaller values would not make much physical
   * sense anyway.
   *
   * Note: rgbmax must be less than or equal to cut so that the pairlist
   * generated from cut may be applied to calculation of the effective
   * radius and its derivatives.
   */

  if (rgbmax > cut) {
    printf("Error in egb2: rgbmax = %f is greater than cutoff = %f\n",
	   rgbmax, cut);
    exit(1);
  }
  rgbmax1i = 1.0/rgbmax;
  rgbmax2i = rgbmax1i*rgbmax1i;
  rgbmaxpsmax2 = (rgbmax+prm->Fsmax)*(rgbmax+prm->Fsmax);

  /* Compute these because they are used a fair amount. */

  n = *natom;
  nn = n * n;
  n3 = 3 * n;
  nn3 = n * n3;

  /* Allocate some temporary vectors. */

  reff = vector(0, n);
  dreff = vector(0, nn3);

  sumepol = vector(0, n);
  sumelec = vector(0, n);
  sumevdw = vector(0, n);
  sumdeijda = vector(0, n);
  sumdeijd2 = vector(0, n);
  sumdeijdn = vector(0, nn);
  sumdeijdg = vector(0, nn3);
  sumdeijdh = vector(0, nn3);

  if ( gb==2 || gb==5 ) psi = vector(0, n);

  count = 0;
  totsasa = 0.0;

  /* 
   * Get the "effective" Born radii via the approximate pairwise method
   * Use Eqs 9-11 of Hawkins, Cramer, Truhlar, J. Phys. Chem. 100:19824
   * (1996).
   *
   * The computation is not executed in parallel for the case of gbsa==1
   * due to the loop-carried dependence of the 'count' variable.
   */

  if (gb_debug)
    printf("Effective Born radii:\n");

#ifdef MULTIPLE_THREADS
#pragma omp parallel for if (gbsa != 1) \
  private (i, xi, yi, zi, ri, ri1i, sumi, j, k, xij, yij, zij, \
	   r2, dij1i, dij, sj, sj2, uij, dij2i, tmpsd, dumbo, theta)
#endif

  /* Loop over all atoms i. */

  for (i = 0; i < n; i++) {

    xi = x[3 * i + 0];
    yi = x[3 * i + 1];
    zi = x[3 * i + 2];
    ri = rborn[i] - BOFFSET;
    ri1i = 1. / ri;
    sumi = 0.;

    /* Select atom j from the pair list. */

    for (k = lprstr[i]; k < lprstr[i] + lpairs[i] + upairs[i]; k++) {

      j = pairlist[k];

      xij = xi - x[3 * j + 0];
      yij = yi - x[3 * j + 1];
      zij = zi - x[3 * j + 2];
      r2 = xij * xij + yij * yij + zij * zij;
      if ( r2 > rgbmaxpsmax2 ) continue;
      dij1i = 1.0 / sqrt(r2);
      dij = r2 * dij1i;
      sj = fs[j] * (rborn[j] - BOFFSET);
      sj2 = sj * sj;

      /*
       * The following equations are from the Appendix of Schaefer and Froemmel,
       * JMB 216:1045-1066, 1990.  Taylor series expansion for d>>s is by Andreas
       * Svrcek-Seiler.  Smooth rgbmax idea is from Andreas Svrcek-Seiler and
       * Alexey Onufriev.
       */
      
      if ( dij > rgbmax + sj ) continue;

      if ( dij > rgbmax - sj ) {

	uij = 1./(dij - sj);

	sumi -= 0.125 * dij1i * (1.0 + 2.0 * dij * uij +
				 rgbmax2i * (r2 - 4.0 * rgbmax * dij - sj2) -
				 2.0 * log(rgbmax * uij));

      } else if ( dij > 4.0 * sj ) {

	dij2i = dij1i * dij1i;
	tmpsd = sj2 * dij2i;
	dumbo = TA + tmpsd*(TB + tmpsd*(TC + tmpsd*(TD + tmpsd*TDD)));

	sumi -= sj2 * sj * dij2i * dij2i * dumbo;

      } else if ( dij > ri + sj ) {

	sumi -= 0.5 * (sj / (r2 - sj2) +
		       0.5 * dij1i * log((dij - sj) / (dij + sj)));

      } else if ( dij > fabs(ri - sj) ) {

	theta = 0.5 * ri1i * dij1i * (r2 + ri * ri - sj2);
	uij = 1. / (dij + sj);

	sumi -= 0.25 * (ri1i * (2. - theta) - uij +
			dij1i * log(ri * uij));

      } else if ( ri < sj ) {

	sumi -= 0.5 * (sj / (r2 - sj2) + 2. * ri1i +
		       0.5 * dij1i * log((sj - dij) / (sj + dij)));

      }

      if (gbsa == 1) {
	if ((P0[i] + P0[j]) > dij) {
	  if (P0[i] > 2.5 && P0[j] > 2.5) {

	    ineighbor[count] = j + 1;
	    /* this +1 is VERY important */
	    count += 1;
	  }
	}
      }

    }

    if ( gb==1 ){

      /* "standard" (HCT) effective radii:  */
      reff[i] = 1.0/(ri1i + sumi);

    } else {

      /* "gbao" formulas:  */

#define GBALPHA 0.8
#define GBBETA 0.0
#define GBGAMMA 2.909125

      psi[i] = -ri*sumi;
      reff[i] = 1.0 / (ri1i - tanh( (GBALPHA + GBGAMMA*psi[i]*psi[i])
				    *psi[i])/rborn[i] );
    }

    if (gbsa == 1) {
      ineighbor[count] = 0;
      count = count + 1;
    }

    /*      reff[ i ] = rborn[i] > reff[i] ?  rborn[i] : reff[i];   */

    if (gb_debug)
      printf("%d\t%15.7f\t%15.7f\n", i + 1, rborn[i], reff[i]);
  }


  /*
   * Main LCPO stuff follows.  No second derivative support for this.
   * Also no parallelization using OpenMP due to the loop-carried
   * dependence of the 'count' variable.
   */

  if (gbsa == 1) {
    totsasa = 0.0;
    count = 0;
    for (i = 0; i < n; i++) {
      if (ineighbor[count] == 0) {
	count = count + 1;
      } else {

	si = PIx4 * P0[i] * P0[i];
	sumAij = 0.0;
	sumAjk = 0.0;
	sumAjk2 = 0.0;
	sumAijAjk = 0.0;
	sumdAijddijdxi = 0.0;
	sumdAijddijdyi = 0.0;
	sumdAijddijdzi = 0.0;
	sumdAijddijdxiAjk = 0.0;
	sumdAijddijdyiAjk = 0.0;
	sumdAijddijdziAjk = 0.0;

	icount = count;

      L70:j = ineighbor[count] - 1;
				/* mind the -1 , fortran again */
      xi = x[3 * i + 0];
      yi = x[3 * i + 1];
      zi = x[3 * i + 2];

      xj = x[3 * j + 0];
      yj = x[3 * j + 1];
      zj = x[3 * j + 2];



      r2 = (xi-xj) * (xi-xj) + (yi-yj) * (yi-yj) + (zi-zj) * (zi-zj);
      dij1i = 1. / sqrt(r2);
      rij = r2 * dij1i;
      tmpaij =
	P0[i] - rij * 0.5 - (P0[i] * P0[i] -
			     P0[j] * P0[j]) * 0.5 * dij1i;

      Aij = PIx2 * P0[i] * tmpaij;

      dAijddij =
	PIx1 * P0[i] * (dij1i * dij1i *
			(P0[i] * P0[i] - P0[j] * P0[j]) - 1.0);

      dAijddijdxj = dAijddij * (xj - xi) * dij1i;
      dAijddijdyj = dAijddij * (yj - yi) * dij1i;
      dAijddijdzj = dAijddij * (zj - zi) * dij1i;

      sumAij = sumAij + Aij;

      count2 = icount;
      sumAjk2 = 0.0;
      sumdAjkddjkdxj = 0.0;
      sumdAjkddjkdyj = 0.0;
      sumdAjkddjkdzj = 0.0;

      p3p4Aij = -surften * (P3[i] + P4[i] * Aij);

      L80:k = ineighbor[count2] - 1;
				/*same as above, -1 ! */
      if (j == k)
	goto L85;

      xk = x[3 * k];
      yk = x[3 * k + 1];
      zk = x[3 * k + 2];

      rjk2 =
	(xj - xk) * (xj - xk) + (yj - yk) * (yj - yk) + (zj -
							 zk) *
	(zj - zk);

      djk1i = 1.0 / sqrt(rjk2);

      rjk = rjk2 * djk1i;

      if (P0[j] + P0[k] > rjk) {
	vdw2dif = P0[j] * P0[j] - P0[k] * P0[k];
	tmpajk = 2.0 * P0[j] - rjk - vdw2dif * djk1i;

	Ajk = PIx1 * P0[j] * tmpajk;

	sumAjk = sumAjk + Ajk;
	sumAjk2 = sumAjk2 + Ajk;

	dAjkddjk =
	  PIx1 * P0[j] * djk1i * (djk1i * djk1i * vdw2dif -
				  1.0);

	dAjkddjkdxj = dAjkddjk * (xj - xk);
	dAjkddjkdyj = dAjkddjk * (yj - yk);
	dAjkddjkdzj = dAjkddjk * (zj - zk);

	f[3 * k] = f[3 * k] + dAjkddjkdxj * p3p4Aij;
	f[3 * k + 1] = f[3 * k + 1] + dAjkddjkdyj * p3p4Aij;
	f[3 * k + 2] = f[3 * k + 2] + dAjkddjkdzj * p3p4Aij;

	sumdAjkddjkdxj = sumdAjkddjkdxj + dAjkddjkdxj;
	sumdAjkddjkdyj = sumdAjkddjkdyj + dAjkddjkdyj;
	sumdAjkddjkdzj = sumdAjkddjkdzj + dAjkddjkdzj;

      }

      L85:count2 = count2 + 1;
      if (ineighbor[count2] != 0) {
	goto L80;
      } else {
	count2 = icount;
      }


      sumAijAjk = sumAijAjk + Aij * sumAjk2;

      sumdAijddijdxi = sumdAijddijdxi - dAijddijdxj;
      sumdAijddijdyi = sumdAijddijdyi - dAijddijdyj;
      sumdAijddijdzi = sumdAijddijdzi - dAijddijdzj;
      sumdAijddijdxiAjk =
	sumdAijddijdxiAjk - dAijddijdxj * sumAjk2;
      sumdAijddijdyiAjk =
	sumdAijddijdyiAjk - dAijddijdyj * sumAjk2;
      sumdAijddijdziAjk =
	sumdAijddijdziAjk - dAijddijdzj * sumAjk2;

      lastxj = dAijddijdxj * sumAjk2 + Aij * sumdAjkddjkdxj;
      lastyj = dAijddijdyj * sumAjk2 + Aij * sumdAjkddjkdyj;
      lastzj = dAijddijdzj * sumAjk2 + Aij * sumdAjkddjkdzj;

      dAidxj =
	surften * (P2[i] * dAijddijdxj +
		   P3[i] * sumdAjkddjkdxj + P4[i] * lastxj);
      dAidyj =
	surften * (P2[i] * dAijddijdyj +
		   P3[i] * sumdAjkddjkdyj + P4[i] * lastyj);
      dAidzj =
	surften * (P2[i] * dAijddijdzj +
		   P3[i] * sumdAjkddjkdzj + P4[i] * lastzj);

      f[3 * j] = f[3 * j] + dAidxj;
      f[3 * j + 1] = f[3 * j + 1] + dAidyj;
      f[3 * j + 2] = f[3 * j + 2] + dAidzj;

      count = count + 1;
      if (ineighbor[count] != 0) {
	goto L70;
      } else {
	count = count + 1;
      }

      Ai = P1[i] * si + P2[i] * sumAij + P3[i] * sumAjk +
	P4[i] * sumAijAjk;

      dAidxi =
	surften * (P2[i] * sumdAijddijdxi +
		   P4[i] * sumdAijddijdxiAjk);
      dAidyi =
	surften * (P2[i] * sumdAijddijdyi +
		   P4[i] * sumdAijddijdyiAjk);
      dAidzi =
	surften * (P2[i] * sumdAijddijdzi +
		   P4[i] * sumdAijddijdziAjk);

      f[3 * i] = f[3 * i] + dAidxi;
      f[3 * i + 1] = f[3 * i + 1] + dAidyi;
      f[3 * i + 2] = f[3 * i + 2] + dAidzi;

      totsasa = totsasa + Ai;
				/*  printf("totsasa up to %d %f\n",i,totsasa); */
      }
    }
    /*      printf("SASA %f , ESURF %f \n",totsasa,totsasa*surften); */
    esurf = totsasa * surften;
  }

  /*
   * Compute the generalized Born energy for all pair interactions, as
   * well as the Van der Waals and Coulombic energy for all pairs that
   * are not on the excluded list.  Compute the 1st and 2nd derivatives
   * of the of the Van der Waals and Coulombic energy for all non-excluded
   * pairs.
   *
   * In addition, compute the 1st derivative of the Born energy for all
   * pairs of atoms (selected by the indices i and j).  David Case and
   * his colleagues have shown that the first derivative for an atom pair
   * may be calculated by first building the following equations in
   * terms of Ri (the effective radius of atom i), Rj (the effective
   * radius of atom j) and Dij (the distance between atoms i and j):
   *
   *      rb2 = Ri * Rj
   *
   *      efac = exp(Dij^2 / (4 * rb2))
   *
   *      fgbi = 1 / sqrt(Dij^2 + rb2 * efac)
   *
   *      fgbk = -kappa * KSCALE / fgbi
   *
   *      expmkf = exp(fgbk) / DIELECTRIC_CONSTANT
   *
   *      dielfac = 1 - expmkf
   *
   *      temp4 = fgbi * fgbi * fgbi
   *
   *      temp6 = Qi * Qj * temp4 * (dielfac + fgbk * expmkf)
   *
   *      temp5 = efac * temp6 * (rb2 + 0.25 * r2) / 2
   *
   * From the above equations it is possible to express the first
   * derivative of the Born energy with respect to any cartesian coordinate
   * (designated by Xl) by summing contributions from the three following
   * equations:
   *
   * 1.   d(Dij)/d(Xl) * A(Dij) where
   *
   *      A(Dij) = (1 - efac / 4) * temp6
   *
   * 2.   d(Ri)/d(Xl) * B(Dij) where
   *
   *      B(Dij) = Ri * temp5
   *
   * 3.   d(Rj)/d(Xl) * C(Dij) where
   *
   *      C(Dij) = Rj * temp5
   *
   * Equation 1 is the simplest of the three to evaluate.  One calculates
   * all of the derivatives of Dij with respect to cartesian coordinates,
   * multiplies these derivatives by A(Dij), and sums the resulting products
   * to the gradient vector.  There are six derivatives involving atoms
   * i and j:
   *
   *      d(Dij)/d(Xi), d(Dij)/d(Yi), d(Dij)/d(Zi)
   *
   *      d(Dij)/d(Xj), d(Dij)/d(Yj), d(Dij)/d(Zj)
   *
   * The derivatives of Dij with respect to the cartesian coordinates
   * contain a factor of Dij in the denominator.  This factor is absorbed
   * by equation 1 (and all of the other equations that incorporate the
   * derivatives) rather than carried with the individual derivatives.
   *
   * Equations 2 and 3 are more complicated because the derivatives of the
   * effective radii Ri and Rj with respect to the cartesian coordinates are
   * sometimes series, but sometimes single terms.  For example:
   *
   *      d(Ri)/d(Xi) = d[f(Dij)]/d(Dij) * d(Dij)/d(Xi)
   *
   *                  + d[f(Dik)]/d(Dik) * d(Dik)/d(Xi) + ...
   *
   *                  + d[f(Din)]/d(Din) * d(Din)/d(Xi)
   *
   *      d(Ri)/d(Xj) = d[f(Dij)]/d(Dij) * d(Dij)/d(Xj)
   *
   *      d(Rj)/d(Xj) = d[f(Dij)]/d(Dij) * d(Dij)/d(Xj)
   *
   *                  + d[f(Djk)]/d(Djk) * d(Djk)/d(Xj) + ...
   *
   *                  + d[f(Djn)]/d(Djn) * d(Djn)/d(Xj)
   *
   *      d(Rj)/d(Xi) = d[f(Dij)]/d(Dij) * d(Dij)/d(Xi)
   *
   * where n is the number of atoms and f() designates a function of the
   * interatomic distance.
   *
   * Moreover, when all of the atom pairs are considered, one can refactor
   * the computation to obtain terms such as:
   *
   *      d(Ri)/d(Xi) * [B(Dij) + B(Dik) + ... + B(Din)]
   *
   *      d(Rj)/d(Xj) * [C(Dij) + C(Djk) + ... + C(Djn)]
   *
   * The square brackets [] that B(Dij) + B(Dik) + ... + B(Din) are summed
   * prior to multiplication by the derivatives of the effective radius.
   *
   * David Case and his colleagues have discovered an algorithm that permits
   * evaluation of the first derivatives while requiring that the first
   * derivatives of the effective radii be evaluated once only.  They
   * use an array called "sumdeijda" into which the above series are summed.
   * For example, the series B(Dij) + B(Dik) + ... + B(Din) is summed into
   * sumdeijda[i], and the series C(Dij) + C(Djk) + ... + C(Din) is summed
   * into sumdeijda[j].  One can think of the contents of sumdeijda[i]
   * as the sum of everything that gets multiplied by d(Ri)/d(Xl), where
   * d(Ri)/d(Xl) represents all of the derivatives of Ri with respect to the
   * cartesian coordinates.  Similarly, one can think of the contents of
   * sumdeijda[j] as the sum of everything that gets multiplied by d(Rj)/d(Xl),
   * where d(Rj)/d(Xl) represents all of the derivatives of Rj with respect
   * to the cartesian coordinates.
   *
   * Once the sumdeijda array is loaded, the d(Ri)/d(Xl) and d(Rj)/d(Xl) are
   * calculated and multiplied by the respective elements of sumdeijda.  For
   * example, for d(Ri)/d(Xl) the individual terms of the d(Ri)/d(Xi) series
   * are calculated and stored into another array called da.  Each term has
   * an x-, y- and z-component.  The components of the jth term are:
   *
   *      d[f(Dij)]/d(Dij) * d(Dij)/dXi
   *
   *      d[f(Dij)]/d(Dij) * d(Dij)/dYi
   *
   *      d[f(Dij)]/d(Dij) * d(Dij)/dZi 
   *
   *      d[f(Dij)]/d(Dij) * d(Dij)/dXj
   *
   *      d[f(Dij)]/d(Dij) * d(Dij)/dYj
   *
   *      d[f(Dij)]/d(Dij) * d(Dij)/dZj
   *
   * Note that the jth term contains derivatives with respect to the cartesian
   * coordinates of atoms i and j.  The derivatives with respect to the cartesian
   * coordinates of atom j are stored in da[j].  The derivatives with respect to
   * the cartesian coordinates of atom i are summed into da[i] with the result
   * that da[i] contains derivatives from f(Dij), f(Dik), ..., f(Din).
   *
   * Each element of the da array is then multiplied by sumdeijda[i] to form:
   *
   * [ d(Rij)/d(Xl) + d(Rik)/d(Xl) + ... ] * [f(Dij) + f(Dik) + ... ]

   * where the series d(Rij)/d(Xl) + d(Rik)/d(Xl) + ... contains n terms each
   * of which has an (x,y,z) coordinate.  Therefore this result is a vector of
   * length 3*n where n is the number of atoms.  This vector represents
   * the derivatives of Ri with respect to all of the cartesian coordinates.
   * It is summed into the gradient vector.
   *
   * The process is then repeated for Rj, Rk, ..., Rn.
   *
   * There is one further refinement that relates to the first derivatives
   * of the "diagonal" terms, i.e., those energy terms that are functions
   * of the effective radius Ri but not of the interatomic distance Dij.
   * One begins by building up the following equations:
   *
   *      expmkf = exp(-kappa * KSCALE * Ri) / DIELECTRIC_CONSTANT
   *
   *      dielfac = 1 - expmkf
   *
   *      qi2h = Qi^2 / 2
   *
   *      qid2h = di2h * DIELECTRIC_CONSTANT
   *
   * Then the first derivative with respect to Ri is the equation:
   *
   * 4.   d(Ri)/d(Xl) * R(Dij) where
   *
   *      R(Dij) = qid2h - kappa * KSCALE * qi2h * expmkf * Ri
   *
   *
   * Equation 4, like equation 2, is multiplied by d(Ri)/d(Xl) and
   * therefore one can add these two equations and store the result
   * in the ith element of the sumdeijda array.
   *
   * Note that equations 2, 3 and 4 include a factor of either -Ri*Ri
   * or -Rj*Rj which arises from the derivative of Ri or Rj, respectively.
   * The negation is introduced when  sumda is multiplied by an element
   * from the dreff array.
   *
   * The above algorithm may be extended to the calculation of the second
   * derivatives.  One proceeds by first building the following equations
   * in terms of Ri, Rj and Dij:
   *
   *      RiInv = 1 / Ri
   *
   *      RjInv = 1 / Rj
   *
   *      RiRjInv = RiInv * RjInv
   *
   *      temp4 = fgbi * fgbi * fgbi
   *
   *      temp5 = Qi * Qj * temp4 * (dielfac + fgbk * expmkf) / 2
   *
   *      temp6 = fgbi * fgbi * temp5
   *
   *      temp7 = Qi * Qj * kappa^2 * KSCALE^2 * temp4 * expmkf
   *
   *      temp8 = 1 - efac / 4
   *
   *      temp9 = temp7 / 4 - 3 * temp6 / 2
   *
   *      temp10 = Dij^2 / 4
   *
   *      temp11 = Ri * Rj + temp10
   *
   *      temp12 = Ri^2 * Rj^2
   *
   *      temp13 = Ri * temp11
   *
   *      temp14 = Rj * temp11
   *
   *      temp15 = temp10 * RjInv
   *
   *      temp16 = temp10 * RiInv
   *
   *      temp17 = temp9 * temp13 * efac
   *
   *      temp18 = temp9 * temp14 * efac
   *
   *      temp19 = temp5 * efac
   *
   *      temp20 = temp8 * temp9
   *
   *      temp21 = temp20 * temp13
   *
   *      temp22 = temp20 * temp14
   *
   *      temp23 = 2 * efac;
   *
   *
   * From the above equations it is possible to express the second
   * derivative of the Born energy with respect to two cartesian coordinates
   * (designated by Xl and Xm) by summing contributions from the following
   * equations:
   *
   * 5.   d(Ri)/d(Xm) * [ d(Dij)/d(Xl) * D(Dij) ] where
   *
   *      D(Dij) = temp23 * (temp21 - temp15 * temp5 / 4)
   *
   *
   * 6.   d(Rj)/d(Xm) * [ d(Dij)/d(Xl) * E(Dij) ] where
   *
   *      E(Dij) = temp23 * (temp22 - temp16 * temp5 / 4)
   *
   *
   * 7.   d(Ri)/d(Xl) * [ d(Dij)/d(Xm) * J(Dij) ] where
   *
   *      J(Dij) = temp23 * ((Ri - RjInv * temp11) * temp5 / 4 + temp21)
   *
   *
   * 8.   d(Rj)/d(Xl) * [ d(Dij)/d(Xm) * N(Dij) ] where
   *
   *      N(Dij) = temp23 * ((Rj - RiInv * temp11) * temp5 / 4 + temp22)
   *
   *
   * 9.   d(Ri)/d(Xm) * [ d(Ri)/d(Xl) * H(Dij) ] where
   *
   *      H(Dij) = efac * (temp10 * (temp13 * RjInv - Ri^2) * temp5 + temp13 * temp17)
   *
   *
   * 10.  d(Rj)/d(Xm) * [ d(Rj)/d(Xl) * M(Dij) ] where
   *
   *      M(Dij) = efac * (temp10 * (temp14 * RiInv - Rj^2) * temp5 + temp14 * temp18)
   *
   *
   * 11.  d(Rj)/d(Xm) * [ d(Ri)/d(Xl) * I(Dij) ] where
   *
   *      I(Dij) = efac * ((temp12 + temp16 * temp13) * temp5 + temp13 * temp18)
   *
   *
   * 12.  d(Ri)/d(Xm) * [ d(Rj)/d(Xl) * L(Dij) ] where
   *
   *      L(Dij) = efac * ((temp12 + temp15 * temp14) * temp5 + temp14 * temp17)
   *
   *
   * 13.  d2(Ri)/d(Xm)d(Xl) * K(Dij) where
   *
   *      K(Dij) = temp13 * temp19
   *
   *
   * 14.  d2(Rj)/d(Xm)d(Xl) * O(Dij) where
   *
   *      O(Dij) = temp14 * temp19
   *
   *
   * 15.  d2(Dij)/d(Xm)d(Xl) * F(Dij) where
   *
   *      F(Dij) = 2 * temp8 * temp5
   *
   *
   * 16.  d(Dij)/d(Xm) * [ d(Dij)/d(Xl) * G(Dij) ] where
   *
   *      G(Dij) = 4 * temp8 * temp20
   *
   *               + (2 * temp8 + efac * temp10 * RiRjInv) * temp5 / Dij^2
   *
   *
   * Using the above equations 5-16 it is possible to calculate the
   * second derivatives of the Born energy in an optimized manner.
   * Note, however, that computing the second derivatives of the effective
   * radius with respect to the cartesian coordinates involves some
   * complexity.  This fact can be appreciated by beginning with the
   * following expression for the effective radius Ri of atom i:
   *
   *
   * Ri = 1 / [1/Rborn(i) + f(Dij) + f(Dik) + ... + f(Din)]
   *
   *
   * where Rborn(i) represents the Born radius of atom i and f() respesents
   * a function of the interatomic distance.  Differentiating with respect to
   * a cartesian coordinate Xl gives:
   *
   *
   * d(Ri)/d(Xl) = -Ri^2 * { d[f(Dij)]/d(Dij) * d(Dij)/d(Xl) +
   * 
   *                         d[f(Dik)]/d(Dik) * d(Dik)/d(Xl) + ... +
   *
   *                         d[f(Din)]/d(Din) * d(Din)/d(Xl) }
   *
   *
   * The above expression is the general case.  Many of the terms are
   * zero, for example, d(Dij)/d(Xl) is zero when l equals neither i nor j.
   *
   * Differentiating again with respect to a cartesian coordinate Xm gives:
   *
   *
   * d2(Ri)/d(Xm)d(Xl) =
   *
   * - Ri^2 * { d[f(Dij)]/d(Dij) * d2(Dij)/d(Xm)d(Xl) +
   *
   *            d[f(Dik)]/d(Dik) * d2(Dik)/d(Xm)d(Xl) + ... +
   *
   *            d[f(Din)]/d(Din) * d2(Din)/d(Xm)d(Xl) }
   *
   *
   * - Ri^2 * { d2[f(Dij)]/d(Dij)2 * d(Dij)/d(Xm) * d(Dij)/d(Xl) +
   *
   *            d2[f(Dik)]/d(Dik)2 * d(Dik)/d(Xm) * d(Dik)/d(Xl) + ... +
   *
   *            d2[f(Din)]/d(Din)2 * d(Din)/d(Xm) * d(Din)/d(Xl) }
   *
   *
   * + 2Ri^3 * { d[f(Dij)]/d(Dij) * d(Dij)/d(Xm) * d[f(Dij)]/d(Dij) * d(Dij)/d(Xl) +
   *
   *             d[f(Dij)]/d(Dij) * d(Dij)/d(Xm) * d[f(Dik)]/d(Dik) * d(Dik)/d(Xl) + ... +
   *
   *             d[f(Dij)]/d(Dij) * d(Dij)/d(Xm) * d[f(Din)]/d(Din) * d(Din)/d(Xl) +
   *
   * 
   *             d[f(Dik)]/d(Dik) * d(Dik)/d(Xm) * d[f(Dij)]/d(Dij) * d(Dij)/d(Xl) +
   *
   *             d[f(Dik)]/d(Dik) * d(Dik)/d(Xm) * d[f(Dik)]/d(Dik) * d(Dik)/d(Xl) + ... +
   *
   *             d[f(Dik)]/d(Dik) * d(Dik)/d(Xm) * d[f(Din)]/d(Din) * d(Din)/d(Xl) + ... +
   *
   *
   *             d[f(Din)]/d(Din) * d(Din)/d(Xm) * d[f(Dij)]/d(Dij) * d(Dij)/d(Xl) +
   *
   *             d[f(Din)]/d(Din) * d(Din)/d(Xm) * d[f(Dik)]/d(Dik) * d(Dik)/d(Xl) + ... +
   *
   *             d[f(Din)]/d(Din) * d(Din)/d(Xm) * d[f(Din)]/d(Din) * d(Din)/d(Xl) }
   *
   *
   * Many of the terms in the above expression are zero, or example, when neither
   * l nor m equals i or j.  However, many non-zero terms remain, and in particular
   * there exists a large number of the terms that are multiplied by 2Ri^3.  But
   * fortunately these particular terms may be computed by simply multiplying
   * the derivative of the Ri with respect to all cartesian coordinates by itself.
   *
   * As can be appreciated from the above analysis, differentiation gives rise to
   * factors of -Ri^2, Ri^3, -Rj^2 and Rj^3.  A factor of Ri^2 or Rj^2 is included
   * equations 5-16, as well as equations 17-18 (below).
   *
   * Square brackets [ ] have been added to many of equations 5-18
   * to indicate the order of evaluation.  The derivative inside of the
   * brackets will be called the inner derivative.  The derivative outside
   * of the brackets will be called the outer derivative.  The inner
   * derivative will be evaluated prior to the outer derivative.  The
   * inner derivative that is multiplied by the accompanying functions
   * of Dij, e.g., H(Dij) of equation 7, produces a gradient vector.
   *
   * Equations 15 and 16 are analogous to equation 1 and are simple to
   * evaluate because they do not involve derivatives of the effective
   * radii Ri and Rj and so do not comprise series.  Instead, they comprise
   * the derivatives of Dij with respect to the cartesian coordinates
   * of atoms i and j only.
   *
   * Equations 13 and 14 are analogous to equations 2 and 3 except that
   * they involve second derivatives of the effective radii Ri and Rj instead
   * of first derivatives.  However, the second derivatives of the effective
   * radius Ri include terms such as d2(Ri)/d(Xl)d(Xm) which affect many Hessian
   * elements when neither l nor m equals i.
   *
   * Equations 9, 10, 11 and 12 imply two applications of the approach used
   * to evaluate equations 2 and 3.  However, the result of the multiplication
   * by the inner derivative (which is enclosed in [ ]) is not a single term
   * but rather is a gradient vector.  Therefore, multiplication by the outer
   * derivative involves multiplying each derivative of Ri or Rj by all of the
   * terms in the gradient vector.  Moreover, this process requires a
   * two-dimensional array, i.e., an array of gradient vectors which is named
   * either "sumdeijdg" or "sumdeijdh", and is analogous to the sumdeijda array
   * that is used to evaluate equations 2, 3 and 4.  The order of evaluation of
   * the derivatives appears to be unimportant for equations 9 and 10, i.e., the
   * inner derivatives may be swappped  with the outer derivatives without
   * affecting the results.
   *
   * However, there is a subtlety involving equations 11 and 12 which requires
   * that if the inner derivatives are swapped with the outer derivatives in
   * equation 11, they must be similarly swapped in equation 12.  This subtlety
   * arises because the the sumdeijdg array is used to collect all inner
   * derivatives with respect to Xl that will subsequently be multiplied by
   * outer derivatives with respect to Xm  An analogous sumdeijdh array is used
   * to collect all inner derivatives with respect to Xm that will subsequently
   * be multiplied by outer derivatives with respect to Xl.  The inner derivatives
   * of equations 11 and 12 are stored in the sumdeijdg array, which implies
   * that for these equations the inner derivatives are taken with respect to Xl.
   * Swapping the inner and outer derivatives for both of these equations preserves
   * the symmetry of the Hessian matrix, but swapping the inner and outer derivatives
   * for only one of these equations breaks the symmetry and leads to erroneous
   * results.  For example, if one were to swap the inner and outer derivatives
   * in equation 11, but not in equation 12, the following pair of equations would
   * obtain:
   *
   * 11'. d(Ri)/d(Xm) * [ d(Rj)/d(Xl) * I(Dij) ]
   *
   * 12.  d(Ri)/d(Xm) * [ d(Rj)/d(Xl) * L(Dij) ]
   *
   * This pair of equations is not symmetric and is therefore incorrect.
   *
   * Equations 5, 6, 7 and 8 are evaluated by first multiplying by
   * an inner derivative of Dij and then by multiplying by an outer
   * derivative of Ri or Rj.  The inner derivatives of Dij are
   * collected into either the sumdeijdg array or sumdeijdh array.
   * The derivatives of a particular Dij produce a gradient vector
   * that is stored in one row of either sumdeijdg or sumdeijdh.
   * Subsequent multiplication by the outer derivative of Ri or Rj
   * involves multiplying each derivative of Ri or Rj by all of the
   * terms in the gradient vector.  This process involves
   * a subtlety.  In equations 5 and 6 the inner derivative is with
   * respect to Xl whereas in equations 7 and 8 the inner derivative
   * is with respect to Xm.  For this reason, the inner derivatives
   * of equations 5 and 6 are collected into the sumdeijdg array,
   * whereas the inner derivatives of equations 7 and 8 are collected
   * into the sumdeijdh array, following the convention that was
   * discussed above for equations 11 and 12.  The outer derivatives
   * taken with respect to Xm are computed using the sumdeijdg array,
   * whereas the outer derivatives taken with respect to Xl are computed
   * using the sumdeijdh array.  These two sets of derivatives are
   * written into transpose-related elements of the Hessian matrix in
   * order to preserve the symmetry of the Hessian.
   *
   * On another topic, it is possible to generate equations for the second
   * derivatives of the non-pairwise ("diagonal") terms.  One begins by
   * building up the following equations:
   *
   *      expmkf = exp(-kappa * KSCALE * Ri) / DIELECTRIC_CONSTANT
   *
   *      dielfac = 1 - expmkf
   *
   *      temp17 = dielfac - kappa * KSCALE * expmkf * Ri
   *
   *      temp18 = Ri * (kappa * KSCALE * expmkf * Ri * (2 + kappa * KSCALE * Ri)
   *
   *                     - 2 * dielfac)
   *
   *
   * Using the above equations it is possible to calculate the second
   * derivatives of the non-pairwise terms via the two following equations:
   *
   * 17.  d2(Ri)/d(Xm)d(Xl) * Q(Dij) where
   *
   *      Q(Dij) = Qi^2 * temp17 / 2
   *
   *
   * 18.  d(Ri)/d(Xm) * [ d(Ri)/d(Xl) * P(Dij) ] where
   *
   *      P(Dij) = Qi^2 * temp18 / 2
   *
   * Equations 17 and 18 are handled in a manner analogous to equation 4.
   *
   * Initialize the arrays used to control the derivative calculation.
   */

  for (i = 0; i < n; i++) {
    sumdeijda[i] = 0.0;
    sumdeijd2[i] = 0.0;
    sumepol[i] = 0.0;
    sumelec[i] = 0.0;
    sumevdw[i] = 0.0;
    }

#ifdef MULTIPLE_THREADS
#pragma omp parallel private (i)
#endif

  {

#ifdef MULTIPLE_THREADS
#pragma omp for
#endif

    for (i = 0; i< nn; i++) {
      sumdeijdn[i] = 0.0;
    }

#ifdef MULTIPLE_THREADS
#pragma omp for
#endif

    for (i = 0; i < nn3; i++) {
      sumdeijdg[i] = 0.0;
      sumdeijdh[i] = 0.0;
    }
  }

#ifndef MULTIPLE_THREADS
  iexcl = 0;
#endif

  /*
   * Calculate the Born, van der Waals and Coulombic energies.  Complete
   * equations 15 and 16, and set up the remaining equations.  When in
   * multi-threaded mode in this nest of loops sum data relative only to
   * atom i or to pairs ii, ij and ji so that OpenMP can be used to parallelize
   * the outer loop.  When in single-threaded mode sum data relative to
   * atoms i and j, as well as to pairs ii, ij, ji and jj, and also perform
   * excluded list management.
   */

#ifdef MULTIPLE_THREADS
#pragma omp parallel \
  private (i, i3, in, in3, xi, yi, zi, qi, qj, ri, ri2, riinv, expmkf, dielfac, \
	   qi2h, qid2h, temp17, temp18, p_dij, q_dij, r_dij, iaci, j, j3, jn, jn3, \
	   xij, yij, zij, r2, r2inv, qiqj, rj, rj2, rjinv, rirjinv, rb2, efac, \
	   fgbi, fgbk, temp4, temp6, a_dij, temp5, b_dij, c_dij, di, dj, d2ii, d2ij, \
	   k, l, df, dg, rinv, r4inv, eel, ic, r6inv, f6, f12, dedx, dedy, dedz, \
	   temp7, temp8, temp9, temp10, temp11, temp12, temp13, temp14, temp15, \
	   temp16, temp19, temp20, temp21, temp22, temp23, d_dij, e_dij, f_dij, \
	   g_dij, h_dij, i_dij, j_dij, k_dij, l_dij, m_dij, n_dij, o_dij, \
	   epol, elec, evdw, jexcl, jexcl_last, nexcl, npairs, xj, yj, zj, thi, \
	   iexcl, iexcl_last, d2jj, ri1i, temp1, dij3i, dx, dy, dz, m, iexw, \
	   datmp, tmpsd, dumbo, temp2, da2tmp, sj, sj2, dij1i, dij2i, dij, \
	   dxj, dyj, dzj, sreffi, dreffi, sdreff, sumdh, sumdn, sumda, sumd2)
#endif
  {

    /*
     * Allocate and initialize the iexw array used for skipping excluded
     * atoms.  Note that because of the manner in which iexw is used, it
     * is necessary to initialize it before only the first iteration of
     * the following loop.
     */

    iexw = ivector(-1, n);
    for (i = -1; i < n; i++) {
      iexw[i] = -1;
    }

#ifdef MULTIPLE_THREADS
#pragma omp for schedule (dynamic, 8)
#endif

    /* Loop over all atoms i. */

    for (i = 0; i < n; i++) {

      in = n * i;

      qi = q[i];
      ri = reff[i];

      /*
       * If atom i is not frozen, compute the "diagonal" energy that
       * is a function of only the effective radius Ri but not of the
       * interatomic distance Dij.  Compute also the contribution of
       * the diagonal energy term to the sum by which the derivatives
       * of Ri will be multiplied.
       */

      if (!frozen[i]) {
	expmkf = exp(-KSCALE * (*kappa) * ri) / (*diel_ext);
	dielfac = 1.0 - expmkf;
	qi2h = 0.5 * qi * qi;
	qid2h = qi2h * dielfac;
	sumepol[i] += -qid2h / ri;
	temp17 = dielfac - (*kappa) * KSCALE * expmkf * ri;
	temp18 = ( (*kappa) * KSCALE * expmkf * ri * (2.0 + (*kappa) * KSCALE * ri)
		   - 2.0 * dielfac ) * ri;
	p_dij = qi2h * temp18;
	q_dij = qi2h * temp17;
	r_dij = qid2h - KSCALE * (*kappa) * qi2h * expmkf * ri;
	sumdeijdn[in + i] += p_dij;
	sumdeijd2[i] += q_dij;
	sumdeijda[i] += r_dij;
      }

      /*
       * Skip the pair calculations if there are no atoms j on the
       * pair list of atom i.
       */

      nexcl = prm->Iblo[i];
      npairs = upairs[i];
      if ( npairs > 0 ) {

	i3 = 3 * i;
	in3 = 3 * in;

	xi = x[i3 + 0];
	yi = x[i3 + 1];
	zi = x[i3 + 2];

	iaci = prm->Ntypes * (prm->Iac[i] - 1);

	ri2 = ri * ri;
	riinv = 1.0 / ri;

	/*
	 * Expand the excluded atom list into the iexw array by storing i
	 * at array address j.  The starting address is either iexcl that
	 * gets updated with each iteration of the i-loop, or Istr[i] that
	 * has been created in the mme_init function to permit parallelization.
	 */

#ifndef MULTIPLE_THREADS
	jexcl = iexcl;
#else
	jexcl = Istr[i];
#endif

	jexcl_last = jexcl + nexcl;
	for (j = jexcl; j < jexcl_last; j++) {
	  iexw[prm->ExclAt[j] - 1] = i;
	}

	epol = elec = evdw = 0.0;

	/* Select atoms j from the pair list. */

	for (m = uprstr[i]; m < uprstr[i] + npairs; m++) {
	
	  j = pairlist[m];

	  j3 = 3 * j;
	  jn = n * j;
	  jn3 = 3 * jn;

	  /* Continue computing the non-diagonal energy term. */

	  xij = xi - x[j3 + 0];
	  yij = yi - x[j3 + 1];
	  zij = zi - x[j3 + 2];
	  r2 = xij * xij + yij * yij + zij * zij;

	  r2inv = 1.0 / r2;
	  qiqj = qi * q[j];
	  rj = reff[j];
	  rj2 = rj * rj;
	  rjinv = 1.0 / rj;
	  rirjinv = riinv * rjinv;
	  rb2 = ri * rj;
	  efac = exp(-r2 / (4.0 * rb2));
	  fgbi = 1.0 / sqrt(r2 + rb2 * efac);
	  fgbk = -(*kappa) * KSCALE / fgbi;

	  /*
	   * Calculate the "non-diagonal" energy term, i.e., the term that is a
	   * function of the effective radius Ri and the interatomic distance Dij.
	   */ 

	  expmkf = exp(fgbk) / (*diel_ext);
	  dielfac = 1.0 - expmkf;
	  epol += -qiqj * dielfac * fgbi;

	  temp4 = fgbi * fgbi * fgbi;
	  temp6 = qiqj * temp4 * (dielfac + fgbk * expmkf);

	  /* The following term is required by equation 1. */

	  a_dij = temp6 * (1.0 - 0.25 * efac);

	  /*
	   * Prepare to compute the 1st derivatives of the effective radii.
	   * The following code performs the sums indicated by equations 2 and 3.
	   * Compute the contribution of the non-diagonal energy term to the sum
	   * by which the derivatives of Ri and Rj will be multiplied.  When in
	   * single-threaded mode sum the contribution for the derivatives of Ri
	   * and Rj.  When in multi-threaded mode sum the contribution for the
	   * derivative of Ri.
	   */

	  temp5 = 0.5 * efac * temp6 * (rb2 + 0.25 * r2);
	  b_dij = ri * temp5;
	  c_dij = rj * temp5;

	  sumdeijda[i] += b_dij;
#ifndef MULTIPLE_THREADS
	  sumdeijda[j] += c_dij;
#endif

	  /*
	   * Calculate the first and second derivatives of the interatomic
	   * distance Dij with respect to the cartesian coordinates of atoms
	   * i and j.  The results are placed into five arrays:
	   *
	   *   di[] for the first derivatives with respect to atom i
	   *   dj[] for the first derivatives with respect to atom j
	   *   d2ii[] for the second derivatives with respect to atom i
	   *   d2jj[] for the second derivatives with respect to atom j
	   *   d2ij[] for the second derivatives with respect to atoms i and j
	   *
	   * Some useful symmetry obtains.  The d2ii and d2ij arrays
	   * are symmetric in that their lower and upper triangles are
	   * the transposes of one another.  Also, d2jj equals d2ii, and
	   * d2ij is the negative of d2ii.
	   *
	   * Note: the additional factor of the interatomic distance Dij that
	   * ought to be included in the denominator of the following equations
	   * is provided by the absence of Dij in equations 15 and 16, as well as
	   * by one less power of Dij in the denominator of the equations for the
	   * derivatives of the Van der Waals and Coulombic energy terms.
	   */
      
	  di[0] = xij;
	  di[1] = yij;
	  di[2] = zij;

	  dj[0] = -xij;
	  dj[1] = -yij;
	  dj[2] = -zij;

	  /* Load the upper triangle of d2ii. */

	  d2ii[0][0] = 1.0 - xij * xij * r2inv;
	  d2ii[0][1] = -xij * yij * r2inv;
	  d2ii[0][2] = -xij * zij * r2inv;

	  d2ii[1][1] = 1.0 - yij * yij * r2inv;
	  d2ii[1][2] = -yij * zij * r2inv;

	  d2ii[2][2] = 1.0 - zij * zij * r2inv;

	  /* Finish loading the rest of all of the matrices. */

	  for (k = 0; k < 3; k++) {

	    /* Load the upper triangle of d2ij. */

	    for (l = k; l < 3; l++) {
#ifndef MULTIPLE_THREADS
	      d2jj[k][l] = d2ii[k][l];
#endif
	      d2ij[k][l] = -d2ii[k][l];
	    }

	    /* Load the symmetric elements of d2ij. */

	    for (l = k+1; l < 3; l++) {
	      d2ij[l][k] = d2ij[k][l];
	    }
	  }

	  /*
	   * Compute the Van der Waals and Coulomb energy here because
	   * if egb2 is called nbond2 is not called.  Remember that the
	   * derivatives of Dij with respect to x, y and z supply one more
	   * power of 1/Dij but that power of 1/Dij is included in the
	   * equations below, instead of with the derivatives of Dij.
	   *
	   * Compute the Van der Waals and Coulombic energies for only
	   * those pairs which are not on the excluded atom list.  Any
	   * pair on the excluded atom list will have atom i stored at
	   * address j of the iexw array.  It is not necessary to reset
	   * the elements of the iexw array to -1 between successive
	   * iterations in i because an i,j pair is uniquely identified
	   * by atom i stored at array address j.  Thus for example, the
	   * i+1,j pair would be stored at the same address as the i,j
	   * pair but after the i,j pair were used.
	   *
	   * Initialize df here instead of within the following if statement
	   * so that df is initialized unconditionally.
	   */

	  df = 0.0;

	  if (iexw[j] != i) {

	    /* Initialize dg. */

	    dg = 0.0;

	    rinv = 1. / sqrt(r2);
	    r4inv = r2inv * r2inv;

	    /*  gas-phase Coulomb energy:  */

	    eel = qiqj * rinv;
	    elec += eel;

	    df -= eel;
	    dg += 2. * eel;

	    /* Lennard-Jones energy:   */

	    ic = prm->Cno[iaci + prm->Iac[j] - 1] - 1;
	    if( ic >= 0 ){
	      r6inv = r2inv * r2inv * r2inv;
	      f6 = prm->Cn2[ic] * r6inv;
	      f12 = prm->Cn1[ic] * r6inv * r6inv;
	      evdw += f12 - f6;

	      df += 6. * f6 - 12. * f12;
	      dg += 156. * f12 - 42. * f6;
	    }

	    df *= r2inv;
	    dg *= r4inv;

	    /*
	     * Calculate the second derivatives of Dij for the Van der Waals
	     * and Coulombic interactions.  This calculation uses df and dg
	     * for these interactions only.
	     *
	     * When in single-threaded mode sum to the Hessian the ii, ij and
	     * jj terms.  When in multi-threaded mode sum to the ii and ij
	     * terms only.
	     *
	     * The ii and jj terms lie on the diagonal so only the upper triangle 
	     * needs to be updated for these 3x3 matrices.  However, the ij term
	     * lies off the diagonal so the full 3x3 matrix must be updated.
	     */

	    for (k = 0; k < 3; k++) {
	      for (l = k; l < 3; l++) {
		g[n3*(i3 + k) + i3 + l] += df*d2ii[k][l] + dg*di[k]*di[l];
#ifndef MULTIPLE_THREADS
		g[n3*(j3 + k) + j3 + l] += df*d2jj[k][l] + dg*dj[k]*dj[l];
#endif
	      }
	      for (l = 0; l < 3; l++) {
		g[n3*(i3 + k) + j3 + l] += df*d2ij[k][l] + dg*di[k]*dj[l];
	      }
	    }
	  }

	  /*
	   * Add to df the contribution to the gradient vector from
	   * equation 1 so that it will be added to the gradient vector
	   * along with the contributions from the Van der Waals and
	   * Coulombic terms.
	   */

	  df += a_dij;

	  /*
	   * Complete the first derivatives of Dij for the Van der Waals
	   * and Coulombic interactions, as well as for the component of
	   * the first derivatives that is contributed by equation 1.
	   *
	   * When in single-threaded mode sum to the gradient vector the
	   * derivatives that are computed relative to the cartesian
	   * coordinates of atoms i and j.  When in multi-threaded mode
	   * sum the derivatives that are computed relative to the cartesian
	   * coordinates of atom i only.
	   */

	  dedx = df * xij;
	  dedy = df * yij;
	  dedz = df * zij;

	  f[i3    ] += dedx;
	  f[i3 + 1] += dedy;
	  f[i3 + 2] += dedz;

#ifndef MULTIPLE_THREADS
	  f[j3    ] -= dedx;
	  f[j3 + 1] -= dedy;
	  f[j3 + 2] -= dedz;
#endif

	  /* The following code sets up the sums for equations 5-16. */

	  temp5 = 0.5 * temp6;
	  temp6 = fgbi * fgbi * temp5;
	  temp7 = qiqj * (*kappa) * (*kappa) * KSCALE * KSCALE * temp4 * expmkf;
	  temp8 = 1 - 0.25 * efac;
	  temp9 = 0.25 * temp7 - 1.5 * temp6;
	  temp10 = 0.25 * r2;
	  temp11 = rb2 + temp10;
	  temp12 = ri2 * rj2;
	  temp13 = ri * temp11;
	  temp14 = rj * temp11;
	  temp15 = temp10 * rjinv;
	  temp16 = temp10 * riinv;
	  temp17 = temp9 * temp13 * efac;
	  temp18 = temp9 * temp14 * efac;
	  temp19 = temp5 * efac;
	  temp20 = temp8 * temp9;
	  temp21 = temp20 * temp13;
	  temp22 = temp20 * temp14;
	  temp23 = 2.0 * efac;

	  d_dij = temp23 * (temp21 - 0.25 * temp15 * temp5);

	  e_dij = temp23 * (temp22 - 0.25 * temp16 * temp5);

	  f_dij = 2.0 * temp8 * temp5;

	  g_dij = 4.0 * temp8 * temp8 * temp9
	    + (2.0 * temp8 + efac * temp10 * rirjinv ) * r2inv * temp5;

	  h_dij = efac * (temp10 * (temp13 * rjinv - ri2) * temp5 + temp13 * temp17);


	  i_dij = efac * ((temp12 + temp16 * temp13) * temp5 + temp13 * temp18);

	  j_dij = temp23 * ((ri - rjinv * temp11) * 0.25 * temp5 + temp21);

	  k_dij = temp13 * temp19;

	  l_dij = efac * ((temp12 + temp15 * temp14) * temp5 + temp14 * temp17);

	  m_dij = efac * (temp10 * (temp14 * riinv - rj2) * temp5 + temp14 * temp18);

	  n_dij = temp23 * ((rj - riinv * temp11) * 0.25 * temp5 + temp22);

	  o_dij = temp14 * temp19;

	  /*
	   * Evaluate equations 15 and 16 and sum the results to the Hessian
	   * matrix.  These equations involve the derivatives d(Dij)/d(Xl),
	   * d(Dij)/d(Xm) and d2(Dij)/d(Xl)d(Xm), which derivatives exist
	   * only for values of k==i, k==j, l==i and l==j.
	   *
	   * When in single-threaded mode sum to the Hessian the ii, ij and
	   * jj terms only.  When in multi-threaded mode sum to the ii and ij
	   * terms only.
	   *
	   * The ii and jj terms lie on the diagonal so only the upper triangle 
	   * needs to be updated for these 3x3 matrices.  However, the ij term
	   * lies off the diagonal so the full 3x3 matrix must be updated.
	   */

	  df = f_dij;
	  dg = g_dij;

	  for (k = 0; k < 3; k++) {
	    for (l = k; l < 3; l++) {
	      g[n3*(i3 + k) + i3 + l] += df*d2ii[k][l] + dg*di[k]*di[l];
#ifndef MULTIPLE_THREADS
	      g[n3*(j3 + k) + j3 + l] += df*d2jj[k][l] + dg*dj[k]*dj[l];
#endif
	    }
	    for (l = 0; l < 3; l++) {
	      g[n3*(i3 + k) + j3 + l] += df*d2ij[k][l] + dg*di[k]*dj[l];
	    }
	  }

	  /*
	   * Setup the multiplication indicated by equations 13 and 14.
	   * The index selects the second derivative to be calculated, i.e.,
	   * d2(Ri) or d2(Rj).  The array sumdeijd2 is used to collect
	   * the terms that subsequently will be multiplied by the
	   * second derivatives of the effective radii.  When in single-
	   * threaded mode sum the terms that will be multiplied by d2(Ri)
	   * and d2(Rj).  When in multi-threaded mode sum only the term
	   * that will be multiplied by d2(Ri).
	   */

	  sumdeijd2[i] += k_dij;

#ifndef MULTIPLE_THREADS
	  sumdeijd2[j] += o_dij;
#endif

	  /*
	   * Setup the inner differentiations indicated by equations 9 and 12.
	   * The column index selects the first derivative to be calculated,
	   * i.e., d(Ri) or d(Rj).  The row index is i because we are
	   * setting up the outer multiplication by d(Ri).
	   */

	  sumdeijdn[in + i] += h_dij;
	  sumdeijdn[in + j] += l_dij;

	  /*
	   * Setup the inner differentiations indicated by equations 10 and 11.
	   * The column index selects the first derivative to be calculated,
	   * i.e., d(Ri) or d(Rj).  The row index is j because we are
	   * setting up the outer multiplication by d(Rj).  When in single-
	   * threaded mode sum all terms.  When in multi-threaded mode sum
	   * only the term that will be multiplied by d(Rj)*d(Ri).
	   */

	  sumdeijdn[jn + i] += i_dij;

#ifndef MULTIPLE_THREADS
	  sumdeijdn[jn + j] += m_dij;
#endif

	  /*
	   * Perform the inner differentiation indicated by equations 5, 6, 7 and 8.
	   * The row or first index is selected by i because we are setting up the
	   * outer differentiation by the derivatives of Ri.  The column or second
	   * index is selected by i or j because the Dij term affects the derivatives
	   * of atoms i and j.  Both the sumdeijdg and sumdeijdh arrays are used
	   * because equations 5 and 6 perform inner differentiation with respect to
	   * Xl, whereas equations 7 and 8 perform inner differentiation with respect
	   * to Xm.  Distinguishing these two cases is accomplished by transposing
	   * the Hessian addresses, as will be done in the final loop below.
	   *
	   * When in multi-threaded mode sum into the sumdeigdg and sumdeijdh arrays
	   * all terms with the exception of those terms that will contribute to the
	   * multiplication of d(Rj) times the j-component of d(Dij).
	   */

	  for (k = 0; k < 3; k++) {
	    sumdeijdg[in3 + i3 + k] += d_dij * di[k];
	    sumdeijdg[in3 + j3 + k] += d_dij * dj[k];
	    sumdeijdg[jn3 + i3 + k] += e_dij * di[k];
	    sumdeijdh[in3 + i3 + k] += j_dij * di[k];
	    sumdeijdh[in3 + j3 + k] += j_dij * dj[k];
	    sumdeijdh[jn3 + i3 + k] += n_dij * di[k];
#ifndef MULTIPLE_THREADS
	    sumdeijdg[jn3 + j3 + k] += e_dij * dj[k];
	    sumdeijdh[jn3 + j3 + k] += n_dij * dj[k];
#endif
	  }
	}

	/* Sum the energy terms. */

	sumepol[i] += epol;
	sumelec[i] += elec;
	sumevdw[i] += evdw;
      }

      /*
       * When in single-threaded mode update iexcl.  This update is unnecessary
       * in multi-threaded mode because jexcl is loaded from Istr[i] not iexcl.
       */

#ifndef MULTIPLE_THREADS
      iexcl += nexcl;
#endif
    }

#ifdef MULTIPLE_THREADS

    /*
     * Initialize the iexw array used for skipping excluded atoms.
     * Note that because of the manner in which iexw is used, it
     * is necessary to initialize it before only the first iteration
     * of the following loop.
     */

    for (i = -1; i < n; i++) {
      iexw[i] = -1;
    }

    /*
     * In this nest of loops sum data relative only to atom j so that
     * OpenMP can be used to parallelize the outer loop.
     */

#pragma omp for schedule (dynamic, 8)

    /* Loop over all atoms j. */

    for (j = 0; j < n; j++) {

      /*
       * Skip the pair calculations if there are no atoms i on the
       * pair list of atom j.
       */

      nexcl = Jblo[j];
      npairs = lpairs[j];
      if ( npairs > 0 ) {

	j3 = 3 * j;
	jn = n * j;
	jn3 = n * j3;

	xj = x[j3 + 0];
	yj = x[j3 + 1];
	zj = x[j3 + 2];

	qj = q[j];
	rj = reff[j];
	rj2 = rj * rj;
	rjinv = 1.0 / rj;

	/*
	 * Expand the excluded list into the iexw array by storing j
	 * at array address i.
	 */

	iexcl = Jstr[j];
	iexcl_last = iexcl + nexcl;
	for (i = iexcl; i < iexcl_last; i++) {
	  iexw[JExclAt[i] - 1] = j;
	}

	/* Select atoms i from the pair list. */

	for (m = lprstr[j]; m < lprstr[j] + npairs; m++) {

	  i = pairlist[m];

	  in = n * i;
	  i3 = 3 * i;
	  in3 = n * i3;

	  xij = x[i3 + 0] - xj;
	  yij = x[i3 + 1] - yj;
	  zij = x[i3 + 2] - zj;
	  r2 = xij * xij + yij * yij + zij * zij;
	  r2inv = 1.0 / r2;

	  iaci = prm->Ntypes * (prm->Iac[i] - 1);

	  qi = q[i];
	  qiqj = q[i] * qj;
	  ri = reff[i];
	  ri2 = ri * ri;
	  riinv = 1.0 / ri;
	  rirjinv = riinv * rjinv;
	  rb2 = ri * rj;
	  efac = exp(-r2 / (4.0 * rb2));
	  fgbi = 1.0 / sqrt(r2 + rb2 * efac);
	  fgbk = -(*kappa) * KSCALE / fgbi;

	  /*
	   * Calculate the "non-diagonal" energy term, i.e., the term that is a
	   * function of the effective radius Ri and the interatomic distance Dij.
	   */ 

	  expmkf = exp(fgbk) / (*diel_ext);
	  dielfac = 1.0 - expmkf;

	  temp4 = fgbi * fgbi * fgbi;
	  temp6 = qiqj * temp4 * (dielfac + fgbk * expmkf);

	  /* The following term is required by equation 1. */

	  a_dij = temp6 * (1.0 - 0.25 * efac);

	  /*
	   * Prepare to compute the 1st derivatives of the effective radii.
	   * The following code performs the sums indicated by equations 2 and 3.
	   * Compute the contribution of the non-diagonal energy term to the sum
	   * by which the derivative of Rj will be multiplied.
	   */

	  temp5 = 0.5 * efac * temp6 * (rb2 + 0.25 * r2);
	  c_dij = rj * temp5;

	  sumdeijda[j] += c_dij;

	  /*
	   * Calculate the first and second derivatives of the interatomic
	   * distance Dij with respect to the cartesian coordinates of atoms
	   * i and j.  The results are placed into two arrays:
	   *
	   *   dj[] for the first derivatives with respect to atom j
	   *   d2jj[] for the second derivatives with respect to atom j
	   *
	   * Some useful symmetry obtains.  The d2ii and d2ij arrays
	   * are symmetric in that their lower and upper triangles are
	   * the transposes of one another.  Also, d2jj equals d2ii, and
	   * d2ij is the negative of d2ii.
	   *
	   * Note: the additional factor of the interatomic distance Dij that
	   * ought to be included in the denominator of the following equations
	   * is provided by the absence of Dij in equations 15 and 16, as well as
	   * by one less power of Dij in the denominator of the equations for the
	   * derivatives of the Van der Waals and Coulombic energy terms.
	   */
      
	  dj[0] = -xij;
	  dj[1] = -yij;
	  dj[2] = -zij;

	  /* Load the upper triangle of d2jj. */

	  d2jj[0][0] = 1.0 - xij * xij * r2inv;
	  d2jj[0][1] = -xij * yij * r2inv;
	  d2jj[0][2] = -xij * zij * r2inv;

	  d2jj[1][1] = 1.0 - yij * yij * r2inv;
	  d2jj[1][2] = -yij * zij * r2inv;

	  d2jj[2][2] = 1.0 - zij * zij * r2inv;

	  /*
	   * Compute the Van der Waals and Coulomb energy here because
	   * if egb2 is called nbond2 is not called.  Remember that the
	   * derivatives of Dij with respect to x, y and z supply one more
	   * power of 1/Dij but that power of 1/Dij is included in the
	   * equations below, instead of with the derivatives of Dij.
	   *
	   * Compute the Van der Waals and Coulombic energies for only
	   * those pairs which are not on the excluded atom list.  Any
	   * pair on the excluded atom list will have atom i stored at
	   * address j of the iexw array.  It is not necessary to reset
	   * the elements of the iexw array to -1 between successive
	   * iterations in i because an i,j pair is uniquely identified
	   * by atom i stored at array address j.  Thus for example, the
	   * i,j+1 pair would be stored at the same address as the i,j
	   * pair but after the i,j pair were used.
	   *
	   * Initialize df here instead of within the following if statement
	   * so that df is initialized unconditionally.
	   */

	  df = 0.0;

	  if (iexw[i] != j) {

	    /* Initialize dg. */

	    dg = 0.0;

	    rinv = 1. / sqrt(r2);
	    r4inv = r2inv * r2inv;

	    /*  gas-phase Coulomb energy:  */

	    eel = qiqj * rinv;

	    df -= eel;
	    dg += 2. * eel;

	    /* Lennard-Jones energy:   */

	    ic = prm->Cno[iaci + prm->Iac[j] - 1] - 1;
	    if( ic >= 0 ){
	      r6inv = r2inv * r2inv * r2inv;
	      f6 = prm->Cn2[ic] * r6inv;
	      f12 = prm->Cn1[ic] * r6inv * r6inv;

	      df += 6. * f6 - 12. * f12;
	      dg += 156. * f12 - 42. * f6;
	    }

	    df *= r2inv;
	    dg *= r4inv;

	    /*
	     * Calculate the second derivatives of Dij for the Van der Waals
	     * and Coulombic interactions.  This calculation uses df and dg
	     * for these interactions only.
	     *
	     * Sum the jj term into the Hessian.
	     *
	     * The jj term lies on the diagonal so only the upper triangle 
	     * needs to be updated for this 3x3 matrix.
	     */

	    for (k = 0; k < 3; k++) {
	      for (l = k; l < 3; l++) {
		g[n3*(j3 + k) + j3 + l] += df*d2jj[k][l] + dg*dj[k]*dj[l];
	      }
	    }
	  }

	  /*
	   * Add to df the contribution to the gradient vector from
	   * equation 1 so that it will be added to the gradient vector
	   * along with the contributions from the Van der Waals and
	   * Coulombic terms.
	   */

	  df += a_dij;

	  /*
	   * Complete the first derivatives of Dij for the Van der Waals
	   * and Coulombic interactions, as well as for the component of
	   * the first derivatives that is contributed by equation 1.
	   *
	   * Sum the derivatives that are computed relative to the cartesian
	   * coordinates of atom j.
	   */

	  dedx = df * xij;
	  dedy = df * yij;
	  dedz = df * zij;

	  f[j3    ] -= dedx;
	  f[j3 + 1] -= dedy;
	  f[j3 + 2] -= dedz;

	  /* The following code sets up the sums for equations 5-16. */

	  temp5 = 0.5 * temp6;
	  temp6 = fgbi * fgbi * temp5;
	  temp7 = qiqj * (*kappa) * (*kappa) * KSCALE * KSCALE * temp4 * expmkf;
	  temp8 = 1 - 0.25 * efac;
	  temp9 = 0.25 * temp7 - 1.5 * temp6;
	  temp10 = 0.25 * r2;
	  temp11 = rb2 + temp10;
	  temp12 = ri2 * rj2;
	  temp13 = ri * temp11;
	  temp14 = rj * temp11;
	  temp15 = temp10 * rjinv;
	  temp16 = temp10 * riinv;
	  temp17 = temp9 * temp13 * efac;
	  temp18 = temp9 * temp14 * efac;
	  temp19 = temp5 * efac;
	  temp20 = temp8 * temp9;
	  temp21 = temp20 * temp13;
	  temp22 = temp20 * temp14;
	  temp23 = 2.0 * efac;

	  e_dij = temp23 * (temp22 - 0.25 * temp16 * temp5);

	  f_dij = 2.0 * temp8 * temp5;

	  g_dij = 4.0 * temp8 * temp8 * temp9
	    + (2.0 * temp8 + efac * temp10 * rirjinv ) * r2inv * temp5;

	  m_dij = efac * (temp10 * (temp14 * riinv - rj2) * temp5 + temp14 * temp18);

	  n_dij = temp23 * ((rj - riinv * temp11) * 0.25 * temp5 + temp22);

	  o_dij = temp14 * temp19;

	  /*
	   * Evaluate equations 15 and 16 and sum the results to the Hessian
	   * matrix.  These equations involve the derivatives d(Dij)/d(Xl),
	   * d(Dij)/d(Xm) and d2(Dij)/d(Xl)d(Xm), which derivatives exist
	   * only for values of k==i, k==j, l==i and l==j.
	   *
	   * Sum the jj term into the Hessian.
	   *
	   * The jj term lies on the diagonal so only the upper triangle 
	   * needs to be updated for this 3x3 matrix.
	   */

	  df = f_dij;
	  dg = g_dij;

	  for (k = 0; k < 3; k++) {
	    for (l = k; l < 3; l++) {
	      g[n3*(j3 + k) + j3 + l] += df*d2jj[k][l] + dg*dj[k]*dj[l];
	    } 
	  }

	  /*
	   * Setup the multiplication indicated by equations 13 and 14.
	   * The index selects the second derivative to be calculated, i.e.,
	   * d2(Rj).
	   */

	  sumdeijd2[j] += o_dij;

	  /*
	   * Setup the inner differentiations indicated by equations 10 and 11.
	   * The column index selects the first derivative to be calculated,
	   * i.e., d(Rj).  The row index is j because we are setting up the
	   * outer multiplication by d(Rj).
	   */

	  sumdeijdn[jn + j] += m_dij;

	  /*
	   * Perform the inner differentiation indicated by equations 5, 6, 7 and 8.
	   * The row or first index is selected by j because we are setting up the
	   * outer differentiation by the derivatives of Rj.  The column or second
	   * index is selected by i or j because the Dij term affects the derivatives
	   * of atoms i and j.  Both the sumdeijdg and sumdeijdh arrays are used
	   * because equations 5 and 6 perform inner differentiation with respect to
	   * Xl, whereas equations 7 and 8 perform inner differentiation with respect
	   * to Xm.  Distinguishing these two cases is accomplished by transposing
	   * the Hessian addresses, as will be done in the final loop below.
	   */

	  for (k = 0; k < 3; k++) {
	    sumdeijdg[jn3 + j3 + k] += e_dij * dj[k];
	    sumdeijdh[jn3 + j3 + k] += n_dij * dj[k];
	  }
	}
      }
    }

#endif

    /* Free the iexw array within this potentially parallel region of code. */

    free_ivector(iexw, -1, n);

    /*
     * Compute the first and second derivatives of the effective radius.
     * Update the gradient vector using the first derivatives.  Complete
     * equations 2, 3 and 4.  Begin equations 13, 14 and 17 which update
     * the Hessian matrix.  When in single-threaded mode update the i and j
     * elements of the gradient vector as well as the ii, ij, ji and jj
     * elements of the Hessian matrix.  When in multi-threaded mode update
     * only the i elements of the gradient vector as well as the ii and ij
     * elements of the Hessian matrix so that OpenMP can be used to parallelize
     * the outer loop.
     */

#ifdef MULTIPLE_THREADS
#pragma omp for
#endif

    /* Loop over all atoms i. */

    for (i = 0; i < n; i++) {

      /*
       * Don't calculate derivatives of the effective radius of atom i
       * if atom i is frozen or if there are no pair atoms j associated
       * with atom i.
       */

      npairs = lpairs[i] + upairs[i];
      if ( frozen[i] || !npairs ) continue;

      /* These variables will hold the sum of first derivatives of Ri. */

      dx = dy = dz = 0.0;

      i3 = 3 * i;
      in = n * i;
      in3 = 3 * in;

      /*
       * The following captures all terms that are to be multiplied by the
       * first derivative of the effective radius (equations 2, 3 and 4).
       */

      sumda = sumdeijda[i];

      if ( gb>1 ) {

	ri = rborn[i] - BOFFSET;
	thi = tanh( (gbalpha - gbbeta*psi[i] + gbgamma*psi[i]*psi[i])*psi[i] );
	sumda *= (gbalpha -2.0*gbbeta*psi[i] + 3.0*gbgamma*psi[i]*psi[i])
	  *(1.0 - thi*thi)*ri/rborn[i];
      }

      /*
       * The following captures all terms that are to be multiplied by the
       * second derivative of the effective radius (equations 13, 14 and 17).
       */

      sumd2 = sumdeijd2[i];

      /* Prepare to compute the derivatives of the effective radius Ri. */

      xi = x[i3 + 0];
      yi = x[i3 + 1];
      zi = x[i3 + 2];
      ri = rborn[i] - BOFFSET;
      ri1i = 1. / ri;

      /* Select atom j from the pair list. */

      for (m = lprstr[i]; m < lprstr[i] + npairs; m++) {

	j = pairlist[m];

	j3 = 3 * j;

	/* The dreff array must be initialized because of the continue statements. */

	dreff[in3 + j3 + 0] = dreff[in3 + j3 + 1] = dreff[in3 + j3 + 2] = 0.0;

	xij = xi - x[j3 + 0];
	yij = yi - x[j3 + 1];
	zij = zi - x[j3 + 2];
	r2 = xij * xij + yij * yij + zij * zij;

	/* Ignore the ij atom pair if their separation exceeds the GB cutoff. */

	if ( r2 > rgbmaxpsmax2 ) continue;

	dij1i = 1.0 / sqrt(r2);
	dij2i = dij1i * dij1i;
	dij = r2 * dij1i;
	sj = fs[j] * (rborn[j] - BOFFSET);
	sj2 = sj * sj;

	/*
	 * The following are the numerator of the first derivatives of the
	 * equations from the Appendix of Schaefer and Froemmel and of the
	 * Taylor series expansion for d>>s by Andreas Svrcek-Seiler.  Smooth
	 * rgbmax idea is from Andreas Svrcek-Seiler and Alexey Onufriev.  The
	 * derivatives are with respect to Dij.  The complete derivative is
	 * formed by multiplying the numerator by -Ri*Ri.  The factor of Ri*Ri
	 * has been moved to equations 5-18.  The negation is deferred until later.
	 * When the chain rule is used to form the first derivatives of the
	 * effective radius with respect to the cartesian coordinates, an
	 * additional factor of Dij appears in the denominator.  That factor
	 * is included in the following expressions.
	 */

	if ( dij > rgbmax + sj ) continue;

	if ( dij > rgbmax - sj ) {

	  temp1 = 1./(dij - sj);
	  dij3i = dij1i * dij2i;

	  datmp = 0.125 * dij3i * ((r2 + sj2) *
				   (temp1 * temp1 - rgbmax2i) -
				   2.0 * log(rgbmax * temp1));

	} else if ( dij > 4.0 * sj ) {

	  tmpsd = sj2 * dij2i;
	  dumbo = TE + tmpsd*(TF + tmpsd*(TG + tmpsd*(TH + tmpsd*THH)));
	  datmp = tmpsd * sj * dij2i * dij2i * dumbo;

	} else if ( dij > ri + sj ) {

	  temp1 = 1. / (r2 - sj2);
	  datmp = temp1 * sj * (-0.5 * dij2i + temp1)
	    + 0.25 * dij1i * dij2i * log((dij - sj) / (dij + sj));

	} else if ( dij > fabs(ri - sj) ) {

	  temp1 = 1. / (dij + sj);
	  dij3i = dij2i * dij1i;
	  datmp =
	    -0.25 * (-0.5 * (r2 - ri * ri + sj2) * dij3i * ri1i *
		     ri1i + dij1i * temp1 * (temp1 - dij1i)
		     - dij3i * log(ri * temp1));

	} else if ( ri < sj ) {

	  temp1 = 1. / (r2 - sj2);
	  datmp =
	    -0.5 * (sj * dij2i * temp1 - 2. * sj * temp1 * temp1 -
		    0.5 * dij2i * dij1i * log((sj - dij) /
					      (sj + dij)));

	} else continue;

	/*
	 * Negate the first derivative of Ri with respect to the
	 * interatomic distance in order to account for the
	 * minus sign of -Ri*Ri.  The factor of Ri*Ri is omitted
	 * because it was included in equations 5-18.  These
	 * derivatives contain the appropriate factor of Dij
	 * in the denominator because it is supplied by datmp.
	 */

	df = -datmp;

	/*
	 * The following are the numerator of the second derivatives of the
	 * equations from the Appendix of Schaefer and Froemmel and of the
	 * Taylor series expansion for d>>s by Andreas Svrcek-Seiler.  Smooth
	 * rgbmax idea is from Andreas Svrcek-Seiler and Alexey Onufriev.  The
	 * derivatives are with respect to Dij.
	 */

	if ( dij > rgbmax + sj ) continue;

	if ( dij > rgbmax - sj ) {

	  temp1 = 1./(dij - sj);
	  temp2 = temp1 * temp1;
	  dij3i = dij1i * dij2i;

	  da2tmp = -0.25 * (dij3i * (sj2 * rgbmax2i - 1.0) + dij1i * temp2 +
			    2.0 * (log(rgbmax * temp1) * dij3i +
				   temp1 * (dij2i - temp2)));

	} else if ( dij > 4.0 * sj ) {

	  tmpsd = sj2 * dij2i;
	  dumbo = TI + tmpsd*(TJ + tmpsd*(TK + tmpsd*(TL + tmpsd*TLL)));
	  da2tmp = tmpsd * sj * dij2i * dij2i * dumbo;


	} else if ( dij > ri + sj ) {

	  temp1 = 2. * dij * sj;
	  temp2 = (dij - sj) * (dij + sj);
	  dij3i = dij2i * dij1i;

	  da2tmp = (0.5 * log((dij - sj) / (dij + sj)) +
		    dij * sj * ((r2 + sj2) * (r2 - sj2) + temp1 * temp1) /
		    (temp2 * temp2 * temp2)) * dij3i;

	} else if ( dij > fabs(ri - sj) ) {

	  temp1 = dij + sj;
	  temp2 = dij * temp1;
	  ri2 = ri * ri;

	  da2tmp = ((2. * ri2 * log(ri / temp1) + sj2) * temp1 * temp1 * temp1
		    + ri2 * sj * (2. * dij * dij - dij * sj - sj2)) /
	    (4. * ri2 * temp2 * temp2 * temp2);

	} else if ( ri < sj ) {

	  temp1 = sj2 - r2;
	  temp2 = 2. * sj * dij;
	  dij3i = dij2i * dij1i;

	  da2tmp = (0.5 * log((sj - dij) / (sj + dij)) + sj * dij *
		    (temp1 * (sj2 + r2) - temp2 * temp2) /
		    (temp1 * temp1 * temp1)) * dij3i;

	} else continue;

	/* 
	 * Calculate the second derivative of the effective radius
	 * with respect to the interatomic distance.  A factor of
	 * the Ri*Ri has been omitted from the second derivative
	 * because it was included in equations 5-18.
	 *
	 * When the chain rule is used to form the second derivatives of the
	 * effective radius with respect to the cartesian coordinates, an
	 * additional factor of Dij*Dij appears in the denominator.  That
	 * factor is included in the following expression.
	 */

	dg = da2tmp * dij2i;

	/*
	 * Calculate the first and second derivatives of the interatomic
	 * distance Dij with respect to the cartesian coordinates of atoms
	 * i and j.  The results are placed into five arrays:
	 *
	 *   di[] for the first derivatives with respect to atom i
	 *   dj[] for the first derivatives with respect to atom j
	 *   d2ii[] for the second derivatives with respect to atom i
	 *   d2jj[] for the second derivatives with respect to atom j
	 *   d2ij[] for the second derivatives with respect to atoms i and j
	 *
	 * Some useful symmetry obtains.  The d2ii and d2ij arrays
	 * are symmetric in that their lower and upper triangles are
	 * the transposes of one another.  Also, d2jj equals d2ii, and
	 * d2ij is the negative of d2ii.
	 *
	 *
	 * Note: the additional factor of the interatomic distance Dij that
	 * ought to be included in the denominator of the following equations
	 * is included instead in the first and second derivatives of the
	 * effective radius with respect to Dij (df and dg) that were calculated
	 * above.
	 */
      
	di[0] = xij;
	di[1] = yij;
	di[2] = zij;

	dj[0] = -xij;
	dj[1] = -yij;
	dj[2] = -zij;

	/* Load the upper triangle of d2ii. */

	r2inv = 1.0 / r2;

	d2ii[0][0] = 1.0 - xij * xij * r2inv;
	d2ii[0][1] = -xij * yij * r2inv;
	d2ii[0][2] = -xij * zij * r2inv;

	d2ii[1][1] = 1.0 - yij * yij * r2inv;
	d2ii[1][2] = -yij * zij * r2inv;

	d2ii[2][2] = 1.0 - zij * zij * r2inv;

	/* Finish loading the rest of all of the matrices. */

	for (k = 0; k < 3; k++) {

	  /* Load the upper triangles of d2jj and d2ij. */

	  for (l = k; l < 3; l++) {
#ifndef MULTIPLE_THREADS
	    d2jj[k][l] = d2ii[k][l];
#endif
	    d2ij[k][l] = -d2ii[k][l];
	  }

	  /* Load the symmetric elements of d2ij. */

	  for (l = k+1; l < 3; l++) {
	    d2ij[l][k] = d2ij[k][l];
	  }
	}

	/* Store the 1st derivatives of Ri with respect to (xj,yj,zj). */

	dxj = df * dj[0];
	dyj = df * dj[1];
	dzj = df * dj[2];

	dreff[in3 + j3 + 0] = dxj;
	dreff[in3 + j3 + 1] = dyj;
	dreff[in3 + j3 + 2] = dzj;

	/*
	 * Update the gradient vector with d(Ri)/d(Xj) when in
	 * single-threaded mode.  When in multi-threaded mode
	 * defer the update until the next nested loops.
	 */

#ifndef MULTIPLE_THREADS
	f[j3    ] += sumda * dxj;
	f[j3 + 1] += sumda * dyj;
	f[j3 + 2] += sumda * dzj;
#endif

	/*
	 * Sum the 1st derivatives of Ri with respect to (xi,yi,zi).
	 * Exploit the fact that d(Dij)/d(Xi) = -d(Dij)/d(Xj).
	 */

	dx -= dxj;
	dy -= dyj;
	dz -= dzj;

	/*
	 * Form the second derivatives of the effective radius with
	 * respect to the cartesian coordinates using the chain rule,
	 * multiply the result by sumd2 and sum to the Hessian matrix.
	 * This step implements equations 13, 14 and 17, except that
	 * terms such as d2(Ri)/d(Dij)d(Dik) are not handled here
	 * but instead are handled in the final loop below.
	 *
	 * When in single-threaded mode update the Hessian using the ii and
	 * jj terms, as well as the ij term if j > i, or the ji term if
	 * j < i.  When in multi-threaded mode update the Hessian using the
	 * ii term, as well as the ij term if j > i.
	 *
	 * The ii term lies on the diagonal so only the upper triangle 
	 * needs to be updated for this 3x3 matrix.  However, the ij
	 * term lies off the diagonal so the full 3x3 matrix must be updated.
	 */

	for (k = 0; k < 3; k++) {
	  for (l = k; l < 3; l++) {
	    g[n3*(i3 + k) + i3 + l] += sumd2*(df*d2ii[k][l] + dg*di[k]*di[l]);
#ifndef MULTIPLE_THREADS
	    g[n3*(j3 + k) + j3 + l] += sumd2*(df*d2jj[k][l] + dg*dj[k]*dj[l]);
#endif
	  }
	  for (l = 0; l < 3; l++) {
	    if (j > i) {
	      g[n3*(i3 + k) + j3 + l] += sumd2*(df*d2ij[k][l] + dg*di[k]*dj[l]);
	    }
#ifndef MULTIPLE_THREADS
	    else {
	      g[n3*(j3 + l) + i3 + k] += sumd2*(df*d2ij[k][l] + dg*di[k]*dj[l]);
	    }
#endif
	  }
	}
      }

      /*
       * Store the sum of first derivatives of Ri with respect to (xi,yi,zi)
       * in the diagonal elements of the dreff array.  These derivatives
       * include the appropriate factor of Dij but not a factor of Ri*Ri.
       * They include the minus sign of -Ri*Ri that was supplied in df.
       */

      dreff[in3 + i3 + 0] = dx;
      dreff[in3 + i3 + 1] = dy;
      dreff[in3 + i3 + 2] = dz;

      /* Update the gradient vector with d(Ri)/d(Xi). */

      f[i3    ] += sumda * dx;
      f[i3 + 1] += sumda * dy;
      f[i3 + 2] += sumda * dz;
    }

    /*
     * This nest of loops sums data relative only to atom j so that
     * OpenMP can be used to parallelize the outer loops.
     * 
     * The control flow in this nest of loops must be identical to 
     * that of the preceeding nest of loops.  Perform the minimum
     * calculation necessary to reconstitute the conditions of
     * the preceeding nest of loops.
     */

#ifdef MULTIPLE_THREADS
#pragma omp for

    /* Loop over all atoms j. */

    for (j = 0; j < n; j++) {

      /*
       * Don't calculate derivatives with respect to atom j if there
       * are no atoms i associated with atom j.
       */

      npairs = lpairs[j] + upairs[j];
      if ( !npairs ) continue;

      j3 = 3 * j;

      xj = x[j3 + 0];
      yj = x[j3 + 1];
      zj = x[j3 + 2];

      sj = fs[j] * (rborn[j] - BOFFSET);
      sj2 = sj * sj;

      /* Select atom i from the pair list. */

      for (m = lprstr[j]; m < lprstr[j] + npairs; m++) {

	i = pairlist[m];

	/*
	 * Don't calculate derivatives of the effective radius of atom i
	 * if atom i is frozen.
	 */

	if ( frozen[i] ) continue;

	i3 = 3 * i;
	in = n * i;

	xij = x[i3    ] - xj;
	yij = x[i3 + 1] - yj;
	zij = x[i3 + 2] - zj;
	r2 = xij * xij + yij * yij + zij * zij;

	/* Ignore the ij atom pair if their separation exceeds the GB cutoff. */

	if ( r2 > rgbmaxpsmax2 ) continue;

	dij1i = 1.0 / sqrt(r2);
	dij2i = dij1i * dij1i;
	dij = r2 * dij1i;

	ri = rborn[i] - BOFFSET;
	ri1i = 1. / ri;

	/*
	 * The following are the numerator of the first derivatives of the
	 * equations from the Appendix of Schaefer and Froemmel and of the
	 * Taylor series expansion for d>>s by Andreas Svrcek-Seiler.  Smooth
	 * rgbmax idea is from Andreas Svrcek-Seiler and Alexey Onufriev.  The
	 * derivatives are with respect to Dij.  The complete derivative is
	 * formed by multiplying the numerator by -Ri*Ri.  The factor of Ri*Ri
	 * has been moved to equations 5-18.  The negation is deferred until later.
	 * When the chain rule is used to form the first derivatives of the
	 * effective radius with respect to the cartesian coordinates, an
	 * additional factor of Dij appears in the denominator.  That factor
	 * is included in the following expressions.
	 */

	if ( dij > rgbmax + sj ) continue;

	if ( dij > rgbmax - sj ) {

	  temp1 = 1./(dij - sj);
	  dij3i = dij1i * dij2i;

	  datmp = 0.125 * dij3i * ((r2 + sj2) *
				   (temp1 * temp1 - rgbmax2i) -
				   2.0 * log(rgbmax * temp1));

	} else if ( dij > 4.0 * sj ) {

	  tmpsd = sj2 * dij2i;
	  dumbo = TE + tmpsd*(TF + tmpsd*(TG + tmpsd*(TH + tmpsd*THH)));
	  datmp = tmpsd * sj * dij2i * dij2i * dumbo;

	} else if ( dij > ri + sj ) {

	  temp1 = 1. / (r2 - sj2);
	  datmp = temp1 * sj * (-0.5 * dij2i + temp1)
	    + 0.25 * dij1i * dij2i * log((dij - sj) / (dij + sj));

	} else if ( dij > fabs(ri - sj) ) {

	  temp1 = 1. / (dij + sj);
	  dij3i = dij2i * dij1i;
	  datmp =
	    -0.25 * (-0.5 * (r2 - ri * ri + sj2) * dij3i * ri1i *
		     ri1i + dij1i * temp1 * (temp1 - dij1i)
		     - dij3i * log(ri * temp1));

	} else if ( ri < sj ) {

	  temp1 = 1. / (r2 - sj2);
	  datmp =
	    -0.5 * (sj * dij2i * temp1 - 2. * sj * temp1 * temp1 -
		    0.5 * dij2i * dij1i * log((sj - dij) /
					      (sj + dij)));

	} else continue;

	/*
	 * Negate the first derivative of Ri with respect to the
	 * interatomic distance in order to account for the
	 * minus sign of -Ri*Ri.  The factor of Ri*Ri is omitted
	 * because it was included in equations 5-18.  These
	 * derivatives contain the appropriate factor of Dij
	 * in the denominator because it is supplied by datmp.
	 */

	df = -datmp;

	/*
	 * The following are the numerator of the second derivatives of the
	 * equations from the Appendix of Schaefer and Froemmel and of the
	 * Taylor series expansion for d>>s by Andreas Svrcek-Seiler.  Smooth
	 * rgbmax idea is from Andreas Svrcek-Seiler and Alexey Onufriev.  The
	 * derivatives are with respect to Dij.
	 */

	if ( dij > rgbmax + sj ) continue;

	if ( dij > rgbmax - sj ) {

	  temp1 = 1./(dij - sj);
	  temp2 = temp1 * temp1;
	  dij3i = dij1i * dij2i;

	  da2tmp = -0.25 * (dij3i * (sj2 * rgbmax2i - 1.0) + dij1i * temp2 +
			    2.0 * (log(rgbmax * temp1) * dij3i +
				   temp1 * (dij2i - temp2)));

	} else if ( dij > 4.0 * sj ) {

	  tmpsd = sj2 * dij2i;
	  dumbo = TI + tmpsd*(TJ + tmpsd*(TK + tmpsd*(TL + tmpsd*TLL)));
	  da2tmp = tmpsd * sj * dij2i * dij2i * dumbo;


	} else if ( dij > ri + sj ) {

	  temp1 = 2. * dij * sj;
	  temp2 = (dij - sj) * (dij + sj);
	  dij3i = dij2i * dij1i;

	  da2tmp = (0.5 * log((dij - sj) / (dij + sj)) +
		    dij * sj * ((r2 + sj2) * (r2 - sj2) + temp1 * temp1) /
		    (temp2 * temp2 * temp2)) * dij3i;

	} else if ( dij > fabs(ri - sj) ) {

	  temp1 = dij + sj;
	  temp2 = dij * temp1;
	  ri2 = ri * ri;

	  da2tmp = ((2. * ri2 * log(ri / temp1) + sj2) * temp1 * temp1 * temp1
		    + ri2 * sj * (2. * dij * dij - dij * sj - sj2)) /
	    (4. * ri2 * temp2 * temp2 * temp2);

	} else if ( ri < sj ) {

	  temp1 = sj2 - r2;
	  temp2 = 2. * sj * dij;
	  dij3i = dij2i * dij1i;

	  da2tmp = (0.5 * log((sj - dij) / (sj + dij)) + sj * dij *
		    (temp1 * (sj2 + r2) - temp2 * temp2) /
		    (temp1 * temp1 * temp1)) * dij3i;

	} else continue;

	/* 
	 * Calculate the second derivative of the effective radius
	 * with respect to the interatomic distance.  A factor of
	 * the Ri*Ri has been omitted from the second derivative
	 * because it was included in equations 5-18.
	 *
	 * When the chain rule is used to form the second derivatives of the
	 * effective radius with respect to the cartesian coordinates, an
	 * additional factor of Dij*Dij appears in the denominator.  That
	 * factor is included in the following expression.
	 */

	dg = da2tmp * dij2i;

	/*
	 * Calculate the first and second derivatives of the interatomic
	 * distance Dij with respect to the cartesian coordinates of atoms
	 * i and j.  The results are placed into four arrays:
	 *
	 *   di[] for the first derivatives with respect to atom i
	 *   dj[] for the first derivatives with respect to atom j
	 *   d2jj[] for the second derivatives with respect to atom j
	 *   d2ij[] for the second derivatives with respect to atoms i and j
	 *
	 * Note: the additional factor of the interatomic distance Dij that
	 * ought to be included in the denominator of the following equations
	 * is provided by the absence of Dij in equations 15 and 16, as well as
	 * by one less power of Dij in the denominator of the equations for the
	 * derivatives of the Van der Waals and Coulombic energy terms.
	 */
      
	di[0] = xij;
	di[1] = yij;
	di[2] = zij;

	dj[0] = -xij;
	dj[1] = -yij;
	dj[2] = -zij;

	/* Load the upper triangle of d2jj. */

	r2inv = 1.0 / r2;

	d2jj[0][0] = 1.0 - xij * xij * r2inv;
	d2jj[0][1] = -xij * yij * r2inv;
	d2jj[0][2] = -xij * zij * r2inv;

	d2jj[1][1] = 1.0 - yij * yij * r2inv;
	d2jj[1][2] = -yij * zij * r2inv;

	d2jj[2][2] = 1.0 - zij * zij * r2inv;

	/* Finish loading the rest of all of the matrices. */

	for (k = 0; k < 3; k++) {

	  /* Load the upper triangle of d2ij. */

	  for (l = k; l < 3; l++) {
	    d2ij[k][l] = -d2jj[k][l];
	  }

	  /* Load the symmetric elements of d2ij. */

	  for (l = k+1; l < 3; l++) {
	    d2ij[l][k] = d2ij[k][l];
	  }
	}

	/*
	 * The following captures all terms that are to be multiplied by the
	 * second derivative of the effective radius (equations 13, 14 and 17).
	 */

	sumd2 = sumdeijd2[i];

	/*
	 * Compute the jj terms and the ji terms thatare selected
	 * in the case of j < i.
	 *
	 * The jj term lies on the diagonal so only the upper triangle 
	 * needs to be updated for this 3x3 matrix.  However, the ji
	 * term lies off the diagonal so the full 3x3 matrix must be updated.
	 */

	for (k = 0; k < 3; k++) {
	  for (l = k; l < 3; l++) {
	    g[n3*(j3 + k) + j3 + l] += sumd2*(df*d2jj[k][l] + dg*dj[k]*dj[l]);
	  }
	  for (l = 0; l < 3; l++) {
	    if (j < i) {
	      g[n3*(j3 + l) + i3 + k] += sumd2*(df*d2ij[k][l] + dg*di[k]*dj[l]);
	    }
	  }
	}

	/*
	 * The following captures all terms that are to be multiplied by the
	 * first derivative of the effective radius (equations 2, 3 and 4).
	 */

	sumda = sumdeijda[i];

	if ( gb>1 ) {

	  ri = rborn[i] - BOFFSET;
	  thi = tanh( (gbalpha - gbbeta*psi[i] + gbgamma*psi[i]*psi[i])*psi[i] );
	  sumda *= (gbalpha -2.0*gbbeta*psi[i] + 3.0*gbgamma*psi[i]*psi[i])
	    *(1.0 - thi*thi)*ri/rborn[i];
	}

	/* Update the gradient vector with d(Ri)/d(Xj). */

	dxj = df * dj[0];
	dyj = df * dj[1];
	dzj = df * dj[2];

	f[j3    ] += sumda * dxj;
	f[j3 + 1] += sumda * dyj;
	f[j3 + 2] += sumda * dzj;
      }
    }
#endif

    /*
     * Perform the inner multiplications of equations 9, 10, 11, 12
     * and 18 in a manner analogous to the multiplications of
     * of equations 2, 3 and 4.  However, the row index of the
     * sumdeijdn array specifies the atom for which the outer
     * multiplications will be performed later.  It is necessary
     * to process every row of sumdeijdn in a manner analogous
     * to the processing of sumdeijda.  All of the derivatives
     * of the effective radii have already been calculated and
     * are stored in the dreff array which operates in a manner
     * analogous to the manner in which the da array operates
     * in the egb function of the eff.c file.
     *
     * The loop index i selects the ith row in sumdeijdn and sumdeijdg,
     * i.e., the atom for which outer multiplications are now set up by
     * summing a gradient vector into a row of sumdeijdg.  The loop
     * index j selects the jth element in the ith row of sumdeijdn, i.e.,
     * the atom for which the derivatives of the inner multiplication
     * were calculated.  The loop index k specifies the atom relative to
     * whose cartesian coordinates the derivatives of the inner multiplication
     * are calculated.  The combination of loops j and k produces a
     * gradient vector that is summed into the ith row of sumdeijdg.
     */

#ifdef MULTIPLE_THREADS
#pragma omp for
#endif

    for (i = 0; i < n; i++) {
      in = n * i;
      in3 = 3 * in;
      for (j = 0; j < n; j++) {
	jn3 = n3 * j;
	sumdn = sumdeijdn[in + j];
	for (k = 0; k < n3; k++) {
	  sumdeijdg[in3 + k] += sumdn * dreff[jn3 + k];
	}
      }
    }

    /*
     * Perform the outer multiplication for equations 5, 6, 7, 8, 9,
     * 10, 11 and 12.  The derivatives of Ri are stored in dreff in a
     * manner analogous to the use of the da array. The results of the
     * inner multiplications were stored in the sumdeijdg and sumdeijdh
     * arrays.  One row of the sumdeijdg or sumdeijdh array is
     * analogous to one element of the sumdeijda array.  Hence each
     * row of the dreff array is multiplied by the corresponding row of
     * the sumdeijdg or sumdeijdh array.  Each combination of two (x,y,z)
     * coordinates is used to create a 3x3 matrix that is summed into the
     * Hessian matrix.
     *
     * Complete equations 13, 14 and 17 by multiplying each row of dreff
     * by itself.  This process handles terms such as d2(Ri)/d(Dij)d(Dik)
     * which give rise to products such as d(Ri)/d(Dij) * d(Ri)/d(Dik).
     *
     * The loop index i selects the ith row in dreff, sumdeijdg and sumdeijdh.
     * The loop index j selects the jth element from the ith row of dreff.
     * The loop index k selects the kth element from the ith row of sumdeijdg
     * and sumdeijdh.  Both sumdeijdg and sumdeijdh are used because the
     * results of multiplication by sumdeijdg and sumdeijdh are summed into
     * transpose-related Hessian addresses.  This approach preserves the
     * symmetry of the Hessian.
     *
     * Only the upper triangle of the Hessian is updated.
     *
     * Use the nowait clause for the following 'for' loop because
     * there are implied barriers at the end of the loop and at the
     * end of the parallel region; hence, the barrier at the end of
     * the loop is redundant.
     */

#ifdef MULTIPLE_THREADS
#pragma omp for schedule (dynamic, 8) nowait
#endif

    for (j = 0; j < n3; j++) {
      for (i = 0; i < n; i++) {
	in3 = n3 * i;
	sreffi = 2.0 * sumdeijd2[i] * reff[i];
	dreffi = dreff[in3 + j];
	sdreff = dreffi * sreffi;
	sumdh = sumdeijdh[in3 + j];
	for (k = j; k < n3; k++) {
	  g[n3*j + k] +=
	    dreffi * sumdeijdg[in3 + k] + dreff[in3 + k] * (sdreff + sumdh);
	}
      }
    }
  }


  /* Perform non-logarithmic reduction of the energy terms. */

  epol = elec = evdw = 0.0;

  for (i = 0; i < n; i++) {
    epol += sumepol[i];
    elec += sumelec[i];
    evdw += sumevdw[i];
  }

  /* Return elec and evdw through the reference parameters eelt and enb. */

  *eelt = elec;
  *enb = evdw;

  /* Free the other temporary vectors. */

  free_vector(sumepol, 0, n);
  free_vector(sumelec, 0, n);
  free_vector(sumevdw, 0, n);
  free_vector(reff, 0, n);
  free_vector(dreff, 0, nn3);
  free_vector(sumdeijda, 0, n);
  free_vector(sumdeijd2, 0, n);
  free_vector(sumdeijdn, 0, nn);
  free_vector(sumdeijdg, 0, nn3);
  free_vector(sumdeijdh, 0, nn3);

  if ( gb==2 || gb==5 ) free_vector(psi, 0, n);

  return (epol);
}

REAL_T mme2(x, f, g, m, natom, iter)
     REAL_T *x, *f, *g, *m;
     int *natom, *iter;
{

  REAL_T ebh, eba, eth, eta, eph, epa, enb, eel, enb14, eel14, ecn, frms;
  REAL_T ene[20];
  int nb_pairs;
  int i, j, k, n;

  n = 3*prm->Natom;
  t1 = second();
  if (*iter <= 1) {
    printf
      ("    iter    Total    bad         vdW     elect.     cons.   genBorn      frms\n");
    tnonb = tpair = tbond = tangl = tphi = tcons = 0.0;
  }
  if (nb_pairs < 0 || (*iter>0 && *iter % nsnb == 0)) {
    nb_pairs = nblist(x, lpairs, upairs, lprstr, uprstr, pairlist);
    t2 = second();
    tpair += t2 - t1;
    t1 = t2;
    /*		printf( "nb_pairs = %d\n", nb_pairs );          */
  }

  /*
   * Initialize the gradient and Hessian to zero.
   */

  for (i = 0; i < n; i++) {
    f[i] = 0.0;
  }

#ifdef MULTIPLE_THREADS
#pragma omp parallel for
#endif

  for (i = 0; i < n*n; i++) {
    g[i] = 0.0;
  }

  ebh = ebond2(prm->Nbonh, prm->BondHAt1, prm->BondHAt2,
	       prm->BondHNum, prm->Rk, prm->Req, x, f, g);
  eba = ebond2(prm->Mbona, prm->BondAt1, prm->BondAt2,
	       prm->BondNum, prm->Rk, prm->Req, x, f, g);
  ene[3] = ebh + eba;
  t2 = second();
  tbond += t2 - t1;
  t1 = t2;

  eth = eangl2(prm->Ntheth, prm->AngleHAt1, prm->AngleHAt2,
	       prm->AngleHAt3, prm->AngleHNum,
	       prm->Tk, prm->Teq, x, f, g);
  eta = eangl2(prm->Ntheta, prm->AngleAt1, prm->AngleAt2,
	       prm->AngleAt3, prm->AngleNum,
	       prm->Tk, prm->Teq, x, f, g);
  ene[4] = eth + eta;
  t2 = second();
  tangl += t2 - t1;
  t1 = t2;

  eph = ephi2(prm->Nphih, prm->DihHAt1, prm->DihHAt2,
	      prm->DihHAt3, prm->DihHAt4, prm->DihHNum,
	      prm->Pk, prm->Pn, prm->Phase, x, f, g);
  epa = ephi2(prm->Mphia, prm->DihAt1, prm->DihAt2,
	      prm->DihAt3, prm->DihAt4, prm->DihNum,
	      prm->Pk, prm->Pn, prm->Phase, x, f, g);
  ene[5] = eph + epa;

  ene[6] = 0.0;        /*  hbond term not in Amber-94 force field */
  nbond2(prm->N14pairs, prm->N14pairlist, uprstr, 0, x, f, g, &enb14, &eel14,
	 scnb, scee);
  if (e_debug) {
    EXPR("%9.3f", enb14);
    EXPR("%9.3f", eel14);
  }
  ene[7] = enb14;
  ene[8] = eel14;
  t2 = second();
  tphi += t2 - t1;
  t1 = t2;

  if (nconstrained) {
    ecn = econs2(x, f, g);
    t2 = second();
    tcons += t2 - t1;
    t1 = t2;
  } else
    ecn = 0.0;
  ene[9] = ecn;

  /* Calculate the generalized Born energy and derivatives. */

  if (gb) {
    ene[10] = egb2(&prm->Natom, x, f, g, prm->Fs, prm->Rborn, prm->Charges,
		   &kappa, &epsext, &enb, &eel);
    ene[1] = enb;
    ene[2] = eel;
  } else {
    nbond2(upairs, pairlist, uprstr, 1, x, f, g, &enb, &eel, 1.0, 1.0);
    ene[1] = enb;
    ene[2] = eel;
    ene[10] = 0.0;
  }
  t2 = second();
  tnonb += t2 - t1;
  t1 = t2;

  ene[0] = 0.0;
  for (k = 1; k <= 10; k++)
    ene[0] += ene[k];

  /*
   * Copy the upper triangle into the lower triangle.  Note that
   * if some of the energy functions update the full Hessian and
   * others update only the upper triangle, a symmetric Hessian
   * will still result from this copying.
   */

#ifdef MULTIPLE_THREADS
#pragma omp parallel for private (i,j)
#endif

  for (j = 0; j < n; j++) {
    for (i = 0; i < j; i++) {
      g[n*j + i] = g[n*i + j];
    }
  }

  /* zero out frozen forces */

#ifdef MULTIPLE_THREADS
#pragma omp parallel for private (k, j)
#endif

  for (k = 0; k < prm->Natom; k++) {
    if (frozen[k]) {
      f[3 * k + 0] = 0.0;
      f[3 * k + 1] = 0.0;
      f[3 * k + 2] = 0.0;
      for (j = 0; j < prm->Natom; j++) {

	/*
	 * Zero all cross derivatives.  Zero both g[k,j] and g[j,k].  Because
	 * newton attempts to build a symmetric matrix by adding the sums in
	 * g[k,j] and g[j,k], if only one is zeroed the other may contribute
	 * a non-zero sum.
	 */

	g[3*k + 0 + n*(3*j + 0)] = g[3*j + 0 + n*(3*k + 0)] = 0.0;
	g[3*k + 1 + n*(3*j + 1)] = g[3*j + 1 + n*(3*k + 1)] = 0.0;
	g[3*k + 2 + n*(3*j + 2)] = g[3*j + 2 + n*(3*k + 2)] = 0.0;
      }
    }
  }

  /* Compare the 1st derivatives from egb to the 1st derivatives from egb2. */

#undef GRADIENT_DEBUG
#ifdef GRADIENT_DEBUG

  REAL_T *f1 = vector(0, n);
  int iter_dum = 1;
  mme(x, f1, &iter_dum);

  frms = 0.0;

  for (i = 0; i < n; i++) {
    frms += (f[i] - f1[i]) * (f[i] - f1[i]);
  }
  printf("rms delta df = %9.2e\n", sqrt(frms) / ((REAL_T)n) );
  
  free_vector(f1, 0, n);

#endif


  /* Construct the 2nd derivatives by finite difference from the 1st derivatives. */

#undef HESSIAN_DEBUG
#ifdef HESSIAN_DEBUG

#define DELX (0.00001)
#define MULX (50000.0)

  ntpr=2;
  int iter_dumm = 3;
  REAL_T eplus, eminus;
  REAL_T *h = vector(0, n*n);
  REAL_T *d = vector(0, n*n);
  REAL_T *fplus = vector(0, n);
  REAL_T *fminus = vector(0, n);
  for (i = 0; i < n; i++ ){
    x[i] += DELX;
    eplus = mme( x, fplus, &iter_dumm);
    x[i] -= 2.0 * DELX;
    eminus = mme( x, fminus, &iter_dumm);
    x[i] += 0.00001;
    for (j = 0; j < n; j++ ){
      h[n*i + j] = MULX * (fplus[j] - fminus[j]);
    }
  }

  /*
   * Calculate the rms difference between the second derivatives calculated
   * analytically and via finite differences.
   */

  k = 0;
  REAL_T maxdev = 0.0;
  REAL_T mindev = 1.0e10;
  REAL_T dev;
  frms = 0.0;
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      k++;
      dev = (h[n*i + j] - g[n*i + j]) * (h[n*i + j] - g[n*i + j]);
      frms += dev;
      dev = sqrt(dev);
      d[n*i + j] = dev;
      if (dev > maxdev) {
	maxdev = dev;
      }
      if (dev < mindev) {
	mindev = dev;
      }
    }
  }
  printf("rms delta dg = %9.2e   maxdev = %9.2e   mindev = %9.2e\n",
	 sqrt(frms) / ((REAL_T)k), maxdev, mindev);

  /* Use the finite-difference Hessian not the analytic Hessian. */

#undef USE_FINITE_DIFFERENCE
#ifdef USE_FINITE_DIFFERENCE
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      g[n*i + j] = h[n*i + j];
    }
  }
#endif

#undef PRINT_DEVIATION
#ifdef PRINT_DEVIATION
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      printf("deviation[%d][%d] = %9.2e\n", i, j, d[i][j]);
    }
    printf("\n");
  }
  printf("\n");
#endif
  
  ntpr=1;
  free_vector(fminus, 0, 3*prm->Natom);
  free_vector(fplus, 0, 3*prm->Natom);
  free_vector(h, 0, n*n);
  free_vector(d, 0, n*n);
#endif

  /* Check that the Hessian is symmetric. */

#undef TEST_SYMMETRY
#ifdef TEST_SYMMETRY
  k = 0;
  frms = 0.0;
  for (i = 0; i < n; i++) {
    for (j = i + 1; j < n; j++) {
      k++;
      frms += (g[n*i + j] - g[n*j + i]) * (g[n*i + j] - g[n*j + i]);
    }
  }
  printf("rms symmetry = %9.2e\n", sqrt(frms / ((REAL_T)k)));
#endif

  frms = 0.0;
  for (i = 0; i < n; i++)
    frms += f[i] * f[i];
  frms = sqrt(frms / n);

  if (*iter == 1 || *iter % ntpr == 0) {
    printf("ff:%4d %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2e\n",
	   *iter, ene[0] + esurf, ene[3] + ene[4] + ene[5],
	   ene[1] + ene[7], ene[2] + ene[8], ene[9], ene[10], frms);
    fflush(stdout);
  }

  /*
   * Load the atomic masses into the array m because the calling
   * function (newton) needs these masses to "level shift" to high
   * frequencies the eigenvalues associated with overall translation
   * and rotation of the molecule.
   *
   * Load the number of atoms as well.
   */

  for (i = 0; i < prm->Natom; i++) {
    m[i] = prm->Masses[i];
  }

  *natom = prm->Natom;

#ifdef MM_DETAILS
	printf( "Inside mme2():\n" );
	printf( "     bond:  %15.9f\n", ene[3] );
	printf( "    angle:  %15.9f\n", ene[4] );
	printf( " dihedral:  %15.9f\n", ene[5] );
	printf( "    enb14:  %15.9f\n", ene[7] );
	printf( "    eel14:  %15.9f\n", ene[8] );
	printf( "      enb:  %15.9f\n", ene[1] );
	printf( "      eel:  %15.9f\n", ene[2] );
	printf( "      egb:  %15.9f\n", ene[10] );
	printf( "    Total:  %15.9f\n", ene[0] );
#endif
  return (ene[0] + esurf);
}


